@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * Config.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -22,10 +22,14 @@
 #include "LangTables.hh"
 #include "MainWindow.hh"
 #include "Utils.hh"
+#include "OutputEditorText.hh"
 
 #include <enchant-provider.h>
 #define USE_STD_NAMESPACE
 #include <tesseract/baseapi.h>
+#if TESSERACT_MAJOR_VERSION < 5
+#include <tesseract/genericvector.h>
+#endif
 #undef USE_STD_NAMESPACE
 
 const std::vector<Config::Lang> Config::LANGUAGES = LangTables::languages<std::vector<Config::Lang>, Glib::ustring>([](const char* str) { return Glib::ustring(str); });
@@ -42,19 +46,25 @@ Config::Config() {
 	ui.setupUi();
 	ui.dialogConfig->set_transient_for(*MAIN->getWindow());
 
-	for(Gtk::TreeView* view : {
-	            ui.treeviewLangsPredef, ui.treeviewLangsCustom
-	        }) {
-		view->set_model(Gtk::ListStore::create(m_langViewCols));
-		view->append_column_editable(_("Filename prefix"), m_langViewCols.prefix);
-		view->append_column_editable(_("Code"), m_langViewCols.code);
-		view->append_column_editable(_("Native name"), m_langViewCols.name);
-		for(int i = 0; i < 3; ++i) {
-			view->get_column(i)->set_sizing(Gtk::TREE_VIEW_COLUMN_FIXED);
-			view->get_column(i)->set_expand(true);
-		}
-		view->set_fixed_height_mode(true);
+	ui.treeviewLangsPredef->set_model(Gtk::ListStore::create(m_langViewCols));
+	ui.treeviewLangsPredef->append_column(_("Filename prefix"), m_langViewCols.prefix);
+	ui.treeviewLangsPredef->append_column(_("Code"), m_langViewCols.code);
+	ui.treeviewLangsPredef->append_column(_("Native name"), m_langViewCols.name);
+	for(int i = 0; i < 3; ++i) {
+		ui.treeviewLangsPredef->get_column(i)->set_sizing(Gtk::TREE_VIEW_COLUMN_FIXED);
+		ui.treeviewLangsPredef->get_column(i)->set_expand(true);
 	}
+	ui.treeviewLangsPredef->set_fixed_height_mode(true);
+
+	ui.treeviewLangsCustom->set_model(Gtk::ListStore::create(m_langViewCols));
+	ui.treeviewLangsCustom->append_column_editable(_("Filename prefix"), m_langViewCols.prefix);
+	ui.treeviewLangsCustom->append_column_editable(_("Code"), m_langViewCols.code);
+	ui.treeviewLangsCustom->append_column_editable(_("Native name"), m_langViewCols.name);
+	for(int i = 0; i < 3; ++i) {
+		ui.treeviewLangsCustom->get_column(i)->set_sizing(Gtk::TREE_VIEW_COLUMN_FIXED);
+		ui.treeviewLangsCustom->get_column(i)->set_expand(true);
+	}
+	ui.treeviewLangsCustom->set_fixed_height_mode(true);
 
 	Glib::RefPtr<Gtk::ListStore> store = Glib::RefPtr<Gtk::ListStore>::cast_static(ui.treeviewLangsPredef->get_model());
 	for(const auto& lang : LANGUAGES) {
@@ -145,6 +155,37 @@ std::string Config::spellingLocation() const {
 	return ui.entrySpelldir->get_text();
 }
 
+std::vector<Glib::ustring> Config::getAvailableLanguages() {
+	Utils::TesseractHandle tess;
+	if(!tess.get()) {
+		return std::vector<Glib::ustring>();
+	}
+#if TESSERACT_MAJOR_VERSION < 5
+	GenericVector<STRING> availLanguages;
+#else
+	std::vector<std::string> availLanguages;
+#endif
+	tess.get()->GetAvailableLanguagesAsVector(&availLanguages);
+	std::vector<Glib::ustring> result;
+	for(int i = 0; i < availLanguages.size(); ++i) {
+#if TESSERACT_MAJOR_VERSION < 5
+		result.push_back(availLanguages[i].string());
+#else
+		result.push_back(availLanguages[i]);
+#endif
+	}
+	std::sort(result.begin(), result.end(), [](const Glib::ustring & s1, const Glib::ustring & s2) {
+		bool s1Script = s1.substr(0, 6) == "script" || s1.substr(0, 1) == s1.substr(0, 1).uppercase();
+		bool s2Script = s2.substr(0, 6) == "script" || s2.substr(0, 1) == s2.substr(0, 1).uppercase();
+		if(s1Script != s2Script) {
+			return !s1Script;
+		} else {
+			return s1 < s2;
+		}
+	});
+	return result;
+}
+
 void Config::showDialog() {
 	toggleAddLanguage(true);
 	while(ui.dialogConfig->run() == Gtk::RESPONSE_HELP);
@@ -203,7 +244,7 @@ std::string Config::tessdataLocation(Location location) {
 	if(location == SystemLocation) {
 #ifdef G_OS_WIN32
 		std::string dataDir = Glib::build_filename(pkgDir, "share");
-		Glib::setenv("TESSDATA_PREFIX",  Glib::build_filename(dataDir, "tessdata");
+		Glib::setenv("TESSDATA_PREFIX",  Glib::build_filename(dataDir, "tessdata"));
 #else
 		Glib::unsetenv("TESSDATA_PREFIX");
 #endif
