@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * CrashHandler.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -103,7 +103,7 @@ void CrashHandler::generate_backtrace_end(bool success) {
 	ui.textview->set_sensitive(true);
 	ui.buttonRegenerate->set_sensitive(true);
 	if(!success) {
-		ui.textview->get_buffer()->set_text(_("Failed to obtain backtrace. Is gdb installed?"));
+		ui.textview->get_buffer()->set_text(_("Failed to obtain backtrace. Is GDB installed?"));
 	} else {
 		std::vector<Glib::ustring> lines = Utils::string_split(ui.textview->get_buffer()->get_text(false), '\n', false);
 		Glib::ustring text = Glib::ustring::compose("%1\n\n", lines[0]);
