@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * FontComboBox.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -22,12 +22,12 @@
 FontComboBox::FontComboColums FontComboBox::s_fontComboCols;
 
 FontComboBox::FontComboBox()
-	: Gtk::ComboBox(true), Glib::ObjectBase("FontComboBox") {
+	: Glib::ObjectBase("FontComboBox"), Gtk::ComboBox(true) {
 	init();
 }
 
 FontComboBox::FontComboBox(BaseObjectType* cobject, const Glib::RefPtr<Gtk::Builder>& builder)
-	: Gtk::ComboBox(cobject), Glib::ObjectBase("FontComboBox") {
+	: Glib::ObjectBase("FontComboBox"), Gtk::ComboBox(cobject) {
 	init();
 }
 
