@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * MainWindow.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -27,6 +27,8 @@
 #include "DisplayerToolSelect.hh"
 #include "OutputEditorHOCR.hh"
 #include "OutputEditorText.hh"
+#include "HOCRBatchExportDialog.hh"
+#include "RecognitionMenu.hh"
 #include "Recognizer.hh"
 #include "SourceManager.hh"
 #include "TessdataManager.hh"
@@ -74,14 +76,16 @@ void MainWindow::tesseractCrash(int signal) {
 void MainWindow::signalHandlerExec(int sig, bool tesseractCrash) {
 	std::signal(sig, nullptr);
 	std::string filename;
-	if(MAIN->getOutputEditor() && MAIN->getOutputEditor()->getModified()) {
+	if(MAIN->getOutputEditor()) {
 		filename = Glib::build_filename(g_get_home_dir(), Glib::ustring::compose("%1_crash-save.txt", PACKAGE_NAME));
 		int i = 0;
 		while(Glib::file_test(filename, Glib::FILE_TEST_EXISTS)) {
 			++i;
 			filename = Glib::build_filename(g_get_home_dir(), Glib::ustring::compose("%1_crash-save_%2.txt", PACKAGE_NAME, i));
 		}
-		MAIN->getOutputEditor()->save(filename);
+		if(!MAIN->getOutputEditor()->crashSave(filename)) {
+			filename.clear();
+		}
 	}
 	Glib::Pid pid;
 	Glib::spawn_async("", std::vector<std::string> {pkgExePath, "crashhandle", Glib::ustring::compose("%1", getpid()), Glib::ustring::compose("%1", tesseractCrash), filename}, Glib::SPAWN_DO_NOT_REAP_CHILD, sigc::slot<void>(), &pid);
@@ -133,6 +137,7 @@ MainWindow::MainWindow() {
 	m_config = new Config;
 	m_acquirer = new Acquirer(ui);
 	m_displayer = new Displayer(ui);
+	m_recognitionMenu = new RecognitionMenu();
 	m_recognizer = new Recognizer(ui);
 	m_sourceManager = new SourceManager(ui);
 
@@ -148,7 +153,6 @@ MainWindow::MainWindow() {
 	m_idlegroup.push_back(ui.spinContrast);
 	m_idlegroup.push_back(ui.spinResolution);
 	m_idlegroup.push_back(ui.buttonAutolayout);
-	m_idlegroup.push_back(ui.menubuttonLanguages);
 
 	CONNECT(ui.windowMain, delete_event, [this](GdkEventAny * ev) {
 		return closeEvent(ev);
@@ -159,8 +163,8 @@ MainWindow::MainWindow() {
 	});
 	CONNECT(m_sourceManager, sourceChanged, [this] { onSourceChanged(); });
 	CONNECT(ui.buttonOutputpane, toggled, [this] { if(m_outputEditor) m_outputEditor->getUI()->set_visible(ui.buttonOutputpane->get_active()); });
-	m_connection_setOCRMode = CONNECT(ui.comboOcrmode, changed, [this] { setOCRMode(ui.comboOcrmode->get_active_row_number()); });
-	CONNECT(m_recognizer, languageChanged, [this] (const Config::Lang & lang) {
+	m_connection_setOCRMode = CONNECT(ui.comboOcrmode, changed, [this] { setOutputMode(static_cast<OutputMode>(ui.comboOcrmode->get_active_row_number())); });
+	CONNECT(m_recognitionMenu, languageChanged, [this] (const Config::Lang & lang) {
 		languageChanged(lang);
 	});
 	CONNECT(ConfigSettings::get<ComboSetting>("outputorient"), changed, [this] {
@@ -168,12 +172,14 @@ MainWindow::MainWindow() {
 	});
 	CONNECT(ui.buttonProgressCancel, clicked, [this] { progressCancel(); });
 	CONNECT(ui.buttonAutolayout, clicked, [this] { m_displayer->autodetectOCRAreas(); });
+	CONNECT(ui.buttonBatchExport, clicked, [this] { batchExport(); });
 
 	ADD_SETTING(VarSetting<std::vector<int>>("wingeom"));
 	ADD_SETTING(SwitchSettingT<Gtk::ToggleButton>("showcontrols", ui.buttonControls));
 	ADD_SETTING(ComboSetting("outputeditor", ui.comboOcrmode));
 
-	m_recognizer->updateLanguagesMenu();
+	ui.menubuttonLanguages->set_menu(*m_recognitionMenu);
+	m_recognitionMenu->rebuild();
 
 	pushState(State::Idle, _("Select an image to begin..."));
 
@@ -185,7 +191,7 @@ MainWindow::MainWindow() {
 	ui.panedOutput->set_orientation(static_cast<Gtk::Orientation>(!ConfigSettings::get<ComboSetting>("outputorient")->getValue()));
 #if ENABLE_VERSIONCHECK
 	if(ConfigSettings::get<SwitchSetting>("updatecheck")->getValue()) {
-		m_newVerThread = Glib::Threads::Thread::create([this] { getNewestVersion(); });
+		m_newVerThread = new std::thread(&MainWindow::getNewestVersion, this);
 	}
 #endif
 }
@@ -194,6 +200,7 @@ MainWindow::~MainWindow() {
 #if ENABLE_VERSIONCHECK
 	if(m_newVerThread) {
 		m_newVerThread->join();
+		delete m_newVerThread;
 	}
 #endif
 	delete m_acquirer;
@@ -204,6 +211,7 @@ MainWindow::~MainWindow() {
 	delete m_displayer;
 	delete m_recognizer;
 	delete m_config;
+	delete m_recognitionMenu;
 	s_instance = nullptr;
 }
 
@@ -213,7 +221,35 @@ void MainWindow::setMenuModel(const Glib::RefPtr<Gio::MenuModel>& menuModel) {
 }
 
 void MainWindow::openFiles(const std::vector<Glib::RefPtr<Gio::File>>& files) {
-	m_sourceManager->addSources(files);
+	std::vector<Glib::RefPtr<Gio::File>> hocrFiles;
+	std::vector<Glib::RefPtr<Gio::File>> otherFiles;
+	for(const Glib::RefPtr<Gio::File>& file : files) {
+		std::string filename = file->get_path();
+		if(Glib::ustring(filename.substr(filename.length() - 5)).lowercase() == ".html") {
+			hocrFiles.push_back(file);
+		} else {
+			otherFiles.push_back(file);
+		}
+	}
+	m_sourceManager->addSources(otherFiles);
+	if(!hocrFiles.empty()) {
+		if(setOutputMode(OutputModeHOCR)) {
+			static_cast<OutputEditorHOCR*>(m_outputEditor)->open(OutputEditorHOCR::InsertMode::Append, hocrFiles);
+			m_outputEditor->getUI()->set_visible(true);
+		}
+	}
+}
+
+void MainWindow::openOutput(const std::string& filename) {
+	if(Utils::string_endswith(Glib::ustring(filename).lowercase(), ".txt")) {
+		if(setOutputMode(OutputModeText)) {
+			m_outputEditor->open(filename);
+		}
+	} else if(Utils::string_endswith(Glib::ustring(filename).lowercase(), ".html")) {
+		if(setOutputMode(OutputModeHOCR)) {
+			m_outputEditor->open(filename);
+		}
+	}
 }
 
 void MainWindow::setOutputPaneVisible(bool visible) {
@@ -272,11 +308,11 @@ void MainWindow::progressUpdate() {
 	}
 }
 
-
 bool MainWindow::closeEvent(GdkEventAny*) {
 	if(!m_outputEditor->clear()) {
 		return true;
 	}
+
 	if((ui.windowMain->get_window()->get_state() & Gdk::WINDOW_STATE_MAXIMIZED) == 0) {
 		std::vector<int> geom(4);
 		ui.windowMain->get_position(geom[0], geom[1]);
@@ -289,8 +325,8 @@ bool MainWindow::closeEvent(GdkEventAny*) {
 
 void MainWindow::onSourceChanged() {
 	std::vector<Source*> sources = m_sourceManager->getSelectedSources();
-	if(m_displayer->setSources(sources)) {
-		ui.headerbar->set_subtitle(sources.size() == 1 ? sources.front()->displayname : _("Multiple sources"));
+	if(m_displayer->setSources(sources) && !sources.empty()) {
+		ui.headerbar->set_subtitle(sources.size() == 1 ? Glib::ustring(sources.front()->displayname) : Glib::ustring::compose(_("Multiple sources (%1)"), sources.size()));
 		if(m_stateStack.back() == State::Idle) {
 			pushState(State::Normal, _("Ready"));
 		}
@@ -330,41 +366,46 @@ void MainWindow::showHelp(const std::string& chapter) {
 
 void MainWindow::showConfig() {
 	m_config->showDialog();
-	m_recognizer->updateLanguagesMenu();
+	m_recognitionMenu->rebuild();
 }
 
-void MainWindow::setOCRMode(int idx) {
+bool MainWindow::setOutputMode(OutputMode mode) {
 	if(m_outputEditor && !m_outputEditor->clear()) {
 		m_connection_setOCRMode.block(true);
 		if(dynamic_cast<OutputEditorText*>(m_outputEditor)) {
-			ui.comboOcrmode->set_active(0);
+			ui.comboOcrmode->set_active(OutputModeText);
 		} else if(dynamic_cast<OutputEditorHOCR*>(m_outputEditor)) {
-			ui.comboOcrmode->set_active(1);
+			ui.comboOcrmode->set_active(OutputModeHOCR);
 		}
 		m_connection_setOCRMode.block(false);
+		return false;
 	} else {
 		if(m_outputEditor) {
 			ui.panedOutput->remove(*m_outputEditor->getUI());
 			delete m_outputEditor;
 		}
 		delete m_displayerTool;
-		if(idx == 0) {
+		if(mode == OutputModeText) {
 			m_displayerTool = new DisplayerToolSelect(m_displayer);
 			m_outputEditor = new OutputEditorText();
-		} else { /*if(idx == 1)*/
+		} else { /*if(mode == OutputModeHOCR)*/
 			m_displayerTool = new DisplayerToolHOCR(m_displayer);
 			m_outputEditor = new OutputEditorHOCR(static_cast<DisplayerToolHOCR*>(m_displayerTool));
 		}
 		ui.buttonAutolayout->set_visible(m_displayerTool->allowAutodetectOCRAreas());
 		m_displayer->setTool(m_displayerTool);
-		m_outputEditor->setLanguage(m_recognizer->getSelectedLanguage());
+		m_outputEditor->setLanguage(m_recognitionMenu->getRecognitionLanguage());
 		ui.panedOutput->pack2(*m_outputEditor->getUI(), true, false);
 		m_outputEditor->getUI()->set_visible(ui.buttonOutputpane->get_active());
+		m_connection_setOCRMode.block(true);
+		ui.comboOcrmode->set_active(mode);
+		m_connection_setOCRMode.block(false);
+		return true;
 	}
 }
 
 void MainWindow::redetectLanguages() {
-	m_recognizer->updateLanguagesMenu();
+	m_recognitionMenu->rebuild();
 }
 
 void MainWindow::manageLanguages() {
@@ -422,7 +463,7 @@ void MainWindow::getNewestVersion() {
 		Glib::RefPtr<Gio::File> file = Gio::File::create_for_uri(CHECKURL);
 		Glib::RefPtr<Gio::FileInputStream> stream = file->read();
 		stream->read(buf, 16);
-	} catch (const Glib::Error&) {
+	} catch (const Glib::Error& e) {
 		return;
 	}
 	std::string newver(buf);
@@ -442,7 +483,7 @@ void MainWindow::checkVersion(const Glib::ustring& newver) {
 		addNotification(_("New version"), Glib::ustring::compose(_("gImageReader %1 is available"), newver), {
 			{_("Download"), [ = ]{ Utils::openUri(DOWNLOADURL); return false; }},
 			{_("Changelog"), [ = ]{ Utils::openUri(CHANGELOGURL); return false; }},
-			{_("Don't notify again"), [this]{ ConfigSettings::get<SwitchSetting>("updatecheck")->setValue(false); return true; }}
+			{_("Don't notify again"), []{ ConfigSettings::get<SwitchSetting>("updatecheck")->setValue(false); return true; }}
 		});
 	}
 }
@@ -463,7 +504,7 @@ void MainWindow::languageChanged(const Config::Lang& lang) {
 		checker.set_language(code);
 	} catch(const GtkSpell::Error& /*e*/) {
 		if(ConfigSettings::get<SwitchSetting>("dictinstall")->getValue()) {
-			NotificationAction actionDontShowAgain = {_("Don't show again"), [this]{ ConfigSettings::get<SwitchSetting>("dictinstall")->setValue(false); return true; }};
+			NotificationAction actionDontShowAgain = {_("Don't show again"), []{ ConfigSettings::get<SwitchSetting>("dictinstall")->setValue(false); return true; }};
 			NotificationAction actionInstall = NotificationAction{_("Install"), [this, lang]{ dictionaryAutoinstall(lang.code); return false; }};
 #ifdef G_OS_UNIX
 			if(getConfig()->useSystemDataLocations()) {
@@ -512,9 +553,9 @@ void MainWindow::dictionaryAutoinstallDone(Glib::RefPtr<Gio::DBus::Proxy> proxy,
 	try {
 		proxy->call_finish(result);
 	} catch (const Glib::Error& e) {
-		Utils::message_dialog(Gtk::MESSAGE_ERROR, _("Error"), Glib::ustring::compose(_("Failed to install spelling dictionary: %1"), e.what()));
+		Utils::messageBox(Gtk::MESSAGE_ERROR, _("Error"), Glib::ustring::compose(_("Failed to install spelling dictionary: %1"), e.what()));
 	}
-	getRecognizer()->updateLanguagesMenu();
+	m_recognitionMenu->rebuild();
 	popState();
 }
 #endif
@@ -533,7 +574,7 @@ void MainWindow::dictionaryAutoinstall(Glib::ustring code) {
 	}
 	if(!dirExists) {
 		popState();
-		Utils::message_dialog(Gtk::MESSAGE_ERROR, _("Error"), _("Failed to create directory for spelling dictionaries."));
+		Utils::messageBox(Gtk::MESSAGE_ERROR, _("Error"), _("Failed to create directory for spelling dictionaries."));
 		return;
 	}
 
@@ -541,7 +582,7 @@ void MainWindow::dictionaryAutoinstall(Glib::ustring code) {
 	Glib::RefPtr<Glib::ByteArray> htmlData = Utils::download(url, messages);
 	if(!htmlData) {
 		popState();
-		Utils::message_dialog(Gtk::MESSAGE_ERROR, _("Error"), Glib::ustring::compose(_("Could not read %1: %2."), url, messages));
+		Utils::messageBox(Gtk::MESSAGE_ERROR, _("Error"), Glib::ustring::compose(_("Could not read %1: %2."), url, messages));
 		return;
 	}
 	Glib::ustring html(reinterpret_cast<const char*>(htmlData->get_data()), htmlData->size());
@@ -589,11 +630,16 @@ void MainWindow::dictionaryAutoinstall(Glib::ustring code) {
 
 	popState();
 	if(!failed.empty()) {
-		Utils::message_dialog(Gtk::MESSAGE_ERROR, _("Error"), Glib::ustring::compose(_("The following dictionaries could not be downloaded:\n%1\n\nCheck the connectivity and directory permissions.\nHint: If you don't have write permissions in system folders, you can switch to user paths in the settings dialog."), Utils::string_join(failed, "\n")));
+		Utils::messageBox(Gtk::MESSAGE_ERROR, _("Error"), Glib::ustring::compose(_("The following dictionaries could not be downloaded:\n%1\n\nCheck the connectivity and directory permissions.\nHint: If you don't have write permissions in system folders, you can switch to user paths in the settings dialog."), Utils::string_join(failed, "\n")));
 	} else if(!downloaded.empty()) {
-		m_recognizer->updateLanguagesMenu();
-		Utils::message_dialog(Gtk::MESSAGE_INFO, _("Dictionaries installed"), Glib::ustring::compose(_("The following dictionaries were installed:\n%1"), Utils::string_join(downloaded, "\n")));
+		m_recognitionMenu->rebuild();
+		Utils::messageBox(Gtk::MESSAGE_INFO, _("Dictionaries installed"), Glib::ustring::compose(_("The following dictionaries were installed:\n%1"), Utils::string_join(downloaded, "\n")));
 	} else {
-		Utils::message_dialog(Gtk::MESSAGE_ERROR, _("Error"), Glib::ustring::compose(_("No spelling dictionaries found for '%1'."), code));
+		Utils::messageBox(Gtk::MESSAGE_ERROR, _("Error"), Glib::ustring::compose(_("No spelling dictionaries found for '%1'."), code));
 	}
 }
+
+void MainWindow::batchExport() {
+	HOCRBatchExportDialog dialog;
+	dialog.run();
+}
