@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * Utils.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -23,7 +23,13 @@
 #include "SourceManager.hh"
 
 #include <clocale>
+#include <csignal>
 #include <fontconfig/fontconfig.h>
+
+#ifdef G_OS_WIN32
+#include <windows.h>
+#endif
+
 #define USE_STD_NAMESPACE
 #include <tesseract/baseapi.h>
 #undef USE_STD_NAMESPACE
@@ -42,40 +48,113 @@ void Utils::popup_positioner(int& x, int& y, bool& push_in, Gtk::Widget* ref, Gt
 	push_in = true;
 }
 
-void Utils::message_dialog(Gtk::MessageType message, const Glib::ustring& title, const Glib::ustring& text, Gtk::Window* parent) {
+Utils::Button::Type Utils::messageBox(Gtk::MessageType type,
+                                      const Glib::ustring& title,
+                                      const Glib::ustring& text,
+                                      const Glib::ustring& body,
+                                      int buttons,
+                                      Gtk::Window* parent,
+                                      Gtk::Widget* bodyWidget) {
 	if(!parent) {
 		parent = MAIN->getWindow();
 	}
-	Gtk::MessageDialog dialog(*parent, title, false, message, Gtk::BUTTONS_OK, true);
-	dialog.set_secondary_text(text);
-	dialog.run();
-}
 
-Utils::Button::Type Utils::question_dialog(const Glib::ustring& title, const Glib::ustring& text, int buttons, Gtk::Window* parent) {
-	if(!parent) {
-		parent = MAIN->getWindow();
+	Gtk::Dialog dialog;
+	dialog.set_transient_for(*parent);
+	dialog.set_title(title);
+	dialog.get_content_area()->set_orientation(Gtk::ORIENTATION_VERTICAL);
+	dialog.get_content_area()->set_spacing(6);
+	dialog.get_content_area()->set_margin_left(6);
+	dialog.get_content_area()->set_margin_right(6);
+	dialog.get_content_area()->set_margin_top(6);
+	dialog.get_content_area()->set_margin_bottom(6);
+
+	Gtk::Grid grid;
+	dialog.get_content_area()->pack_start(grid, true, true);
+
+	Glib::ustring iconName;
+	switch (type) {
+	case Gtk::MESSAGE_INFO:
+		iconName = "dialog-information-symbolic";
+		break;
+	case Gtk::MESSAGE_WARNING:
+		iconName = "dialog-warning-symbolic";
+		break;
+	case Gtk::MESSAGE_QUESTION:
+		iconName = "dialog-question-symbolic";
+		break;
+	case Gtk::MESSAGE_ERROR:
+		iconName = "dialog-error-symbolic";
+		break;
+	case Gtk::MESSAGE_OTHER:
+		break;
 	}
-	Gtk::MessageDialog dialog(*parent, title, false, Gtk::MESSAGE_QUESTION, Gtk::BUTTONS_NONE, true);
-	dialog.set_position(Gtk::WIN_POS_CENTER_ON_PARENT);
+	Gtk::Image icon(iconName, Gtk::ICON_SIZE_DIALOG);
+	grid.attach(icon, 0, 0, 1, body.empty() ? 1 : 2);
+
+
+	Gtk::Label label(text);
+	label.set_xalign(0);
+	grid.attach(label, 1, 0);
+	grid.set_column_spacing(6);
+	grid.set_row_spacing(6);
+
+	Gtk::ScrolledWindow scrollArea;
+	scrollArea.set_hexpand(true);
+	scrollArea.set_vexpand(true);
+	scrollArea.set_shadow_type(Gtk::SHADOW_IN);
+	scrollArea.set_min_content_width(320);
+	scrollArea.set_min_content_height(160);
+
+	Gtk::TextView textView;
+	if(!body.empty()) {
+		textView.set_editable(false);
+		textView.set_cursor_visible(false);
+		Glib::RefPtr<Gtk::TextBuffer> buffer = Gtk::TextBuffer::create();
+		buffer->set_text(body);
+		textView.set_buffer(buffer);
+		scrollArea.add(textView);
+		grid.attach(scrollArea, 1, 1);
+	} else if (bodyWidget) {
+		scrollArea.add(*bodyWidget);
+		grid.attach(scrollArea, 1, 1);
+	}
+
 	if((buttons & Button::Ok) != 0) {
 		dialog.add_button(_("OK"), Button::Type::Ok);
 	}
 	if((buttons & Button::Yes) != 0) {
 		dialog.add_button(_("Yes"), Button::Type::Yes);
 	}
-	if((buttons & Button::No) != 0) {
-		dialog.add_button(_("No"), Button::Type::No);
-	}
-	if((buttons & Button::Cancel) != 0) {
-		dialog.add_button(_("Cancel"), Button::Type::Cancel);
+	if((buttons & Button::YesAll) != 0) {
+		dialog.add_button(_("Yes to all"), Button::Type::YesAll);
 	}
 	if((buttons & Button::Save) != 0) {
 		dialog.add_button(_("Save"), Button::Type::Save);
 	}
+	if((buttons & Button::No) != 0) {
+		dialog.add_button(_("No"), Button::Type::No);
+	}
+	if((buttons & Button::NoAll) != 0) {
+		dialog.add_button(_("No to all"), Button::Type::NoAll);
+	}
 	if((buttons & Button::Discard) != 0) {
 		dialog.add_button(_("Discard"), Button::Type::Discard);
 	}
-	dialog.set_secondary_text(text);
+	if((buttons & Button::DiscardAll) != 0) {
+		dialog.add_button(_("Discard all"), Button::Type::DiscardAll);
+	}
+	if((buttons & Button::Text) != 0) {
+		dialog.add_button(_("Text"), Button::Type::Text);
+	}
+	if((buttons & Button::HOCR) != 0) {
+		dialog.add_button(_("hOCR"), Button::Type::HOCR);
+	}
+	if((buttons & Button::Cancel) != 0) {
+		dialog.add_button(_("Cancel"), Button::Type::Cancel);
+	}
+
+	dialog.show_all();
 	return static_cast<Button::Type>(dialog.run());
 }
 
@@ -132,9 +211,11 @@ std::string Utils::make_absolute_path(const std::string& path, const std::string
 		return path;
 	}
 	std::string abspath = Glib::build_filename(basepath, path);
+#ifdef G_OS_UNIX
 	char* realabspath = realpath(abspath.c_str(), nullptr);
 	abspath = std::string(realabspath);
 	free(realabspath);
+#endif
 	return abspath;
 }
 
@@ -168,7 +249,7 @@ std::string Utils::make_output_filename(const std::string& filename) {
 	// Generate non-existing file
 	int i = 0;
 	std::pair<std::string, std::string> parts = split_filename(newfilename);
-	parts.first = Glib::Regex::create("_[0-9]+$")->replace(parts.first, 0, "", static_cast<Glib::RegexMatchFlags>(0));
+	parts.first = Glib::Regex::create(Glib::ustring::compose("_[0-9]+.%1$", parts.second))->replace(parts.first, 0, "", static_cast<Glib::RegexMatchFlags>(0));
 	newfilename = Glib::ustring::compose("%1.%2", parts.first, parts.second);
 	while(Glib::file_test(newfilename, Glib::FILE_TEST_EXISTS)) {
 		newfilename = Glib::ustring::compose("%1_%2.%3", parts.first, ++i, parts.second);
@@ -176,10 +257,34 @@ std::string Utils::make_output_filename(const std::string& filename) {
 	return newfilename;
 }
 
+void Utils::list_dir(const std::string& path, const std::set<std::string>& filters, std::vector<Glib::RefPtr<Gio::File>>& output) {
+	try {
+		Glib::Dir dir(path);
+		std::vector<Glib::RefPtr<Gio::File>> files;
+		for(const std::string& filename : dir) {
+			std::string subpath = Glib::build_filename(path, filename);
+			if(Glib::file_test(subpath, Glib::FILE_TEST_IS_DIR)) {
+				list_dir(subpath, filters, output);
+			} else {
+				std::pair<std::string, std::string> parts = Utils::split_filename(filename);
+				if(filters.find(parts.second) != filters.end()) {
+					output.push_back(Gio::File::create_for_path(subpath));
+				}
+			}
+		}
+	} catch(const Glib::FileError& e) { }
+}
+
 std::vector<Glib::ustring> Utils::string_split(const Glib::ustring& text, char delim, bool keepEmpty) {
 	std::vector<Glib::ustring> parts;
 	Glib::ustring::size_type startPos = 0, endPos = 0;
 	Glib::ustring::size_type npos = Glib::ustring::npos;
+	if(text.empty()) {
+		if(keepEmpty) {
+			parts.push_back("");
+		}
+		return parts;
+	}
 	while(true) {
 		startPos = endPos;
 		endPos = text.find(delim, startPos);
@@ -214,10 +319,31 @@ Glib::ustring Utils::string_trim(const Glib::ustring& str, const Glib::ustring&
 	return ret;
 }
 
+Glib::ustring Utils::string_rstrip(const Glib::ustring& str, const Glib::ustring& what) {
+	Glib::ustring ret = str;
+	std::size_t rpos = ret.find_last_not_of(what);
+	if(rpos != Glib::ustring::npos) {
+		ret.erase(rpos + 1);
+	}
+	return ret;
+}
+
 bool Utils::strings_equal(const Glib::ustring& str1, const Glib::ustring& str2, bool matchCase) {
 	return matchCase ? (str1 == str2) : (str1.casefold() == str2.casefold());
 }
 
+bool Utils::string_endswith(const Glib::ustring& str, const Glib::ustring& what) {
+	int n = str.length();
+	int m = what.length();
+	return n >= m && str.substr(n - m) == what;
+}
+
+bool Utils::string_startswith(const Glib::ustring& str, const Glib::ustring& what) {
+	int n = str.length();
+	int m = what.length();
+	return n >= m && str.substr(0, m) == what;
+}
+
 std::size_t Utils::string_firstIndex(const Glib::ustring& str, const Glib::ustring& search, int pos, bool matchCase) {
 	std::size_t res = Glib::ustring::npos;
 	if(matchCase) {
@@ -282,14 +408,20 @@ Glib::ustring Utils::string_html_escape(const Glib::ustring& str) {
 }
 
 std::vector<std::pair<Glib::ustring, int>> Utils::string_split_pos(const Glib::ustring& str, const Glib::RefPtr<Glib::Regex>& splitRe) {
-	int strpos = 0;
+	Glib::ustring::size_type strpos = 0;
 	Glib::MatchInfo info;
 	std::vector<std::pair<Glib::ustring, int>> result;
-	while(splitRe->match(str, strpos, info)) {
+
+	Glib::ustring matchstr = str;
+	while(splitRe->match(matchstr, 0, info)) {
+		// Fetch pos returns positions in bytes, no characters... Need to extract substring on raw string, and then deduce the new position by comparing the lengths
 		int start = 0, end = 0;
 		info.fetch_pos(0, start, end);
-		result.push_back(std::make_pair(str.substr(strpos, start - strpos), strpos));
-		strpos = end;
+		result.push_back(std::make_pair(str.raw().substr(0, start), strpos));
+
+		Glib::ustring newmatchstr = matchstr.raw().substr(end);
+		strpos += matchstr.length() - newmatchstr.length();
+		matchstr = newmatchstr;
 	}
 	if(strpos < str.length()) {
 		result.push_back(std::make_pair(str.substr(strpos), strpos));
@@ -331,7 +463,8 @@ Glib::RefPtr<Glib::ByteArray> Utils::download(const std::string& url, Glib::ustr
 			try {
 				stream = file->read_finish(asyncResult);
 				status = stream ? Ready : Failed;
-			} catch(Glib::Error&) {
+			} catch(Glib::Error& e) {
+				messages = e.what();
 				status = Failed;
 			}
 		}, cancellable);
@@ -380,19 +513,33 @@ Glib::RefPtr<Glib::ByteArray> Utils::download(const std::string& url, Glib::ustr
 }
 
 
-Glib::ustring Utils::getSpellingLanguage(const Glib::ustring& lang) {
+Glib::ustring Utils::getSpellingLanguage(const Glib::ustring& lang, const Glib::ustring& defaultLanguage) {
 	// If it is already a valid code, return it
-	static Glib::RefPtr<Glib::Regex> langPattern = Glib::Regex::create("^[a-z]{2,}(_[A-Z]{2,})?$");
+	static Glib::RefPtr<Glib::Regex> langPattern = Glib::Regex::create("^[a-z]{2}(_[A-Z]{2})?$");
 	if(langPattern->match(lang)) {
 		return lang;
 	}
 	// Treat the language as a tesseract lang spec and try to find a matching code
-	Config::Lang langspec = {lang};
+	Config::Lang langspec = {lang, "", ""};
 	if(!lang.empty() && MAIN->getConfig()->searchLangSpec(langspec)) {
+		// If default language starts with the returned code, return default language
+		if(!defaultLanguage.empty() && Utils::string_startswith(defaultLanguage, langspec.code)) {
+			return defaultLanguage;
+		}
+		// Return any one language culture code
+		std::vector<Glib::ustring> langCultures = MAIN->getConfig()->searchLangCultures(langspec.code);
+		if(!langCultures.empty()) {
+			return langCultures.front();
+		}
+		// Return incomplete code (only country)
 		return langspec.code;
 	}
+	// Return default language if specified
+	if(!defaultLanguage.empty()) {
+		return defaultLanguage;
+	}
 	// Use the application locale, if specified, otherwise fall back to en
-	Glib::ustring syslocale = g_getenv ("LANG");
+	Glib::ustring syslocale = Glib::getenv("LANG");
 	if(syslocale == "c" || syslocale == "C" || syslocale.empty()) {
 		return "en_US";
 	}
@@ -478,3 +625,24 @@ void Utils::runInMainThreadBlocking(const std::function<void()>& f) {
 	}
 	mutex.unlock();
 }
+
+Utils::TesseractHandle::TesseractHandle(const char* language) {
+	// unfortunately tesseract creates deliberate aborts when an error occurs
+	std::signal(SIGABRT, MainWindow::tesseractCrash);
+	std::string current = setlocale(LC_ALL, NULL);
+	setlocale(LC_ALL, "C");
+	m_tess = new tesseract::TessBaseAPI();
+	int ret = m_tess->Init(nullptr, language);
+	setlocale(LC_ALL, current.c_str());
+
+	if(ret == -1) {
+		delete m_tess;
+		m_tess = nullptr;
+	}
+}
+
+Utils::TesseractHandle::~TesseractHandle() {
+	delete m_tess;
+	std::signal(SIGABRT, MainWindow::signalHandler);
+}
+
