@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * HOCRDocument.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -19,6 +19,7 @@
 
 #include "common.hh"
 #include "HOCRDocument.hh"
+#include "HOCRSpellChecker.hh"
 #include "XmlUtils.hh"
 #include "Utils.hh"
 
@@ -28,20 +29,70 @@
 
 #define DEBUG(...) //__VA_ARGS__
 
-HOCRDocument::HOCRDocument(GtkSpell::Checker* spell)
-	: Glib::ObjectBase("HOCRDocument")
-	, m_spell(spell) {
+HOCRDocument::HOCRDocument()
+	: Glib::ObjectBase("HOCRDocument") {
+	m_spell = new HOCRSpellChecker();
 }
 
 HOCRDocument::~HOCRDocument() {
 	for(HOCRPage* page : m_pages) {
 		delete page;
 	}
+	delete m_spell;
 }
 
-void HOCRDocument::recheckSpelling() {
-	for(HOCRPage* page : m_pages) {
-		recursiveDataChanged(get_iter(get_root_path(page->index())), {"ocrx_word"});
+void HOCRDocument::resetMisspelled(const Gtk::TreeIter& index) {
+	if(!index->children().empty()) {
+		for(const Gtk::TreeIter& child : index->children()) {
+			resetMisspelled(child);
+		}
+	} else {
+		mutableItemAtIndex(index)->setMisspelled(-1);
+		row_changed(get_path(index), index);
+	}
+}
+
+void HOCRDocument::addSpellingActions(Gtk::Menu* menu, const Gtk::TreeIter& index) {
+	std::vector<Glib::ustring> suggestions;
+	Glib::ustring trimmedWord;
+	bool valid = getItemSpellingSuggestions(index, trimmedWord, suggestions, 16);
+	for(const Glib::ustring& suggestion : suggestions) {
+		Glib::ustring replacement = suggestion;
+		Gtk::MenuItem* item = Gtk::manage(new Gtk::MenuItem(suggestion));
+		CONNECT(item, activate, [this, replacement, index] { editItemText(index, replacement); });
+		menu->append(*item);
+	}
+
+	if(!trimmedWord.empty()) {
+		if(suggestions.empty()) {
+			Gtk::MenuItem* item = Gtk::manage(new Gtk::MenuItem(_("No suggestions")));
+			item->set_sensitive(false);
+			menu->append(*item);
+		}
+		if(!valid) {
+			menu->append(*Gtk::manage(new Gtk::SeparatorMenuItem));
+			Gtk::MenuItem* additem = Gtk::manage(new Gtk::MenuItem(_("Add to dictionary")));
+			CONNECT(additem, activate, [this, index, trimmedWord] {
+				m_spell->add_to_dictionary(trimmedWord);
+				resetMisspelled(index);
+			});
+			menu->append(*additem);
+			Gtk::MenuItem* ignoreitem = Gtk::manage(new Gtk::MenuItem(_("Ignore word")));
+			CONNECT(ignoreitem, activate, [this, index, trimmedWord] {
+				m_spell->ignore_word(trimmedWord);
+				resetMisspelled(index);
+			});
+			menu->append(*ignoreitem);
+		}
+	}
+}
+
+void HOCRDocument::addWordToDictionary(const Gtk::TreeIter& index) {
+	const HOCRItem* item = itemAtIndex(index);
+	if(item && item->isMisspelled()) {
+		Glib::ustring trimmedWord = HOCRItem::trimmedWord(item->text());
+		m_spell->add_to_dictionary(trimmedWord);
+		resetMisspelled(index);
 	}
 }
 
@@ -54,9 +105,15 @@ Glib::ustring HOCRDocument::toHTML() {
 	return html;
 }
 
-Gtk::TreeIter HOCRDocument::addPage(const xmlpp::Element* pageElement, bool cleanGraphics) {
-	m_pages.push_back(new HOCRPage(pageElement, ++m_pageIdCounter, m_defaultLanguage, cleanGraphics, m_pages.size()));
-	Gtk::TreeIter iter = get_iter(get_root_path(m_pages.back()->index()));
+Gtk::TreeIter HOCRDocument::insertPage(int beforeIdx, const xmlpp::Element* pageElement, bool cleanGraphics, const std::string& sourceBasePath) {
+	m_pages.insert(m_pages.begin() + beforeIdx, new HOCRPage(pageElement, ++m_pageIdCounter, m_defaultLanguage, cleanGraphics, beforeIdx));
+	if(!sourceBasePath.empty()) {
+		m_pages[beforeIdx]->convertSourcePath(sourceBasePath, true);
+	}
+	for(std::size_t i = beforeIdx + 1; i < m_pages.size(); ++i) {
+		m_pages[i]->m_index = i;
+	}
+	Gtk::TreeIter iter = get_iter(get_root_path(beforeIdx));
 	recursiveRowInserted(iter);
 	for(int i = 0, n = m_pages.size(); i < n; ++i) {
 		Gtk::TreePath path = get_root_path(i);
@@ -65,6 +122,16 @@ Gtk::TreeIter HOCRDocument::addPage(const xmlpp::Element* pageElement, bool clea
 	return iter;
 }
 
+Gtk::TreeIter HOCRDocument::indexAtItem(const HOCRItem* item) const {
+	Gtk::TreePath path;
+	const HOCRItem* cur = item;
+	while(cur) {
+		path.push_front(cur->index());
+		cur = cur->parent();
+	}
+	return get_iter(path);
+}
+
 bool HOCRDocument::editItemAttribute(const Gtk::TreeIter& index, const Glib::ustring& name, const Glib::ustring& value, const Glib::ustring& attrItemClass) {
 	HOCRItem* item = mutableItemAtIndex(index);
 	if(!item) {
@@ -74,7 +141,7 @@ bool HOCRDocument::editItemAttribute(const Gtk::TreeIter& index, const Glib::ust
 	item->setAttribute(name, value, attrItemClass);
 	row_changed(get_path(index), index);
 	if(name == "lang") {
-		recursiveDataChanged(index, {"ocrx_word"});
+		resetMisspelled(index);
 	}
 	m_signal_item_attribute_changed.emit(index, name, value);
 	if(name == "title:bbox") {
@@ -153,7 +220,9 @@ Gtk::TreeIter HOCRDocument::mergeItems(const Gtk::TreeIter& parent, int startRow
 			row_deleted(childPath);
 		}
 		targetItem->setText(text);
-		row_changed(get_path(targetIndex), targetIndex);
+		for(const Gtk::TreeIter& changedIndex : recheckItemSpelling(targetIndex)) {
+			row_changed(get_path(changedIndex), changedIndex);
+		}
 	} else {
 		// Merge other items: merge dom trees and bounding boxes
 		std::vector<HOCRItem*> moveChilds;
@@ -217,6 +286,69 @@ Gtk::TreeIter HOCRDocument::splitItem(const Gtk::TreeIter& index, int startRow,
 	return newIndex;
 }
 
+Gtk::TreeIter HOCRDocument::splitItemText(const Gtk::TreeIter& itemIndex, int pos, const Glib::RefPtr<Pango::Context>& pangoContext) {
+	HOCRItem* item = mutableItemAtIndex(itemIndex);
+	if(!item || item->itemClass() != "ocrx_word") {
+		return Gtk::TreeIter();
+	}
+	// Don't split if it would create empty words
+	if(pos == 0 || pos == int(item->text().length())) {
+		return itemIndex;
+	}
+	// Compute new bounding box using font metrics with default font
+	Glib::RefPtr<Pango::Layout> layout = Pango::Layout::create(pangoContext);
+	layout->set_text(item->text().substr(0, pos));
+	int substrWidth = layout->get_width() * Pango::SCALE;
+	layout->set_text(item->text());
+	int fullWidth = layout->get_width() * Pango::SCALE;
+	double fraction = substrWidth / double(fullWidth);
+	Geometry::Rectangle bbox = item->bbox();
+	Geometry::Rectangle leftBBox = bbox;
+	leftBBox.width = bbox.width * fraction;
+	Geometry::Rectangle rightBBox = bbox;
+	rightBBox.setLeft(bbox.x + bbox.width * fraction);
+
+	xmlpp::Document doc;
+	xmlpp::Element* newElement = doc.create_root_node("span");
+	newElement->add_child_text(item->text().substr(pos));
+	newElement->set_attribute("class", "ocrx_word");
+	std::map<Glib::ustring, Glib::ustring> attrs = item->getTitleAttributes();
+	attrs["bbox"] = Glib::ustring::compose("%1 %2 %3 %4", rightBBox.left(), rightBBox.top(), rightBBox.right(), rightBBox.bottom());
+	newElement->set_attribute("title", HOCRItem::serializeAttrGroup(attrs));
+	newElement->set_attribute("lang", item->lang());
+	HOCRItem* newItem = new HOCRItem(newElement, item->page(), item->parent(), item->index() + 1);
+	Gtk::TreePath insertPath = get_path(itemIndex);
+	insertPath.back() += 1;
+	insertItem(item->parent(), newItem, item->index() + 1);
+	row_inserted(insertPath, get_iter(insertPath));
+	item->setText(item->text().substr(0, pos));
+	editItemAttribute(itemIndex, "title:bbox", Glib::ustring::compose("%1 %2 %3 %4", leftBBox.left(), leftBBox.top(), leftBBox.right(), leftBBox.bottom()));
+	row_changed(get_path(itemIndex), itemIndex);
+	return get_iter(insertPath);
+}
+
+Gtk::TreeIter HOCRDocument::mergeItemText(const Gtk::TreeIter& itemIndex, bool mergeNext) {
+	HOCRItem* item = mutableItemAtIndex(itemIndex);
+	if(!item || item->itemClass() != "ocrx_word") {
+		return Gtk::TreeIter();
+	}
+	if((!mergeNext && item->index() == 0) || (mergeNext && item->index() == int(item->parent()->children().size()) - 1)) {
+		return Gtk::TreeIter();
+	}
+	int offset = mergeNext ? 1 : -1;
+	HOCRItem* otherItem = item->parent()->children()[item->index() + offset];
+	Glib::ustring newText = mergeNext ? item->text() + otherItem->text() : otherItem->text() + item->text();
+	Geometry::Rectangle newBbox = item->bbox().unite(otherItem->bbox());
+	item->setText(newText);
+	row_changed(get_path(itemIndex), itemIndex);
+	editItemAttribute(itemIndex, "title:bbox", Glib::ustring::compose("%1 %2 %3 %4", newBbox.left(), newBbox.top(), newBbox.right(), newBbox.bottom()));
+	Gtk::TreePath deletePath = get_path(itemIndex);
+	deletePath.back() += offset;
+	deleteItem(otherItem);
+	row_deleted(deletePath);
+	return itemIndex;
+}
+
 Gtk::TreeIter HOCRDocument::addItem(const Gtk::TreeIter& parent, const xmlpp::Element* element) {
 	HOCRItem* parentItem = mutableItemAtIndex(parent);
 	if(!parentItem) {
@@ -272,7 +404,11 @@ Gtk::TreeIter HOCRDocument::prevIndex(const Gtk::TreeIter& current) const {
 	Gtk::TreeIter iter = current;
 	// If the current index is invalid return last index
 	if(!iter) {
-		return get_iter(get_root_path(m_pages.size() - 1));
+		if(m_pages.empty()) {
+			return Gtk::TreeIter();
+		} else {
+			return get_iter(get_root_path(m_pages.size() - 1));
+		}
 	}
 	// Return last possible leaf of previous sibling, if any, or parent
 	Gtk::TreeIter parent = iter->parent();
@@ -289,18 +425,106 @@ Gtk::TreeIter HOCRDocument::prevIndex(const Gtk::TreeIter& current) const {
 	return iter;
 }
 
-bool HOCRDocument::indexIsMisspelledWord(const Gtk::TreeIter& index) const {
-	return !checkItemSpelling(index);
+Gtk::TreeIter HOCRDocument::prevOrNextIndex(bool next, const Gtk::TreeIter& current, const Glib::ustring& ocrClass, bool misspelled, bool lowconf) const {
+	Gtk::TreeIter start = current;
+	if(!start) {
+		start = get_iter(get_root_path(0));
+	}
+	Gtk::TreeIter curr = next ? nextIndex(start) : prevIndex(start);
+	while(curr && curr != start) {
+		const HOCRItem* item = itemAtIndex(curr);
+		if(item && item->itemClass() == ocrClass && (!misspelled || indexIsMisspelledWord(curr)) && (!lowconf || std::atoi(item->getTitleAttribute("x_wconf").c_str()) < 90)) {
+			break;
+		}
+		curr = next ? nextIndex(curr) : prevIndex(curr);
+	};
+	return curr;
 }
 
-bool HOCRDocument::checkItemSpelling(const Gtk::TreeIter& index, std::vector<Glib::ustring>* suggestions, int limit) const {
+bool HOCRDocument::indexIsMisspelledWord(const Gtk::TreeIter& index) const {
+	const HOCRItem* item = itemAtIndex(index);
+	if(item->isMisspelled() == -1) {
+		recheckItemSpelling(index);
+	}
+	return item->isMisspelled() == 1;
+}
+
+std::vector<Gtk::TreeIter> HOCRDocument::recheckItemSpelling(const Gtk::TreeIter& index) const {
+	HOCRItem* item = mutableItemAtIndex(index);
+	if(item->itemClass() != "ocrx_word") { return {}; }
+
+	Glib::ustring prefix, suffix, trimmed = HOCRItem::trimmedWord(item->text(), &prefix, &suffix);
+	if(trimmed.empty()) {
+		item->setMisspelled(false);
+		return {index};
+	}
+	Glib::ustring lang = item->spellingLang();
+	if(lang.empty()) {
+		item->setMisspelled(false);
+		return {index};
+	}
+	if(m_spell->get_language() != lang) {
+		try {
+			m_spell->set_language(lang);
+		} catch(const GtkSpell::Error& /*e*/) {
+			item->setMisspelled(false);
+			return {index};
+		}
+	}
+
+	// check word, including (if requested) setting suggestions; handle hyphenated phrases correctly
+	if(m_spell->checkSpelling(trimmed)) {
+		item->setMisspelled(false);
+		return {index};
+	}
+	item->setMisspelled(true);
+
+	// handle some hyphenated words
+	// don't bother with words not broken over sibling text lines (ie interrupted by other blocks), it's human hard
+	// don't bother with words broken over three or more lines, it's implausible and this treatment is ^ necessarily incomplete
+	HOCRItem* line = item->parent();
+	if(!line) { return {index}; }
+	HOCRItem* paragraph = line->parent();
+	if(!paragraph) { return {index}; }
+	int lineIdx = item->index();
+	std::vector<HOCRItem*> paragraphLines = paragraph->children();
+	if(item == line->children().front() && lineIdx > 0) {
+		HOCRItem* prevLine = paragraphLines.at(lineIdx - 1);
+		if(!prevLine || prevLine->children().empty()) { return {index}; }
+		HOCRItem* prevWord = prevLine->children().back();
+		if(!prevWord || prevWord->itemClass() != "ocrx_word") { return {index}; }
+		Glib::ustring prevText = prevWord->text();
+		if(Utils::string_endswith(prevText, "-")) { return {index}; }
+
+		// don't bother with (reassembled) suggestions for broken words since we can't re-break them
+		bool valid = m_spell->checkSpelling(HOCRItem::trimmedWord(prevText) + trimmed);
+		item->setMisspelled(!valid);
+		prevWord->setMisspelled(!valid);
+		return {index, indexAtItem(prevWord)};
+	} else if(item == line->children().back() && lineIdx + 1 < int(paragraphLines.size()) && Utils::string_endswith(item->text(), "-")) {
+		HOCRItem* nextLine = paragraphLines.at(lineIdx + 1);
+		if(!nextLine || nextLine->children().empty()) { return {index}; }
+		HOCRItem* nextWord = nextLine->children().front();
+		if(!nextWord || nextWord->itemClass() != "ocrx_word") { return {index}; }
+
+		bool valid = m_spell->checkSpelling(trimmed + HOCRItem::trimmedWord(nextWord->text()));
+		item->setMisspelled(!valid);
+		nextWord->setMisspelled(!valid);
+		return {index, indexAtItem(nextWord)};
+	}
+	return {index};
+}
+
+bool HOCRDocument::getItemSpellingSuggestions(const Gtk::TreeIter& index, Glib::ustring& trimmedWord, std::vector<Glib::ustring>& suggestions, int limit) const {
 	const HOCRItem* item = itemAtIndex(index);
 	if(item->itemClass() != "ocrx_word") { return true; }
 
-	Glib::ustring prefix, suffix, trimmed = HOCRItem::trimmedWord(item->text(), &prefix, &suffix);
-	if(trimmed.empty()) { return true; }
+	Glib::ustring prefix, suffix;
+	trimmedWord = HOCRItem::trimmedWord(item->text(), &prefix, &suffix);
+	if(trimmedWord.empty()) {
+		return true;
+	}
 	Glib::ustring lang = item->spellingLang();
-	bool haveLanguage = true;
 	if(m_spell->get_language() != lang) {
 		try {
 			m_spell->set_language(lang);
@@ -309,44 +533,11 @@ bool HOCRDocument::checkItemSpelling(const Gtk::TreeIter& index, std::vector<Gli
 		}
 	}
 
-	// check word, including (if requested) setting suggestions; handle hyphenated phrases correctly
-	if(checkSpelling(trimmed, suggestions, limit)) { return true; }
-
-	// handle some hyphenated words
-	// don't bother with words not broken over sibling text lines (ie interrupted by other blocks), it's human hard
-	// don't bother with words broken over three or more lines, it's implausible and this treatment is ^ necessarily incomplete
-	HOCRItem* parent = item->parent();
-	if(!parent) { return false; }
-	HOCRItem* grandparent = parent->parent();
-	if(!grandparent) { return false; }
-	int idx = item->index();
-	int parentIdx = parent->index();
-	std::vector<HOCRItem*> parentSiblings = grandparent->children();
-	if(idx == 0 && parentIdx > 0) {
-		HOCRItem* parentPrevSibling = parentSiblings.at(parentIdx - 1);
-		if(!parentPrevSibling) { return false; }
-		std::vector<HOCRItem*> cousins = parentPrevSibling->children();
-		if(cousins.size() < 1) { return false; }
-		HOCRItem* prevCousin = cousins.back();
-		if(!prevCousin || prevCousin->itemClass() != "ocrx_word") { return false; }
-		Glib::ustring prevText = prevCousin->text();
-		if(prevText.empty() || prevText[prevText.length() - 1] != '-') { return false; }
-
-		// don't bother with (reassembled) suggestions for broken words since we can't re-break them
-		return checkSpelling(HOCRItem::trimmedWord(prevText) + trimmed);
+	bool valid = m_spell->checkSpelling(trimmedWord, &suggestions, limit);
+	for(int i = 0, n = suggestions.size(); i < n; ++i) {
+		suggestions[i] = prefix + suggestions[i] + suffix;
 	}
-	Glib::ustring text = item->text();
-	if(idx + 1 == parent->children().size() && parentIdx + 1 < parentSiblings.size() && !text.empty() && text[text.length() - 1] == '-') {
-		HOCRItem* parentNextSibling = parentSiblings.at(parentIdx + 1);
-		if(!parentNextSibling) { return false; }
-		std::vector<HOCRItem*> cousins = parentNextSibling->children();
-		if(cousins.size() < 1) { return false; }
-		HOCRItem* nextCousin = cousins.front();
-		if(!nextCousin || nextCousin->itemClass() != "ocrx_word") { return false; }
-
-		return checkSpelling(trimmed + HOCRItem::trimmedWord(nextCousin->text()));
-	}
-	return false;
+	return valid;
 }
 
 bool HOCRDocument::referencesSource(const Glib::ustring& filename) const {
@@ -389,18 +580,6 @@ void HOCRDocument::convertSourcePaths(const std::string& basepath, bool absolute
 	}
 }
 
-void HOCRDocument::recursiveDataChanged(const Gtk::TreeIter& index, const std::vector<Glib::ustring>& itemClasses) {
-	if(index && !index->children().empty()) {
-		bool emitChanged = (itemClasses.empty() || std::find(itemClasses.begin(), itemClasses.end(), itemAtIndex(index)->itemClass()) != itemClasses.end());
-		for(const Gtk::TreeIter& childIndex : index->children()) {
-			if(emitChanged) {
-				row_changed(get_path(childIndex), childIndex);
-			}
-			recursiveDataChanged(childIndex);
-		}
-	}
-}
-
 void HOCRDocument::recursiveRowInserted(const Gtk::TreeIter& index) {
 	DEBUG(std::cout << "Inserted: " << get_path(index).to_string() << std::endl;)
 	row_inserted(get_path(index), index);
@@ -435,6 +614,8 @@ GType HOCRDocument::get_column_type_vfunc(int index) const {
 		return Glib::Value<Glib::RefPtr<Gdk::Pixbuf>>::value_type();
 	} else if(index == COLUMN_TEXT) {
 		return Glib::Value<Glib::ustring>::value_type();
+	} else if(index == COLUMN_TOOLTIP) {
+		return Glib::Value<Glib::ustring>::value_type();
 	} else if(index == COLUMN_TEXT_COLOR) {
 		return Glib::Value<Glib::ustring>::value_type();
 	} else if(index == COLUMN_WCONF) {
@@ -453,9 +634,9 @@ bool HOCRDocument::iter_next_vfunc(const iterator& iter, iterator& iter_next) co
 	HOCRItem* nextItem = nullptr;
 	if(item) {
 		if(item->parent()) {
-			nextItem = item->index() < item->parent()->children().size() - 1 ? item->parent()->children()[item->index() + 1] : nullptr;
+			nextItem = item->index() < int(item->parent()->children().size()) - 1 ? item->parent()->children()[item->index() + 1] : nullptr;
 		} else {
-			nextItem = item->index() < m_pages.size() - 1 ? m_pages[item->index() + 1] : nullptr;
+			nextItem = item->index() < int(m_pages.size()) - 1 ? m_pages[item->index() + 1] : nullptr;
 		}
 	}
 	iter_next.gobj()->user_data = nextItem;
@@ -473,7 +654,7 @@ bool HOCRDocument::get_iter_vfunc(const Path& path, iterator& iter) const {
 	int idx = 0, size = path.size();
 	HOCRItem* item = m_pages[path[idx++]];
 	while(item && idx < size) {
-		item = path[idx] >= item->children().size() ? nullptr : item->children()[path[idx++]];
+		item = path[idx] >= int(item->children().size()) ? nullptr : item->children()[path[idx++]];
 	}
 	iter.gobj()->user_data = item;
 	iter.set_stamp(iter.gobj()->user_data != nullptr);
@@ -502,7 +683,7 @@ bool HOCRDocument::iter_parent_vfunc(const iterator& child, iterator& iter) cons
 bool HOCRDocument::iter_nth_child_vfunc(const iterator& parent, int n, iterator& iter) const {
 	DEBUG(std::cout << "iter_nth_child_vfunc " << get_path(parent).to_string() << "@" << n << ": ";)
 	HOCRItem* parentItem = mutableItemAtIndex(parent);
-	iter.gobj()->user_data = parentItem && n < parentItem->children().size() ? parentItem->children()[n] : nullptr;
+	iter.gobj()->user_data = parentItem && n < int(parentItem->children().size()) ? parentItem->children()[n] : nullptr;
 	iter.set_stamp(iter.gobj()->user_data != nullptr);
 	DEBUG(std::cout << (iter.gobj()->user_data != nullptr) << std::endl;)
 	return iter.gobj()->user_data != nullptr;
@@ -510,7 +691,7 @@ bool HOCRDocument::iter_nth_child_vfunc(const iterator& parent, int n, iterator&
 
 bool HOCRDocument::iter_nth_root_child_vfunc(int n, iterator& iter) const {
 	DEBUG(std::cout << "iter_nth_root_child_vfunc " << n << ": ";)
-	iter.gobj()->user_data =  n < m_pages.size() ? m_pages[n] : nullptr;
+	iter.gobj()->user_data =  n < int(m_pages.size()) ? m_pages[n] : nullptr;
 	iter.set_stamp(iter.gobj()->user_data != nullptr);
 	DEBUG(std::cout << (iter.gobj()->user_data != nullptr) << std::endl;)
 	return iter.gobj()->user_data != nullptr;
@@ -566,6 +747,8 @@ void HOCRDocument::get_value_vfunc(const iterator& iter, int column, Glib::Value
 		setValue(value, decorationRoleForItem(item));
 	} else if(column == COLUMN_TEXT) {
 		setValue(value, displayRoleForItem(item));
+	} else if(column == COLUMN_TOOLTIP) {
+		setValue(value, tooltipRoleForItem(item));
 	} else if(column == COLUMN_TEXT_COLOR) {
 		bool enabled = item->isEnabled();
 		const HOCRItem* parent = item->parent();
@@ -575,9 +758,9 @@ void HOCRDocument::get_value_vfunc(const iterator& iter, int column, Glib::Value
 		}
 		Glib::ustring color;
 		if(enabled) {
-			color = checkItemSpelling(iter) ? "#000" : "#F00";
+			color = indexIsMisspelledWord(iter) ? "#F00" : "#000";
 		} else {
-			color = checkItemSpelling(iter) ? "#a0a0a4" : "#d05052";
+			color = indexIsMisspelledWord(iter) ? "#d05052" : "#a0a0a4";
 		}
 		setValue(value, color);
 	} else if(column == COLUMN_WCONF) {
@@ -591,12 +774,18 @@ void HOCRDocument::set_value_impl(const iterator& row, int column, const Glib::V
 		return;
 	} else if(column == COLUMN_TEXT) {
 		item->setText(static_cast<const Glib::Value<Glib::ustring>&>(value).get());
-		row_changed(get_path(row), row);
+		for(const Gtk::TreeIter& changedIndex : recheckItemSpelling(row)) {
+			row_changed(get_path(changedIndex), changedIndex);
+		}
 	} else if(column == COLUMN_CHECKED) {
 		item->setEnabled(static_cast<const Glib::Value<bool>&>(value).get());
-		Gtk::TreePath path = get_path(row);
-		row_changed(path, row);
-		recursiveDataChanged(row);
+		// Get leaf
+		Gtk::TreeIter leaf = row;
+		while(!leaf->children().empty()) {
+			leaf = leaf->children()[leaf->children().size() - 1];
+		}
+		Gtk::TreePath path = get_path(leaf);
+		row_changed(path, leaf);
 	}
 }
 
@@ -620,7 +809,7 @@ Glib::ustring HOCRDocument::displayRoleForItem(const HOCRItem* item) const {
 	} else if(itemClass == "ocr_graphic") {
 		return _("Graphic");
 	}
-	return "";
+	return Glib::ustring();
 }
 
 Glib::RefPtr<Gdk::Pixbuf> HOCRDocument::decorationRoleForItem(const HOCRItem* item) const {
@@ -638,67 +827,18 @@ Glib::RefPtr<Gdk::Pixbuf> HOCRDocument::decorationRoleForItem(const HOCRItem* it
 	} else if(itemClass == "ocr_graphic") {
 		return Gdk::Pixbuf::create_from_resource("/org/gnome/gimagereader/item_halftone.png");
 	}
-	return Gdk::Pixbuf::create(Gdk::COLORSPACE_RGB, false, 8, 0, 0);
+	auto pixbuf = Gdk::Pixbuf::create(Gdk::COLORSPACE_RGB, false, 8, 16, 16);
+	pixbuf->fill(255);
+	return pixbuf;
 }
 
-bool HOCRDocument::checkSpelling(const Glib::ustring& trimmed, std::vector<Glib::ustring>* suggestions, int limit) const {
-	static const Glib::RefPtr<Glib::Regex> splitRe = Glib::Regex::create("[\u2013\u2014]+");
-	std::vector<std::pair<Glib::ustring, int>> words = Utils::string_split_pos(trimmed, splitRe);
-
-	std::size_t perWordLimit = 0;
-	// s = p^w => w = log_p(c) = log(c)/log(p) => p = 10^(log(c)/w)
-	if(limit > 0) { perWordLimit = std::size_t(std::pow(10, std::log10(limit) / words.size())); }
-	std::vector<std::vector<Glib::ustring>> wordSuggestions;
-	bool valid = true;
-	bool multipleWords = words.size() > 1;
-	for(const std::pair<Glib::ustring, int>& word : words) {
-		bool wordValid = m_spell->check_word(word.first);
-		valid &= wordValid;
-		if(suggestions) {
-			std::vector<Glib::ustring> ws = m_spell->get_suggestions(word.first);
-			if(wordValid && multipleWords) { ws.insert(ws.begin(), word.first); }
-			if(limit == -1) {
-				wordSuggestions.push_back(ws);
-			} else if(limit > 0) {
-				ws.resize(std::min(ws.size(), perWordLimit));
-				wordSuggestions.push_back(ws);
-			}
-		}
-	}
-	if(suggestions) {
-		suggestions->clear();
-		std::vector<std::vector<Glib::ustring>> suggestionCombinations;
-		generateCombinations(wordSuggestions, suggestionCombinations, 0, std::vector<Glib::ustring>());
-		for(const std::vector<Glib::ustring>& combination : suggestionCombinations) {
-			Glib::ustring s = "";
-			auto originalWord = words.begin();
-			int last = 0;
-			int next;
-			for(const Glib::ustring& suggestedWord : combination) {
-				next = originalWord->second;
-				s.append(trimmed.substr(last, next - last));
-				s.append(suggestedWord);
-				last = next + originalWord->first.length();
-				originalWord++;
-			}
-			s.append(trimmed.substr(last));
-			suggestions->push_back(s);
-		}
-	}
-	return valid;
-}
-
-// each suggestion for each word => each word in each suggestion
-void HOCRDocument::generateCombinations(const std::vector<std::vector<Glib::ustring>>& lists, std::vector<std::vector<Glib::ustring>>& results, int depth, const std::vector<Glib::ustring>& c) const {
-	if(depth == lists.size()) {
-		results.push_back(c);
-		return;
-	}
-	for(int i = 0; i < lists[depth].size(); ++i) {
-		std::vector<Glib::ustring> newc = c;
-		newc.insert(newc.end(), lists[depth][i]);
-		generateCombinations(lists, results, depth + 1, newc);
+Glib::ustring HOCRDocument::tooltipRoleForItem(const HOCRItem* item) const {
+	Glib::ustring itemClass = item->itemClass();
+	if(itemClass == "ocr_page") {
+		const HOCRPage* page = static_cast<const HOCRPage*>(item);
+		return Glib::ustring::compose("%1 (%2 %3/%4)", page->title(), _("Page"), page->index() + 1, m_pages.size());
 	}
+	return Glib::ustring();
 }
 
 void HOCRDocument::insertItem(HOCRItem* parent, HOCRItem* item, int i) {
@@ -760,7 +900,7 @@ Glib::ustring HOCRItem::serializeAttrGroup(const std::map<Glib::ustring, Glib::u
 
 Glib::ustring HOCRItem::trimmedWord(const Glib::ustring& word, Glib::ustring* prefix, Glib::ustring* suffix) {
 	// correctly trim words with apostrophes or hyphens within them, initialisms/acronyms, and numeric citations
-	static const Glib::RefPtr<Glib::Regex> wordRe = Glib::Regex::create("^(\\W*)(\\w?|\\w(\\w|[-\u2013\u2014'’])*\\w|(\\w+\\.){2,})([\\W\u00b2\u00b3\u00b9\u2070-\u207e]*)$");
+	static const Glib::RefPtr<Glib::Regex> wordRe = Glib::Regex::create("^(\\W*)(\\w?|\\w(\\w|[-\u2013\u2014'’])*\\w|(\\w+\\.){2,})([\\W\u00b2\u00b3\u00b9\u2070\\-\u207e]*)$");
 	Glib::MatchInfo matchInfo;
 	if(wordRe->match(word, matchInfo)) {
 		if(prefix) {
@@ -786,6 +926,10 @@ HOCRItem::HOCRItem(const xmlpp::Element* element, HOCRPage* page, HOCRItem* pare
 			m_attrs[attrName] = attr->get_value();
 		}
 	}
+	// Map ocr_header/ocr_caption/ocr_textfloat to ocr_line
+	if(m_attrs["class"] == "ocr_header" || m_attrs["class"] == "ocr_caption" || m_attrs["class"] == "ocr_textfloat") {
+		m_attrs["class"] = "ocr_line";
+	}
 	// Adjust item id based on pageId
 	if(parent) {
 		Glib::ustring idClass = itemClass().substr(itemClass().find_first_of('_') + 1);
@@ -808,6 +952,9 @@ HOCRItem::HOCRItem(const xmlpp::Element* element, HOCRPage* page, HOCRItem* pare
 	} else if(itemClass() == "ocr_line") {
 		// Depending on the locale, tesseract can use a comma instead of a dot as decimal separator in the baseline...
 		Utils::string_replace(m_titleAttrs["baseline"], ",", ".", true);
+		m_misspelled = false;
+	} else {
+		m_misspelled = false;
 	}
 }
 
@@ -989,13 +1136,13 @@ std::pair<double, double> HOCRItem::baseLine() const {
 	return std::make_pair(0.0, 0.0);
 }
 
-bool HOCRItem::parseChildren(const xmlpp::Element* element, Glib::ustring language) {
+bool HOCRItem::parseChildren(const xmlpp::Element* element, Glib::ustring language, const Glib::ustring& defaultLanguage) {
 	// Determine item language (inherit from parent if not specified)
 	Glib::ustring elemLang = getAttribute("lang");
 	if(!elemLang.empty()) {
 		auto it = s_langCache.find(elemLang);
 		if(it == s_langCache.end()) {
-			it = s_langCache.insert(std::make_pair(elemLang, Utils::getSpellingLanguage(elemLang))).first;
+			it = s_langCache.insert(std::make_pair(elemLang, Utils::getSpellingLanguage(elemLang, defaultLanguage))).first;
 		}
 		m_attrs.erase(m_attrs.find("lang"));
 		language = it->second;
@@ -1009,7 +1156,7 @@ bool HOCRItem::parseChildren(const xmlpp::Element* element, Glib::ustring langua
 	const xmlpp::Element* childElement = XmlUtils::firstChildElement(element);
 	while(childElement) {
 		m_childItems.push_back(new HOCRItem(childElement, m_pageItem, this, m_childItems.size()));
-		haveWords |= m_childItems.back()->parseChildren(childElement, language);
+		haveWords |= m_childItems.back()->parseChildren(childElement, language, defaultLanguage);
 		childElement = XmlUtils::nextSiblingElement(childElement);
 	}
 	return haveWords;
@@ -1027,7 +1174,7 @@ Glib::ustring HOCRItem::getTitleAttribute(const Glib::ustring& key) const {
 
 ///////////////////////////////////////////////////////////////////////////////
 
-HOCRPage::HOCRPage(const xmlpp::Element* element, int pageId, const Glib::ustring& language, bool cleanGraphics, int index)
+HOCRPage::HOCRPage(const xmlpp::Element* element, int pageId, const Glib::ustring& defaultLanguage, bool cleanGraphics, int index)
 	: HOCRItem(element, this, nullptr, index), m_pageId(pageId) {
 	m_attrs["id"] = Glib::ustring::compose("page_%1", pageId);
 
@@ -1041,6 +1188,9 @@ HOCRPage::HOCRPage(const xmlpp::Element* element, int pageId, const Glib::ustrin
 	}
 	m_angle = std::atof(m_titleAttrs["rot"].c_str());
 	m_resolution = std::atoi(m_titleAttrs["res"].c_str());
+	if(m_resolution == 0) {
+		m_resolution = 100;
+	}
 
 	const xmlpp::Element* childElement = XmlUtils::firstChildElement(element, "div");
 	while(childElement) {
@@ -1048,7 +1198,7 @@ HOCRPage::HOCRPage(const xmlpp::Element* element, int pageId, const Glib::ustrin
 		const xmlpp::Element* nextSibling = XmlUtils::nextSiblingElement(childElement);
 		HOCRItem* item = new HOCRItem(childElement, this, this, m_childItems.size());
 		m_childItems.push_back(item);
-		if(!item->parseChildren(childElement, language)) {
+		if(!item->parseChildren(childElement, defaultLanguage, defaultLanguage)) {
 			// No word children -> treat as graphic
 			if(cleanGraphics && (item->bbox().width < 10 || item->bbox().height < 10)) {
 				// Ignore graphics which are less than 10 x 10
