@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * OutputEditorHOCR.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -25,6 +25,7 @@
 #include <tesseract/ocrclass.h>
 #undef USE_STD_NAMESPACE
 #include <libxml++/libxml++.h>
+#include <gtkspellmm.h>
 
 #include "ConfigSettings.hh"
 #include "DisplayerToolHOCR.hh"
@@ -115,6 +116,22 @@ protected:
 			});
 			combo->show();
 			return Gtk::manage(combo);
+		} else if(it && (*it)[m_cols.attr] == "lang") {
+			Gtk::ComboBoxText* combo = new Gtk::ComboBoxText();
+			for(const Glib::ustring& code : GtkSpell::Checker::get_language_list()) {
+				Glib::ustring text = GtkSpell::Checker::decode_language_code(code);
+				combo->append(code, text);
+			}
+			if((*it)[m_cols.multiple] == true) {
+				combo->set_active(-1);
+			} else {
+				combo->set_active_id((*it)[m_cols.value]);
+			}
+			combo->signal_editing_done().connect([ = ] {
+				edited(path, combo->get_active_id());
+			});
+			combo->show();
+			return Gtk::manage(combo);
 		} else {
 			Gtk::Entry* entry = new Gtk::Entry;
 			entry->set_text(property_text());
@@ -135,11 +152,17 @@ class OutputEditorHOCR::HOCRTreeView : public Gtk::TreeView {
 public:
 	HOCRTreeView(BaseObjectType* cobject, const Glib::RefPtr<Gtk::Builder>& /*builder*/)
 		: Gtk::TreeView(cobject) {
-		CONNECT(get_selection(), changed, [this] {
+		m_connectionSelectionChanged = CONNECT(get_selection(), changed, [this] {
 			Gtk::TreeIter oldIndex = m_currentIndex;
 			m_currentIndex = currentIndex();
 			m_signal_current_index_changed(m_currentIndex, oldIndex);
 		});
+		CONNECT(this, button_press_event, [this](GdkEventButton * ev) {
+			return button_press_event(ev);
+		}, false);
+		CONNECT(this, key_press_event, [this](GdkEventKey * ev) {
+			return key_press_event(ev);
+		}, false);
 	}
 	sigc::signal<void, Gtk::TreeIter, Gtk::TreeIter> signal_current_index_changed() {
 		return m_signal_current_index_changed;
@@ -158,6 +181,20 @@ public:
 		}
 		return Gtk::TreeIter();
 	}
+	void setCurrentIndex(const Gtk::TreeIter& index) {
+		if (index && m_currentIndex && index == m_currentIndex) {
+			return;
+		}
+		m_connectionSelectionChanged.block();
+		get_selection()->unselect_all();
+		m_connectionSelectionChanged.unblock();
+		if(index) {
+			expand_to_path(get_model()->get_path(index));
+			get_selection()->select(index);
+			set_cursor(get_model()->get_path(index));
+			scroll_to_row(get_model()->get_path(index));
+		}
+	}
 	Gtk::TreeIter indexAtPos(int evx, int evy) {
 		Gtk::TreePath path;
 		Gtk::TreeViewColumn* col;
@@ -170,7 +207,7 @@ public:
 	}
 
 protected:
-	bool on_button_press_event(GdkEventButton* button_event) {
+	bool button_press_event(GdkEventButton* button_event) {
 		bool rightclick = (button_event->type == GDK_BUTTON_PRESS && button_event->button == 3);
 
 		Gtk::TreePath path;
@@ -191,7 +228,7 @@ protected:
 		}
 		return true;
 	}
-	bool on_key_press_event(GdkEventKey* key_event) {
+	bool key_press_event(GdkEventKey* key_event) {
 		if(key_event->keyval == GDK_KEY_Delete) {
 			m_signal_delete.emit();
 			return true;
@@ -205,10 +242,50 @@ private:
 	sigc::signal<void, Gtk::TreeIter, Gtk::TreeIter> m_signal_current_index_changed;
 	sigc::signal<void, GdkEventButton*> m_signal_context_menu;
 	sigc::signal<void> m_signal_delete;
+	sigc::connection m_connectionSelectionChanged;
 };
 
 ///////////////////////////////////////////////////////////////////////////////
 
+void OutputEditorHOCR::HOCRBatchProcessor::writeHeader(std::ostream& dev, tesseract::TessBaseAPI* tess, const PageInfo& pageInfo) const {
+	Glib::ustring header = Glib::ustring::compose(
+	                           "<!DOCTYPE html>\n"
+	                           "<html>\n"
+	                           "<head>\n"
+	                           " <title>%1</title>\n"
+	                           " <meta charset=\"utf-8\" /> \n"
+	                           " <meta name='ocr-system' content='tesseract %2' />\n"
+	                           " <meta name='ocr-capabilities' content='ocr_page ocr_carea ocr_par ocr_line ocrx_word'/>\n"
+	                           "</head><body>\n", Glib::path_get_basename(pageInfo.filename), tess->Version());
+	dev << header.raw();
+}
+
+void OutputEditorHOCR::HOCRBatchProcessor::writeFooter(std::ostream& dev) const {
+	dev << "</body></html>\n";
+}
+
+void OutputEditorHOCR::HOCRBatchProcessor::appendOutput(std::ostream& dev, tesseract::TessBaseAPI* tess, const PageInfo& pageInfos, bool /*firstArea*/) const {
+	char* text = tess->GetHOCRText(pageInfos.page);
+	xmlpp::DomParser parser;
+	parser.parse_memory(text);
+	xmlpp::Document* doc = parser.get_document();
+
+	xmlpp::Element* pageDiv = doc->get_root_node();
+	if(!pageDiv || pageDiv->get_name() != "div") {
+		return;
+	}
+	std::map<Glib::ustring, Glib::ustring> attrs = HOCRItem::deserializeAttrGroup(pageDiv->get_attribute_value("title"));
+	// This works because in batch mode the output is created next to the source image
+	attrs["image"] = Glib::ustring::compose("'%1'", Glib::path_get_basename(pageInfos.filename));
+	attrs["ppageno"] = Glib::ustring::compose("%1", pageInfos.page);
+	attrs["rot"] = Glib::ustring::compose("%1", pageInfos.angle);
+	attrs["res"] = Glib::ustring::compose("%1", pageInfos.resolution);
+	pageDiv->set_attribute("title", HOCRItem::serializeAttrGroup(attrs));
+	doc->write_to_stream(dev);
+}
+
+///////////////////////////////////////////////////////////////////////////////
+
 OutputEditorHOCR::OutputEditorHOCR(DisplayerToolHOCR* tool) {
 	ui.builder->get_widget_derived("treeviewItems", m_treeView);
 	ui.setupUi();
@@ -220,17 +297,20 @@ OutputEditorHOCR::OutputEditorHOCR(DisplayerToolHOCR* tool) {
 
 	m_tool = tool;
 
-	m_document = Glib::RefPtr<HOCRDocument>(new HOCRDocument(&m_spell));
+	m_document = Glib::RefPtr<HOCRDocument>(new HOCRDocument);
 
 	// HOCR tree view
 	m_treeView->set_model(m_document);
 	m_treeView->get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);
+	m_treeView->set_tooltip_column(HOCRDocument::COLUMN_TOOLTIP);
 	// First column: [checkbox][icon][text]
 	m_treeViewTextCell = Gtk::manage(new HOCRCellRendererText);
 	Gtk::TreeViewColumn* itemViewCol1 = Gtk::manage(new Gtk::TreeViewColumn(""));
 	Gtk::CellRendererToggle& checkRenderer = *Gtk::manage(new Gtk::CellRendererToggle);
 	Gtk::CellRendererPixbuf& iconRenderer = *Gtk::manage(new Gtk::CellRendererPixbuf);
+	iconRenderer.set_fixed_size(22, 22);
 	Gtk::CellRendererText& textRenderer = *m_treeViewTextCell;
+	textRenderer.property_ellipsize().set_value(Pango::ELLIPSIZE_MIDDLE);
 	itemViewCol1->pack_start(checkRenderer, false);
 	itemViewCol1->pack_start(iconRenderer, false);
 	itemViewCol1->pack_start(textRenderer, true);
@@ -239,6 +319,8 @@ OutputEditorHOCR::OutputEditorHOCR(DisplayerToolHOCR* tool) {
 	itemViewCol1->add_attribute(textRenderer, "text", HOCRDocument::COLUMN_TEXT);
 	itemViewCol1->add_attribute(textRenderer, "editable", HOCRDocument::COLUMN_EDITABLE);
 	itemViewCol1->add_attribute(textRenderer, "foreground", HOCRDocument::COLUMN_TEXT_COLOR);
+	itemViewCol1->set_sizing(Gtk::TREE_VIEW_COLUMN_FIXED);
+	itemViewCol1->set_expand(true);
 	CONNECT(&checkRenderer, toggled, [this](const Glib::ustring & path) {
 		Gtk::TreeIter it = m_document->get_iter(path);
 		bool prev;
@@ -258,6 +340,9 @@ OutputEditorHOCR::OutputEditorHOCR(DisplayerToolHOCR* tool) {
 	m_treeView->append_column(*itemViewCol2);
 	itemViewCol2->set_visible(false);
 	itemViewCol2->set_fixed_width(32);
+	itemViewCol2->set_sizing(Gtk::TREE_VIEW_COLUMN_FIXED);
+
+	m_treeView->set_fixed_height_mode();
 
 	// Navigation target combo box
 	Glib::RefPtr<Gtk::ListStore> navigationComboModel = Gtk::ListStore::create(m_navigationComboCols);
@@ -280,6 +365,9 @@ OutputEditorHOCR::OutputEditorHOCR(DisplayerToolHOCR* tool) {
 	row = *(navigationComboModel->append());
 	row[m_navigationComboCols.label] = _("Misspelled word");
 	row[m_navigationComboCols.itemClass] = "ocrx_word_bad";
+	row = *(navigationComboModel->append());
+	row[m_navigationComboCols.label] = _("Low confidence word");
+	row[m_navigationComboCols.itemClass] = "ocrx_word_lowconf";
 	ui.comboNavigation->pack_start(m_navigationComboCols.label);
 	ui.comboNavigation->set_active(0);
 
@@ -334,7 +422,11 @@ OutputEditorHOCR::OutputEditorHOCR(DisplayerToolHOCR* tool) {
 	ui.buttonSave->set_sensitive(false);
 	ui.buttonExport->set_sensitive(false);
 
-	CONNECT(ui.buttonOpen, clicked, [this] { open(); });
+	CONNECT(ui.menuItemInsertAppend, activate, [this] { setInsertMode(InsertMode::Append, "ins_hocr_append.png"); });
+	CONNECT(ui.menuItemInsertBefore, activate, [this] { setInsertMode(InsertMode::InsertBefore, "ins_hocr_before.png"); });
+	CONNECT(ui.buttonOpen, clicked, [this] { open(InsertMode::Replace); });
+	CONNECT(ui.menuItemOpenAppend, activate, [this] { open(InsertMode::Append); });
+	CONNECT(ui.menuItemOpenInsert, activate, [this] { open(InsertMode::InsertBefore); });
 	CONNECT(ui.buttonSave, clicked, [this] { save(); });
 	CONNECT(ui.menuitemExportText, activate, [this] { exportToText(); });
 	CONNECT(ui.menuitemExportPdf, activate, [this] { exportToPDF(); });
@@ -342,7 +434,7 @@ OutputEditorHOCR::OutputEditorHOCR(DisplayerToolHOCR* tool) {
 	CONNECT(ui.buttonClear, clicked, [this] { clear(); });
 	CONNECT(ui.buttonFindreplace, toggled, [this] { m_searchFrame->clear(); m_searchFrame->getWidget()->set_visible(ui.buttonFindreplace->get_active()); });
 	CONNECT(ui.buttonWconf, toggled, [this] { m_treeView->get_column(1)->set_visible(ui.buttonWconf->get_active());});
-	CONNECT(ui.buttonPreview, toggled, [this] { updatePreview(); });
+	CONNECT(ui.buttonPreview, toggled, [this] { previewToggled(); });
 	CONNECT(m_treeView, context_menu, [this](GdkEventButton * ev) {
 		showTreeWidgetContextMenu(ev);
 	});
@@ -354,7 +446,7 @@ OutputEditorHOCR::OutputEditorHOCR(DisplayerToolHOCR* tool) {
 	CONNECT(m_searchFrame, apply_substitutions, sigc::mem_fun(this, &OutputEditorHOCR::applySubstitutions));
 	m_connectionCustomFont = CONNECT(ConfigSettings::get<FontSetting>("customoutputfont"), changed, [this] { setFont(); });
 	m_connectionDefaultFont = CONNECT(ConfigSettings::get<SwitchSetting>("systemoutputfont"), changed, [this] { setFont(); });
-	m_connectionSelectionChanged = CONNECT(m_treeView, current_index_changed, sigc::mem_fun(this, &OutputEditorHOCR::showItemProperties));
+	CONNECT(m_treeView, current_index_changed, sigc::mem_fun(this, &OutputEditorHOCR::showItemProperties));
 	CONNECT(ui.notebook, switch_page, [this](Gtk::Widget*, guint) {
 		updateSourceText();
 	});
@@ -371,6 +463,7 @@ OutputEditorHOCR::OutputEditorHOCR(DisplayerToolHOCR* tool) {
 		setModified();
 		updateSourceText();
 	});
+	CONNECTX(m_document, item_attribute_changed, [this](const Gtk::TreeIter & itemIndex, const Glib::ustring & name, const Glib::ustring & value) { itemAttributeChanged(itemIndex, name, value); });
 	CONNECTX(m_document, item_attribute_changed, [this](const Gtk::TreeIter&, const Glib::ustring&, const Glib::ustring&) {
 		setModified();
 	});
@@ -383,6 +476,7 @@ OutputEditorHOCR::OutputEditorHOCR(DisplayerToolHOCR* tool) {
 	CONNECT(ui.buttonNavigationPrev, clicked, [this] { navigateNextPrev(false); });
 	CONNECT(ui.buttonExpandAll, clicked, [this] { expandCollapseItemClass(true); });
 	CONNECT(ui.buttonCollapseAll, clicked, [this] { expandCollapseItemClass(false); });
+	m_connectionSourceChanged = CONNECT(MAIN->getDisplayer(), imageChanged, [this] { sourceChanged(); });
 
 	setFont();
 }
@@ -404,6 +498,11 @@ void OutputEditorHOCR::setFont() {
 	}
 }
 
+void OutputEditorHOCR::setInsertMode(InsertMode mode, const std::string& iconName) {
+	m_insertMode = mode;
+	ui.imageInsert->set(Gdk::Pixbuf::create_from_resource(Glib::ustring::compose("/org/gnome/gimagereader/%1", iconName)));
+}
+
 void OutputEditorHOCR::setModified() {
 	ui.buttonSave->set_sensitive(m_document->pageCount() > 0);
 	ui.buttonExport->set_sensitive(m_document->pageCount() > 0);
@@ -415,30 +514,34 @@ void OutputEditorHOCR::setModified() {
 
 OutputEditorHOCR::ReadSessionData* OutputEditorHOCR::initRead(tesseract::TessBaseAPI& tess) {
 	tess.SetPageSegMode(tesseract::PSM_AUTO_ONLY);
-	return new HOCRReadSessionData;
+	HOCRReadSessionData* data = new HOCRReadSessionData;
+	data->insertIndex = m_insertMode == InsertMode::Append ? m_document->pageCount() : currentPage();
+	return data;
 }
 
 void OutputEditorHOCR::read(tesseract::TessBaseAPI& tess, ReadSessionData* data) {
 	tess.SetVariable("hocr_font_info", "true");
-	char* text = tess.GetHOCRText(data->page);
-	Utils::runInMainThreadBlocking([&] { addPage(text, *data); });
+	char* text = tess.GetHOCRText(data->pageInfo.page);
+	HOCRReadSessionData* hdata = static_cast<HOCRReadSessionData*>(data);
+	Utils::runInMainThreadBlocking([&] { addPage(text, *hdata); });
 	delete[] text;
+	++hdata->insertIndex;
 }
 
 void OutputEditorHOCR::readError(const Glib::ustring& errorMsg, ReadSessionData* data) {
-	static_cast<HOCRReadSessionData*>(data)->errors.push_back(Glib::ustring::compose("%1[%2]: %3", data->file, data->page, errorMsg));
+	static_cast<HOCRReadSessionData*>(data)->errors.push_back(Glib::ustring::compose("%1[%2]: %3", data->pageInfo.filename, data->pageInfo.page, errorMsg));
 }
 
 void OutputEditorHOCR::finalizeRead(ReadSessionData* data) {
 	HOCRReadSessionData* hdata = static_cast<HOCRReadSessionData*>(data);
 	if(!hdata->errors.empty()) {
 		Glib::ustring message = Glib::ustring::compose(_("The following pages could not be processed:\n%1"), Utils::string_join(hdata->errors, "\n"));
-		Utils::message_dialog(Gtk::MESSAGE_WARNING, _("Recognition errors"), message);
+		Utils::messageBox(Gtk::MESSAGE_WARNING, _("Recognition errors"), message);
 	}
 	OutputEditor::finalizeRead(data);
 }
 
-void OutputEditorHOCR::addPage(const Glib::ustring& hocrText, ReadSessionData data) {
+void OutputEditorHOCR::addPage(const Glib::ustring& hocrText, HOCRReadSessionData data) {
 	xmlpp::DomParser parser;
 	parser.parse_memory(hocrText);
 	xmlpp::Document* doc = parser.get_document();
@@ -448,13 +551,13 @@ void OutputEditorHOCR::addPage(const Glib::ustring& hocrText, ReadSessionData da
 		return;
 	}
 	std::map<Glib::ustring, Glib::ustring> attrs = HOCRItem::deserializeAttrGroup(pageDiv->get_attribute_value("title"));
-	attrs["image"] = Glib::ustring::compose("'%1'", data.file);
-	attrs["ppageno"] = Glib::ustring::compose("%1", data.page);
-	attrs["rot"] = Glib::ustring::compose("%1", data.angle);
-	attrs["res"] = Glib::ustring::compose("%1", data.resolution);
+	attrs["image"] = Glib::ustring::compose("'%1'", data.pageInfo.filename);
+	attrs["ppageno"] = Glib::ustring::compose("%1", data.pageInfo.page);
+	attrs["rot"] = Glib::ustring::compose("%1", data.pageInfo.angle);
+	attrs["res"] = Glib::ustring::compose("%1", data.pageInfo.resolution);
 	pageDiv->set_attribute("title", HOCRItem::serializeAttrGroup(attrs));
 
-	Gtk::TreeIter index = m_document->addPage(pageDiv, true);
+	Gtk::TreeIter index = m_document->insertPage(data.insertIndex, pageDiv, true);
 	expandCollapseChildren(index, true);
 	m_treeView->expand_to_path(m_document->get_path(index));
 	m_treeView->expand_row(m_document->get_path(index), true);
@@ -462,6 +565,16 @@ void OutputEditorHOCR::addPage(const Glib::ustring& hocrText, ReadSessionData da
 	m_modified = true;
 }
 
+bool OutputEditorHOCR::containsSource(const std::string& source, int sourcePage) const {
+	for(int i = 0, n = m_document->pageCount(); i < n; ++i) {
+		const HOCRPage* page = m_document->page(i);
+		if(page->pageNr() == sourcePage && page->sourceFile() == source) {
+			return true;
+		}
+	}
+	return false;
+}
+
 void OutputEditorHOCR::navigateTargetChanged() {
 	Glib::ustring target = (*ui.comboNavigation->get_active())[m_navigationComboCols.itemClass];
 	bool allowExpandCollapse = target.substr(0, 9) != "ocrx_word";
@@ -496,24 +609,17 @@ void OutputEditorHOCR::expandCollapseItemClass(bool expand) {
 void OutputEditorHOCR::navigateNextPrev(bool next) {
 	Glib::ustring target = (*ui.comboNavigation->get_active())[m_navigationComboCols.itemClass];
 	bool misspelled = false;
+	bool lowconf = false;
 	if(target == "ocrx_word_bad") {
 		target = "ocrx_word";
 		misspelled = true;
+	} else if(target == "ocrx_word_lowconf") {
+		target = "ocrx_word";
+		lowconf = true;
 	}
+
 	Gtk::TreeIter start = m_treeView->currentIndex();
-	if(!start) {
-		start = m_document->get_iter("0");
-	}
-	Gtk::TreeIter curr = next ? m_document->nextIndex(start) : m_document->prevIndex(start);
-	while(curr != start) {
-		const HOCRItem* item = m_document->itemAtIndex(curr);
-		if(item && item->itemClass() == target && (!misspelled || m_document->indexIsMisspelledWord(curr))) {
-			break;
-		}
-		curr = next ? m_document->nextIndex(curr) : m_document->prevIndex(curr);
-	};
-	// Sets current index
-	showItemProperties(curr);
+	m_treeView->setCurrentIndex(m_document->prevOrNextIndex(next, start, target, misspelled, lowconf));
 }
 
 void OutputEditorHOCR::expandCollapseChildren(const Gtk::TreeIter& index, bool expand) const {
@@ -526,24 +632,30 @@ void OutputEditorHOCR::expandCollapseChildren(const Gtk::TreeIter& index, bool e
 }
 
 bool OutputEditorHOCR::showPage(const HOCRPage* page) {
-	return page && MAIN->getSourceManager()->addSource(Gio::File::create_for_path(page->sourceFile()), true) && MAIN->getDisplayer()->setup(&page->pageNr(), &page->resolution(), &page->angle());
+	m_connectionSourceChanged.block();
+	bool success = page && MAIN->getSourceManager()->addSource(Gio::File::create_for_path(page->sourceFile()), true) && MAIN->getDisplayer()->setup(&page->pageNr(), &page->resolution(), &page->angle());
+	m_connectionSourceChanged.unblock();
+	if(success) {
+		sourceChanged();
+	}
+	return success;
+}
+
+int OutputEditorHOCR::currentPage() {
+	std::vector<Gtk::TreeModel::Path> selected = m_treeView->get_selection()->get_selected_rows();
+	if(selected.empty()) {
+		return m_document->pageCount();
+	}
+	Gtk::TreeModel::Path path = selected.front();
+	if(path.empty()) {
+		return m_document->pageCount();
+	}
+	return path.front();
 }
 
 void OutputEditorHOCR::showItemProperties(const Gtk::TreeIter& index, const Gtk::TreeIter& prev) {
 	m_tool->setAction(DisplayerToolHOCR::ACTION_NONE);
 	const HOCRItem* prevItem = m_document->itemAtIndex(prev);
-	Gtk::TreeIter current = m_treeView->currentIndex();
-	if(!current || current != index) {
-		m_connectionSelectionChanged.block(true);
-		m_treeView->get_selection()->unselect_all();
-		if(index) {
-			m_treeView->expand_to_path(m_document->get_path(index));
-			m_treeView->get_selection()->select(index);
-			m_treeView->set_cursor(m_document->get_path(index));
-			m_treeView->scroll_to_row(m_document->get_path(index));
-		}
-		m_connectionSelectionChanged.block(false);
-	}
 	m_propStore->clear();
 	ui.textviewSource->get_buffer()->set_text("");
 
@@ -674,6 +786,22 @@ void OutputEditorHOCR::updateSourceText() {
 	}
 }
 
+void OutputEditorHOCR::itemAttributeChanged(const Gtk::TreeIter& itemIndex, const Glib::ustring& name, const Glib::ustring& /*value*/) {
+	const HOCRItem* currentItem = m_document->itemAtIndex(itemIndex);
+	if(name == "title:bbox" && currentItem) {
+		// Minimum bounding box
+		Geometry::Rectangle minBBox;
+		if(currentItem->itemClass() == "ocr_page") {
+			minBBox = currentItem->bbox();
+		} else {
+			for(const HOCRItem* child : currentItem->children()) {
+				minBBox = minBBox.unite(child->bbox());
+			}
+		}
+		m_tool->setSelection(currentItem->bbox(), minBBox);
+	}
+}
+
 void OutputEditorHOCR::bboxDrawn(const Geometry::Rectangle& bbox, int action) {
 	xmlpp::Document doc;
 	Gtk::TreeIter current = m_treeView->currentIndex();
@@ -722,7 +850,7 @@ void OutputEditorHOCR::bboxDrawn(const Geometry::Rectangle& bbox, int action) {
 		newElement->set_attribute("class", "ocrx_word");
 		std::map<Glib::ustring, std::set<Glib::ustring>> propWord = propAttrs["ocrx_word"];
 
-		newElement->set_attribute("lang", propWord["lang"].size() == 1 ? *propWord["lang"].begin() : m_spell.get_language());
+		newElement->set_attribute("lang", propWord["lang"].size() == 1 ? *propWord["lang"].begin() : m_document->defaultLanguage());
 		std::map<Glib::ustring, Glib::ustring> titleAttrs;
 		titleAttrs["bbox"] = Glib::ustring::compose("%1 %2 %3 %4", bbox.x, bbox.y, bbox.x + bbox.width, bbox.y + bbox.height);
 		titleAttrs["x_font"] = propWord["title:x_font"].size() == 1 ? *propWord["title:x_font"].begin() : "sans";
@@ -736,13 +864,8 @@ void OutputEditorHOCR::bboxDrawn(const Geometry::Rectangle& bbox, int action) {
 		return;
 	}
 	Gtk::TreeIter index = m_document->addItem(current, newElement);
-	if(index) {
-		m_connectionSelectionChanged.block(true);
-		m_treeView->get_selection()->unselect_all();
-		m_connectionSelectionChanged.block(false);
-		m_treeView->expand_to_path(m_document->get_path(index));
-		m_treeView->get_selection()->select(index);
-		m_treeView->scroll_to_row(m_document->get_path(index));
+	if (index) {
+		m_treeView->setCurrentIndex(index);
 	}
 }
 
@@ -779,16 +902,14 @@ void OutputEditorHOCR::showTreeWidgetContextMenu(GdkEventButton* ev) {
 			menu.append(*mergeItem);
 			CONNECT(mergeItem, activate, [&] {
 				Gtk::TreeIter newIndex = m_document->mergeItems(m_document->get_iter(paths.front())->parent(), rows.front(), rows.back());
-				m_treeView->get_selection()->unselect_all();
-				m_treeView->get_selection()->select(newIndex);
+				m_treeView->setCurrentIndex(newIndex);
 			});
 			if(firstItem->itemClass() != "ocr_carea") {
 				Gtk::MenuItem* splitItem = Gtk::manage(new Gtk::MenuItem(_("Split from parent")));
 				menu.append(*splitItem);
 				CONNECT(splitItem, activate, [&] {
 					Gtk::TreeIter newIndex = m_document->splitItem(m_document->get_iter(paths.front())->parent(), rows.front(), rows.back());
-					m_treeView->get_selection()->unselect_all();
-					m_treeView->get_selection()->select(newIndex);
+					m_treeView->setCurrentIndex(newIndex);
 					expandCollapseChildren(newIndex, true);
 				});
 			}
@@ -798,8 +919,7 @@ void OutputEditorHOCR::showTreeWidgetContextMenu(GdkEventButton* ev) {
 			menu.append(*swapItem);
 			CONNECT(swapItem, activate, [&] {
 				Gtk::TreeIter newIndex = m_document->swapItems(m_document->get_iter(paths.front())->parent(), rows.front(), rows.back());
-				m_treeView->get_selection()->unselect_all();
-				m_treeView->get_selection()->select(newIndex);
+				m_treeView->setCurrentIndex(newIndex);
 			});
 		}
 		CONNECT(&menu, hide, [&] { loop->quit(); });
@@ -837,38 +957,7 @@ void OutputEditorHOCR::showTreeWidgetContextMenu(GdkEventButton* ev) {
 		menu.append(*addWordItem);
 		CONNECT(addWordItem, activate, [this] { m_tool->setAction(DisplayerToolHOCR::ACTION_DRAW_WORD_RECT); });
 	} else if(itemClass == "ocrx_word") {
-		Glib::ustring prefix, suffix, trimmedWord = HOCRItem::trimmedWord(item->text(), &prefix, &suffix);
-		std::vector<Glib::ustring> suggestions;
-		bool valid = m_document->checkItemSpelling(index, &suggestions, 16);
-		for(const Glib::ustring& suggestion : suggestions) {
-			Glib::ustring replacement = prefix + suggestion + suffix;
-			Gtk::MenuItem* item = Gtk::manage(new Gtk::MenuItem(replacement));
-			CONNECT(item, activate, [this, replacement, index] { m_document->editItemText(index, replacement); });
-			menu.append(*item);
-		}
-
-		if(!trimmedWord.empty()) {
-			if(menu.get_children().empty()) {
-				Gtk::MenuItem* item = Gtk::manage(new Gtk::MenuItem(_("No suggestions")));
-				item->set_sensitive(false);
-				menu.append(*item);
-			}
-			if(!valid) {
-				menu.append(*Gtk::manage(new Gtk::SeparatorMenuItem));
-				Gtk::MenuItem* additem = Gtk::manage(new Gtk::MenuItem(_("Add to dictionary")));
-				CONNECT(additem, activate, [this, index, trimmedWord] {
-					m_spell.add_to_dictionary(trimmedWord);
-					m_document->recheckSpelling();
-				});
-				menu.append(*additem);
-				Gtk::MenuItem* ignoreitem = Gtk::manage(new Gtk::MenuItem(_("Ignore word")));
-				CONNECT(ignoreitem, activate, [this, index, trimmedWord] {
-					m_spell.ignore_word(trimmedWord);
-					m_document->recheckSpelling();
-				});
-				menu.append(*ignoreitem);
-			}
-		}
+		m_document->addSpellingActions(&menu, index);
 	}
 	if(!menu.get_children().empty()) {
 		menu.append(*Gtk::manage(new Gtk::SeparatorMenuItem));
@@ -877,8 +966,7 @@ void OutputEditorHOCR::showTreeWidgetContextMenu(GdkEventButton* ev) {
 		Gtk::MenuItem* splitItem = Gtk::manage(new Gtk::MenuItem(_("Split from parent")));
 		CONNECT(splitItem, activate, [this, index, item] {
 			Gtk::TreeIter newIndex = m_document->splitItem(index->parent(), item->index(), item->index());
-			m_treeView->get_selection()->unselect_all();
-			m_treeView->get_selection()->select(newIndex);
+			m_treeView->setCurrentIndex(newIndex);
 			expandCollapseChildren(newIndex, true);
 		});
 		menu.append(*splitItem);
@@ -915,43 +1003,84 @@ void OutputEditorHOCR::pickItem(const Geometry::Point& point) {
 	double scale = double(page->resolution()) / double(MAIN->getDisplayer()->getCurrentResolution());
 	Geometry::Point newPoint( scale * (point.x * std::cos(alpha) - point.y * std::sin(alpha)) + 0.5 * page->bbox().width,
 	                          scale * (point.x * std::sin(alpha) + point.y * std::cos(alpha)) + 0.5 * page->bbox().height);
-	showItemProperties(m_document->searchAtCanvasPos(pageIndex, newPoint));
+	m_treeView->setCurrentIndex(m_document->searchAtCanvasPos(pageIndex, newPoint));
 	m_treeView->grab_focus();
 }
 
-void OutputEditorHOCR::open() {
-	if(!clear(false)) {
-		return;
+bool OutputEditorHOCR::open(InsertMode mode, std::vector<Glib::RefPtr<Gio::File>> files) {
+	if(mode == InsertMode::Replace && !clear(false)) {
+		return false;
 	}
-	FileDialogs::FileFilter filter = {_("hOCR HTML Files"), {"text/html", "text/xml", "text/plain"}, {"*.html"}};
-	std::vector<Glib::RefPtr<Gio::File>> files = FileDialogs::open_dialog(_("Open hOCR File"), "", "outputdir", filter, false);
 	if(files.empty()) {
-		return;
+		FileDialogs::FileFilter filter = {_("hOCR HTML Files"), {"text/html", "text/xml", "text/plain"}, {"*.html"}};
+		files = FileDialogs::open_dialog(_("Open hOCR File"), "", "outputdir", filter, true);
 	}
-	std::string filename = files.front()->get_path();
-	std::string source;
-	try {
-		source = Glib::file_get_contents(filename);
-	} catch(Glib::Error&) {
-		Utils::message_dialog(Gtk::MESSAGE_ERROR, _("Failed to open file"), Glib::ustring::compose(_("The file could not be opened: %1."), filename));
-		return;
+	if(files.empty()) {
+		return false;
 	}
+	std::vector<Glib::ustring> failed;
+	std::vector<Glib::ustring> invalid;
+	int added = 0;
+	for(const Glib::RefPtr<Gio::File>& file : files) {
+		std::string filename = file->get_path();
+		std::string source;
+		try {
+			source = Glib::file_get_contents(filename);
+		} catch(const Glib::Error&) {
+			failed.push_back(filename);
+			continue;
+		}
 
-	xmlpp::DomParser parser;
-	parser.parse_memory(source);
-	xmlpp::Document* doc = parser.get_document();
-	const xmlpp::Element* div = XmlUtils::firstChildElement(XmlUtils::firstChildElement(doc->get_root_node(), "body"), "div");
-	if(!div || div->get_attribute_value("class") != "ocr_page") {
-		Utils::message_dialog(Gtk::MESSAGE_ERROR, _("Invalid hOCR file"), Glib::ustring::compose(_("The file does not appear to contain valid hOCR HTML: %1"), filename));
-		return;
+		xmlpp::DomParser parser;
+		try {
+			parser.parse_memory(source);
+		} catch(const xmlpp::exception&) {
+			failed.push_back(filename);
+			continue;
+		}
+
+		xmlpp::Document* doc = parser.get_document();
+		const xmlpp::Element* div = XmlUtils::firstChildElement(XmlUtils::firstChildElement(doc->get_root_node(), "body"), "div");
+		if(!div || div->get_attribute_value("class") != "ocr_page") {
+			invalid.push_back(filename);
+			continue;
+		}
+		int pos = mode == InsertMode::InsertBefore ? currentPage() : m_document->pageCount();
+		while(div) {
+			m_document->insertPage(pos++, div, false, Glib::path_get_dirname(filename));
+			div = XmlUtils::nextSiblingElement(div, "div");
+			++added;
+		}
 	}
-	while(div) {
-		m_document->addPage(div, false);
-		div = XmlUtils::nextSiblingElement(div, "div");
+	if(added > 0) {
+		m_modified = mode != InsertMode::Replace;
+		if(mode == InsertMode::Replace && m_filebasename.empty()) {
+			m_filebasename = Utils::split_filename(files.front()->get_path()).first;
+		}
+		MAIN->setOutputPaneVisible(true);
 	}
-	m_document->convertSourcePaths(Glib::path_get_dirname(filename), true);
-	m_modified = false;
-	m_filebasename = Utils::split_filename(filename).first;
+	std::vector<Glib::ustring> errorMsg;
+	if(!failed.empty()) {
+		errorMsg.push_back(Glib::ustring::compose(_("The following files could not be opened:\n%1"), Utils::string_join(failed, "\n")));
+	}
+	if(!invalid.empty()) {
+		errorMsg.push_back(Glib::ustring::compose(_("The following files are not valid hOCR HTML:\n%1"), Utils::string_join(invalid, "\n")));
+	}
+	if(!errorMsg.empty()) {
+		Utils::messageBox(Gtk::MESSAGE_ERROR, _("Unable to open files"), Utils::string_join(errorMsg, "\n\n"));
+	}
+	return added > 0;
+}
+
+bool OutputEditorHOCR::selectPage(int nr) {
+	if(!m_document || nr >= m_document->pageCount()) {
+		return false;
+	}
+	Gtk::TreeIter index = m_document->indexAtItem(m_document->page(nr));
+	if(index) {
+		m_treeView->setCurrentIndex(index);
+	}
+	return index;
 }
 
 bool OutputEditorHOCR::save(const std::string& filename) {
@@ -960,7 +1089,11 @@ bool OutputEditorHOCR::save(const std::string& filename) {
 		Glib::ustring suggestion = m_filebasename;
 		if(suggestion.empty()) {
 			std::vector<Source*> sources = MAIN->getSourceManager()->getSelectedSources();
-			suggestion = !sources.empty() ? Utils::split_filename(sources.front()->displayname).first : _("output");
+			if(!sources.empty()) {
+				suggestion = Utils::split_filename(sources.front()->file->get_path()).first;
+			} else {
+				suggestion = _("output");
+			}
 		}
 		FileDialogs::FileFilter filter = {_("hOCR HTML Files"), {"text/html"}, {"*.html"}};
 		outname = FileDialogs::save_dialog(_("Save hOCR Output..."), suggestion + ".html", "outputdir", filter);
@@ -970,7 +1103,7 @@ bool OutputEditorHOCR::save(const std::string& filename) {
 	}
 	std::ofstream file(outname);
 	if(!file.is_open()) {
-		Utils::message_dialog(Gtk::MESSAGE_ERROR, _("Failed to save output"), _("Check that you have writing permissions in the selected folder."));
+		Utils::messageBox(Gtk::MESSAGE_ERROR, _("Failed to save output"), _("Check that you have writing permissions in the selected folder."));
 		return false;
 	}
 	std::string current = setlocale(LC_ALL, NULL);
@@ -998,11 +1131,47 @@ bool OutputEditorHOCR::save(const std::string& filename) {
 	return true;
 }
 
+bool OutputEditorHOCR::crashSave(const std::string& filename) const {
+	std::ofstream file(filename);
+	if(file.is_open()) {
+		Glib::ustring header = Glib::ustring::compose(
+		                           "<!DOCTYPE html>\n"
+		                           "<html>\n"
+		                           "<head>\n"
+		                           " <title>%1</title>\n"
+		                           " <meta charset=\"utf-8\" /> \n"
+		                           " <meta name='ocr-capabilities' content='ocr_page ocr_carea ocr_par ocr_line ocrx_word'/>\n"
+		                           "</head>\n", Glib::path_get_basename(filename));
+		m_document->convertSourcePaths(Glib::path_get_dirname(filename), false);
+		Glib::ustring body = m_document->toHTML();
+		m_document->convertSourcePaths(Glib::path_get_dirname(filename), true);
+		Glib::ustring footer = "</html>\n";
+		file.write(header.data(), header.bytes());
+		file.write(body.data(), body.bytes());
+		file.write(footer.data(), footer.bytes());
+		return true;
+	}
+	return false;
+}
+
 bool OutputEditorHOCR::exportToODT() {
-	if(m_document->pageCount() == 0) {
+	std::string suggestion = m_filebasename;
+	if(suggestion.empty()) {
+		std::vector<Source*> sources = MAIN->getSourceManager()->getSelectedSources();
+		suggestion = !sources.empty() ? Glib::path_get_basename(sources.front()->displayname) : _("output");
+	}
+
+	FileDialogs::FileFilter filter = {_("OpenDocument Text Documents"), {"application/vnd.oasis.opendocument.text"}, {"*.odt"}};
+	std::string outname = FileDialogs::save_dialog(_("Save ODT Output..."), suggestion + ".odt", "outputdir", filter);
+	if(outname.empty()) {
 		return false;
 	}
-	return HOCROdtExporter(m_tool).run(m_document, m_filebasename);
+
+	m_treeView->grab_focus(); // Ensure any item editor loses focus and commits its changes
+	MAIN->getDisplayer()->setBlockAutoscale(true);
+	bool success = HOCROdtExporter().run(m_document.get(), m_filebasename);
+	MAIN->getDisplayer()->setBlockAutoscale(false);
+	return success;
 }
 
 bool OutputEditorHOCR::exportToPDF() {
@@ -1012,27 +1181,70 @@ bool OutputEditorHOCR::exportToPDF() {
 	ui.buttonPreview->set_active(false); // Disable preview because if conflicts with preview from PDF dialog
 	const HOCRItem* item = m_document->itemAtIndex(m_treeView->currentIndex());
 	const HOCRPage* page = item ? item->page() : m_document->page(0);
-	if(showPage(page)) {
-		return HOCRPdfExporter(m_document, page, m_tool).run(m_filebasename);
+	bool success = false;
+	if(!showPage(page)) {
+		return false;
 	}
-	return false;
+	HOCRPdfExportDialog dialog(m_tool, m_document.get(), page, MAIN->getWindow());
+	if(dialog.run() != Gtk::RESPONSE_OK) {
+		return false;
+	}
+	HOCRPdfExporter::PDFSettings settings = dialog.getPdfSettings();
+
+	std::string suggestion = m_filebasename;
+	if(suggestion.empty()) {
+		std::vector<Source*> sources = MAIN->getSourceManager()->getSelectedSources();
+		suggestion = !sources.empty() ? Glib::path_get_basename(sources.front()->displayname) : _("output");
+	}
+
+	std::string outname;
+	while(true) {
+		FileDialogs::FileFilter filter = {_("PDF Files"), {"application/pdf"}, {"*.pdf"}};
+		outname = FileDialogs::save_dialog(_("Save PDF Output..."), suggestion + ".pdf", "outputdir", filter);
+		if(outname.empty()) {
+			break;
+		}
+		if(m_document->referencesSource(outname)) {
+			Utils::messageBox(Gtk::MESSAGE_ERROR, _("Invalid Output"), _("Cannot overwrite a file which is a source image of this document."));
+			continue;
+		}
+		break;
+	}
+	if(outname.empty()) {
+		return false;
+	}
+
+	MAIN->getDisplayer()->setBlockAutoscale(true);
+	success = HOCRPdfExporter().run(m_document.get(), outname, &settings);
+	MAIN->getDisplayer()->setBlockAutoscale(false);
+	return success;
 }
 
 bool OutputEditorHOCR::exportToText() {
-	if(m_document->pageCount() == 0) {
+	std::string suggestion = m_filebasename;
+	if(suggestion.empty()) {
+		std::vector<Source*> sources = MAIN->getSourceManager()->getSelectedSources();
+		suggestion = !sources.empty() ? Glib::path_get_basename(sources.front()->displayname) : _("output");
+	}
+
+	FileDialogs::FileFilter filter = {_("Text Files"), {"text/plain"}, {"*.txt"}};
+	std::string outname = FileDialogs::save_dialog(_("Save Text Output..."), suggestion + ".txt", "outputdir", filter);
+	if(outname.empty()) {
 		return false;
 	}
-	return HOCRTextExporter().run(m_document, m_filebasename);
+
+	m_treeView->grab_focus(); // Ensure any item editor loses focus and commits its changes
+	return HOCRTextExporter().run(m_document.get(), m_filebasename);
 }
 
 bool OutputEditorHOCR::clear(bool hide) {
 	m_connectionPreviewTimer.disconnect();
-	m_preview->setVisible(false);
+	ui.buttonPreview->set_active(false);
 	if(!ui.boxEditorHOCR->get_visible()) {
 		return true;
 	}
-	if(getModified()) {
-		int response = Utils::question_dialog(_("Output not saved"), _("Save output before proceeding?"), Utils::Button::Save | Utils::Button::Discard | Utils::Button::Cancel);
+	if(m_modified) {
+		int response = Utils::messageBox(Gtk::MESSAGE_QUESTION, _("Output not saved"), _("Save output before proceeding?"), "", Utils::Button::Save | Utils::Button::Discard | Utils::Button::Cancel);
 		if(response == Utils::Button::Save) {
 			if(!save()) {
 				return false;
@@ -1041,7 +1253,7 @@ bool OutputEditorHOCR::clear(bool hide) {
 			return false;
 		}
 	}
-	m_document = Glib::RefPtr<HOCRDocument>(new HOCRDocument(&m_spell));
+	m_document = Glib::RefPtr<HOCRDocument>(new HOCRDocument());
 	CONNECTX(m_document, row_inserted, [this](const Gtk::TreePath&, const Gtk::TreeIter&) {
 		setModified();
 	});
@@ -1122,8 +1334,7 @@ bool OutputEditorHOCR::findReplaceInItem(const Gtk::TreeIter& index, const Glib:
 		if(m_treeViewTextCell->get_entry()) {
 			m_treeViewTextCell->get_entry()->editing_done(); // Commit previous changes
 		}
-		m_treeView->get_selection()->unselect_all();
-		m_treeView->get_selection()->select(path);
+		m_treeView->setCurrentIndex(m_document->get_iter(path));
 		m_treeView->set_cursor(path, *m_treeView->get_column(0), *m_treeViewTextCell, true);
 		g_assert(m_treeViewTextCell->get_current_path() == path.to_string() && m_treeViewTextCell->get_entry());
 		m_treeViewTextCell->get_entry()->select_region(pos, pos + searchstr.length());
@@ -1201,6 +1412,34 @@ void OutputEditorHOCR::applySubstitutions(const std::map<Glib::ustring, Glib::us
 	MAIN->popState();
 }
 
+void OutputEditorHOCR::sourceChanged() {
+	int page;
+	std::string path = MAIN->getDisplayer()->getCurrentImage(page);
+	// Check if source is in document tree
+	Gtk::TreeIter pageIndex = m_document->searchPage(path, page);
+	if(!pageIndex) {
+		ui.buttonPreview->set_active(false);
+		m_treeView->setCurrentIndex(Gtk::TreeIter());
+	} else {
+		Gtk::TreeIter curIndex = m_treeView->currentIndex();
+		while(curIndex && curIndex != pageIndex && curIndex->parent()) {
+			curIndex = curIndex->parent();
+		}
+		if(curIndex && curIndex != pageIndex) {
+			m_treeView->setCurrentIndex(pageIndex);
+		}
+	}
+}
+
+void OutputEditorHOCR::previewToggled() {
+	Gtk::TreeIter index = m_treeView->currentIndex();
+	if(ui.buttonPreview->get_active() && !index && m_document->pageCount() > 0) {
+		m_treeView->setCurrentIndex(m_document->indexAtItem(m_document->page(0)));
+	} else {
+		updatePreview();
+	}
+}
+
 void OutputEditorHOCR::updatePreview() {
 	const HOCRItem* item = m_document->itemAtIndex(m_treeView->currentIndex());
 	if(!ui.buttonPreview->get_active() || !item) {
@@ -1210,7 +1449,6 @@ void OutputEditorHOCR::updatePreview() {
 
 	const HOCRPage* page = item->page();
 	const Geometry::Rectangle& bbox = page->bbox();
-	int pageDpi = page->resolution();
 
 	Cairo::RefPtr<Cairo::ImageSurface> image = Cairo::ImageSurface::create(Cairo::FORMAT_ARGB32, bbox.width, bbox.height);
 	Cairo::RefPtr<Cairo::Context> context = Cairo::Context::create(image);
