@@ -5,7 +5,7 @@
  *   Copyright (C) 2009-2013 Canonical Ltd.
  *   Author: Robert Ancell <robert.ancell@canonical.com>
  * Modifications are:
- *   Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ *   Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -23,10 +23,11 @@
 
 #include "ScannerSane.hh"
 #include <cstring>
+#include <thread>
 #include <sane/saneopts.h>
 
 void ScannerSane::init() {
-	m_thread = Glib::Threads::Thread::create([this] { run(); });
+	m_thread = new std::thread(&ScannerSane::run, this);
 }
 
 void ScannerSane::redetect() {
@@ -50,6 +51,7 @@ void ScannerSane::close() {
 		}
 	}
 	m_thread->join();
+	delete m_thread;
 	m_thread = nullptr;
 }
 
@@ -422,7 +424,7 @@ void ScannerSane::doGetParameters() {
 
 	m_job->height = std::max(1, m_job->parameters.lines);
 	m_job->rowstride = m_job->parameters.pixels_per_line * 3; // Buffer is for a 24 bit RGB image
-	m_job->imgbuf.resize(m_job->rowstride * m_job->parameters.lines);
+	m_job->imgbuf.resize(std::size_t(m_job->rowstride) * m_job->parameters.lines);
 
 	/* Prepare for read */
 	m_job->lineBuffer.resize(m_job->parameters.bytes_per_line);
@@ -472,7 +474,7 @@ void ScannerSane::doRead() {
 		// Resize image if necessary
 		if(m_job->lineCount > m_job->height) {
 			m_job->height = m_job->lineCount;
-			m_job->imgbuf.resize(m_job->rowstride * m_job->height);
+			m_job->imgbuf.resize(std::size_t(m_job->rowstride) * m_job->height);
 		}
 
 		// Write data
@@ -511,7 +513,8 @@ void ScannerSane::doCompletePage() {
 	std::string filename = m_job->params.filename;
 	if(m_job->params.type != ScanType::SINGLE) {
 		std::pair<std::string, std::string> parts = Utils::split_filename(filename);
-		filename = Glib::ustring::compose("%1_%2.%3", parts.first, m_job->pageNumber, parts.second);
+		filename = Glib::ustring::compose("%1-%2.%3", parts.first, m_job->pageNumber, parts.second);
+		filename = Utils::make_output_filename(filename);
 	}
 	std::string ext = Utils::split_filename(filename).second;
 	try {
