@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * Config.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -26,13 +26,14 @@
 #include <QDesktopServices>
 #include <QDir>
 #include <QMultiMap>
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
 #include <QStandardPaths>
-#endif
 #include <QUrl>
 #include <enchant-provider.h>
 #define USE_STD_NAMESPACE
 #include <tesseract/baseapi.h>
+#if TESSERACT_MAJOR_VERSION < 5
+#include <tesseract/genericvector.h>
+#endif
 #undef USE_STD_NAMESPACE
 
 const QList<Config::Lang> Config::LANGUAGES = LangTables::languages<QList<Config::Lang>, QString>([](const char* str) { return QString::fromUtf8(str); });
@@ -51,13 +52,8 @@ Config::Config(QWidget* parent)
 	: QDialog(parent) {
 	ui.setupUi(this);
 
-#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
-	ui.tableWidgetPredefLang->horizontalHeader()->setResizeMode(QHeaderView::Stretch);
-	ui.tableWidgetAdditionalLang->horizontalHeader()->setResizeMode(QHeaderView::Stretch);
-#else
 	ui.tableWidgetPredefLang->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);
 	ui.tableWidgetAdditionalLang->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);
-#endif
 	ui.widgetAddLang->setVisible(false);
 
 #if !ENABLE_VERSIONCHECK
@@ -72,19 +68,19 @@ Config::Config(QWidget* parent)
 		ui.tableWidgetPredefLang->setItem(row, 2, new QTableWidgetItem(lang.name));
 	}
 
-	connect(ui.checkBoxDefaultOutputFont, SIGNAL(toggled(bool)), ui.pushButtonOutputFont, SLOT(setDisabled(bool)));
-	connect(ui.pushButtonOutputFont, SIGNAL(clicked()), &m_fontDialog, SLOT(exec()));
-	connect(&m_fontDialog, SIGNAL(fontSelected(QFont)), this, SLOT(updateFontButton(QFont)));
-	connect(ui.pushButtonAddLang, SIGNAL(clicked()), this, SLOT(toggleAddLanguage()));
-	connect(ui.pushButtonRemoveLang, SIGNAL(clicked()), this, SLOT(removeLanguage()));
-	connect(ui.pushButtonAddLangOk, SIGNAL(clicked()), this, SLOT(addLanguage()));
-	connect(ui.pushButtonAddLangCancel, SIGNAL(clicked()), this, SLOT(toggleAddLanguage()));
-	connect(ui.tableWidgetAdditionalLang->selectionModel(), SIGNAL(selectionChanged(QItemSelection, QItemSelection)), this, SLOT(langTableSelectionChanged(QItemSelection, QItemSelection)));
-	connect(ui.buttonBox->button(QDialogButtonBox::Help), SIGNAL(clicked()), MAIN, SLOT(showHelp()));
-	connect(ui.lineEditLangPrefix, SIGNAL(textChanged(QString)), this, SLOT(clearLineEditErrorState()));
-	connect(ui.lineEditLangName, SIGNAL(textChanged(QString)), this, SLOT(clearLineEditErrorState()));
-	connect(ui.lineEditLangCode, SIGNAL(textChanged(QString)), this, SLOT(clearLineEditErrorState()));
-	connect(ui.comboBoxDataLocation, SIGNAL(currentIndexChanged(int)), this, SLOT(setDataLocations(int)));
+	connect(ui.checkBoxDefaultOutputFont, &QCheckBox::toggled, ui.pushButtonOutputFont, &QPushButton::setDisabled);
+	connect(ui.pushButtonOutputFont, &QPushButton::clicked, &m_fontDialog, &QFontDialog::exec);
+	connect(&m_fontDialog, &QFontDialog::fontSelected, this, &Config::updateFontButton);
+	connect(ui.pushButtonAddLang, &QPushButton::clicked, this, &Config::toggleAddLanguage);
+	connect(ui.pushButtonRemoveLang, &QPushButton::clicked, this, &Config::removeLanguage);
+	connect(ui.pushButtonAddLangOk, &QPushButton::clicked, this, &Config::addLanguage);
+	connect(ui.pushButtonAddLangCancel, &QPushButton::clicked, this, &Config::toggleAddLanguage);
+	connect(ui.tableWidgetAdditionalLang->selectionModel(), &QItemSelectionModel::selectionChanged, this, &Config::langTableSelectionChanged);
+	connect(ui.buttonBox->button(QDialogButtonBox::Help), &QPushButton::clicked, MAIN, [] { MAIN->showHelp(); });
+	connect(ui.lineEditLangPrefix, &QLineEdit::textChanged, this, &Config::clearLineEditErrorState);
+	connect(ui.lineEditLangName, &QLineEdit::textChanged, this, &Config::clearLineEditErrorState);
+	connect(ui.lineEditLangCode, &QLineEdit::textChanged, this, &Config::clearLineEditErrorState);
+	connect(ui.comboBoxDataLocation, qOverload<int>(&QComboBox::currentIndexChanged), this, &Config::setDataLocations);
 
 	ADD_SETTING(SwitchSetting("dictinstall", ui.checkBoxDictInstall, true));
 	ADD_SETTING(SwitchSetting("updatecheck", ui.checkBoxUpdateCheck, true));
@@ -119,7 +115,7 @@ bool Config::searchLangSpec(Lang& lang) const {
 	return false;
 }
 
-QList<QString> Config::searchLangCultures(const QString& code) const {
+QStringList Config::searchLangCultures(const QString& code) const {
 	return LANGUAGE_CULTURES.values(code);
 }
 
@@ -145,6 +141,33 @@ QString Config::spellingLocation() const {
 	return ui.lineEditSpellLocation->text();
 }
 
+QStringList Config::getAvailableLanguages() {
+	Utils::TesseractHandle tess;
+	if(!tess.get()) {
+		return QStringList();
+	}
+#if TESSERACT_MAJOR_VERSION < 5
+	GenericVector<STRING> availLanguages;
+#else
+	std::vector<std::string> availLanguages;
+#endif
+	tess.get()->GetAvailableLanguagesAsVector(&availLanguages);
+	QStringList result;
+	for(int i = 0; i < availLanguages.size(); ++i) {
+		result.append(availLanguages[i].c_str());
+	}
+	std::sort(result.begin(), result.end(), [](const QString & s1, const QString & s2) {
+		bool s1Script = s1.startsWith("script") || s1.left(1) == s1.left(1).toUpper();
+		bool s2Script = s2.startsWith("script") || s2.left(1) == s2.left(1).toUpper();
+		if(s1Script != s2Script) {
+			return !s1Script;
+		} else {
+			return s1 < s2;
+		}
+	});
+	return result;
+}
+
 void Config::disableDictInstall() {
 	ConfigSettings::get<SwitchSetting>("dictinstall")->setValue(false);
 }
@@ -191,15 +214,7 @@ QString Config::spellingLocation(Location location) {
 		return dataDir.absoluteFilePath("myspell/dicts");
 #endif
 	} else {
-#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
-# ifdef Q_OS_WIN
-		QDir configDir = QDir(QDir::home().absoluteFilePath("Local Settings/Application Data"));
-# else
-		QDir configDir = QDir(QDir::home().absoluteFilePath(".config"));
-# endif
-#else
 		QDir configDir = QDir(QStandardPaths::writableLocation(QStandardPaths::GenericConfigLocation));
-#endif
 #if HAVE_ENCHANT2
 		return configDir.absoluteFilePath("enchant/hunspell");
 #else
@@ -214,22 +229,10 @@ QString Config::tessdataLocation(Location location) {
 		QDir dataDir = QDir(QString("%1/../share/").arg(QApplication::applicationDirPath()));
 		qputenv("TESSDATA_PREFIX", dataDir.absoluteFilePath("tessdata").toLocal8Bit());
 #else
-# if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
-		unsetenv("TESSDATA_PREFIX");
-# else
 		qunsetenv("TESSDATA_PREFIX");
-# endif
 #endif
 	} else {
-#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
-# ifdef Q_OS_WIN
-		QDir configDir = QDir(QDir::home().absoluteFilePath("Local Settings/Application Data"));
-# else
-		QDir configDir = QDir(QDir::home().absoluteFilePath(".config"));
-# endif
-#else
 		QDir configDir = QDir(QStandardPaths::writableLocation(QStandardPaths::GenericConfigLocation));
-#endif
 		qputenv("TESSDATA_PREFIX", configDir.absoluteFilePath("tessdata").toLocal8Bit());
 	}
 	QByteArray current = setlocale(LC_ALL, NULL);
@@ -260,15 +263,15 @@ void Config::toggleAddLanguage(bool forceHide) {
 void Config::addLanguage() {
 	QString errorStyle = "background: #FF7777; color: #FFFFFF;";
 	bool invalid = false;
-	if(!QRegExp("\\w+").exactMatch(ui.lineEditLangPrefix->text())) {
+	if(!QRegularExpression("^\\w+$").match(ui.lineEditLangPrefix->text()).hasMatch()) {
 		invalid = true;
 		ui.lineEditLangPrefix->setStyleSheet(errorStyle);
 	}
-	if(!QRegExp(".+").exactMatch(ui.lineEditLangName->text())) {
+	if(!QRegularExpression("^.+$").match(ui.lineEditLangName->text()).hasMatch()) {
 		invalid = true;
 		ui.lineEditLangName->setStyleSheet(errorStyle);
 	}
-	if(!QRegExp("[a-z]{2,}(_[A-Z]{2,})?").exactMatch(ui.lineEditLangCode->text())) {
+	if(!QRegularExpression("^[a-z]{2,}(_[A-Z]{2,})?$").match(ui.lineEditLangCode->text()).hasMatch()) {
 		invalid = true;
 		ui.lineEditLangCode->setStyleSheet(errorStyle);
 	}
