@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * DisplayerToolSelect.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -151,7 +151,7 @@ void DisplayerToolSelect::autodetectLayout(bool noDeskew) {
 	QImage img = m_displayer->getImage(m_displayer->getSceneBoundingRect());
 
 	// Perform layout analysis
-	Utils::busyTask([this, &nDeskew, &avgDeskew, &rects, &img] {
+	Utils::busyTask([&nDeskew, &avgDeskew, &rects, &img] {
 		QByteArray current = setlocale(LC_ALL, NULL);
 		setlocale(LC_ALL, "C");
 		tesseract::TessBaseAPI tess;
@@ -230,7 +230,7 @@ void NumberedDisplayerSelection::contextMenuEvent(QGraphicsSceneContextMenuEvent
 	orderSpin->setRange(1, static_cast<DisplayerToolSelect*>(m_tool)->m_selections.size());
 	orderSpin->setValue(m_number);
 	orderSpin->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
-	connect(orderSpin, SIGNAL(valueChanged(int)), this, SLOT(reorderSelection(int)));
+	connect(orderSpin, qOverload<int>(&QSpinBox::valueChanged), this, &NumberedDisplayerSelection::reorderSelection);
 	layout->addWidget(orderSpin);
 
 	QWidgetAction* spinAction = new QWidgetAction(&menu);
