@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * MainWindow.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -28,6 +28,7 @@
 #include <QProgressBar>
 #include <QStatusBar>
 #include <QUrl>
+#include <QtConcurrent/QtConcurrentRun>
 #include <csignal>
 #include <iostream>
 #ifdef Q_OS_LINUX
@@ -44,8 +45,10 @@
 #include "Displayer.hh"
 #include "DisplayerToolSelect.hh"
 #include "DisplayerToolHOCR.hh"
+#include "HOCRBatchExportDialog.hh"
 #include "OutputEditorText.hh"
 #include "OutputEditorHOCR.hh"
+#include "RecognitionMenu.hh"
 #include "Recognizer.hh"
 #include "SourceManager.hh"
 #include "TessdataManager.hh"
@@ -69,21 +72,27 @@ void MainWindow::signalHandlerExec(int signal, bool tesseractCrash) {
 	std::signal(signal, nullptr);
 
 	QString filename;
-	if(MAIN->getOutputEditor() && MAIN->getOutputEditor()->getModified()) {
+	if(MAIN->getOutputEditor()) {
 		filename = QDir(Utils::documentsFolder()).absoluteFilePath(QString("%1_crash-save.txt").arg(PACKAGE_NAME));
 		int i = 0;
 		while(QFile(filename).exists()) {
 			++i;
 			filename = QDir(Utils::documentsFolder()).absoluteFilePath(QString("%1_crash-save_%2.txt").arg(PACKAGE_NAME).arg(i));
 		}
-		MAIN->getOutputEditor()->save(filename);
+		if(!MAIN->getOutputEditor()->crashSave(filename)) {
+			filename.clear();
+		}
 	}
 
 	QProcess process;
 	process.start(QApplication::applicationFilePath(), QStringList() << "crashhandle" << QString::number(QApplication::applicationPid()) << QString::number(tesseractCrash) << filename);
 #ifdef Q_OS_LINUX
 	// Allow crash handler spawned debugger to attach to the crashed process
+#if QT_VERSION >= QT_VERSION_CHECK(5, 15, 0)
+	prctl(PR_SET_PTRACER, process.processId(), 0, 0, 0);
+#else
 	prctl(PR_SET_PTRACER, process.pid(), 0, 0, 0);
+#endif
 #endif
 	process.waitForFinished(-1);
 	std::raise(signal);
@@ -132,6 +141,7 @@ MainWindow::MainWindow(const QStringList& files)
 	m_config = new Config(this);
 	m_acquirer = new Acquirer(ui);
 	m_displayer = new Displayer(ui);
+	m_recognitionMenu = new RecognitionMenu(this);
 	m_recognizer = new Recognizer(ui);
 	m_sourceManager = new SourceManager(ui);
 
@@ -153,25 +163,30 @@ MainWindow::MainWindow(const QStringList& files)
 	m_idleWidgets.append(ui.spinBoxResolution);
 	m_idleWidgets.append(ui.toolButtonRecognize);
 
-	connect(ui.actionRedetectLanguages, SIGNAL(triggered()), m_recognizer, SLOT(updateLanguagesMenu()));
-	connect(ui.actionManageLanguages, SIGNAL(triggered()), this, SLOT(manageLanguages()));
-	connect(ui.actionPreferences, SIGNAL(triggered()), this, SLOT(showConfig()));
-	connect(ui.actionHelp, SIGNAL(triggered()), this, SLOT(showHelp()));
-	connect(ui.actionAbout, SIGNAL(triggered()), this, SLOT(showAbout()));
-	connect(ui.actionImageControls, SIGNAL(toggled(bool)), ui.widgetImageControls, SLOT(setVisible(bool)));
-	connect(m_acquirer, SIGNAL(scanPageAvailable(QString)), m_sourceManager, SLOT(addSource(QString)));
-	connect(m_sourceManager, SIGNAL(sourceChanged()), this, SLOT(onSourceChanged()));
-	connect(ui.actionToggleOutputPane, SIGNAL(toggled(bool)), ui.dockWidgetOutput, SLOT(setVisible(bool)));
-	connect(ui.comboBoxOCRMode, SIGNAL(currentIndexChanged(int)), this, SLOT(setOCRMode(int)));
-	connect(m_recognizer, SIGNAL(languageChanged(Config::Lang)), this, SLOT(languageChanged(Config::Lang)));
-	connect(ui.actionAutodetectLayout, SIGNAL(triggered()), m_displayer, SLOT(autodetectOCRAreas()));
+	ui.comboBoxOCRMode->addItem(_("Plain text"), OutputModeText);
+	ui.comboBoxOCRMode->addItem(_("hOCR, PDF"), OutputModeHOCR);
+	ui.comboBoxOCRMode->setCurrentIndex(-1);
+
+	connect(ui.actionRedetectLanguages, &QAction::triggered, m_recognitionMenu, &RecognitionMenu::rebuild);
+	connect(ui.actionManageLanguages, &QAction::triggered, this, &MainWindow::manageLanguages);
+	connect(ui.actionPreferences, &QAction::triggered, this, &MainWindow::showConfig);
+	connect(ui.actionHelp, &QAction::triggered, this, [this] { showHelp(); });
+	connect(ui.actionAbout, &QAction::triggered, this, &MainWindow::showAbout);
+	connect(ui.actionImageControls, &QAction::toggled, ui.widgetImageControls, &QWidget::setVisible);
+	connect(m_acquirer, &Acquirer::scanPageAvailable, m_sourceManager, [this](const QString & source) { m_sourceManager->addSource(source); });
+	connect(m_sourceManager, &SourceManager::sourceChanged, this, &MainWindow::onSourceChanged);
+	connect(ui.actionToggleOutputPane, &QAction::toggled, ui.dockWidgetOutput, &QDockWidget::setVisible);
+	connect(ui.comboBoxOCRMode, qOverload<int>(&QComboBox::currentIndexChanged), this, [this] { setOutputMode(static_cast<OutputMode>(ui.comboBoxOCRMode->currentData().toInt())); });
+	connect(m_recognitionMenu, &RecognitionMenu::languageChanged, this, &MainWindow::languageChanged);
+	connect(ui.actionAutodetectLayout, &QAction::triggered, m_displayer, &Displayer::autodetectOCRAreas);
+	connect(ui.actionBatchExport, &QAction::triggered, this, &MainWindow::batchExport);
 
 	ADD_SETTING(VarSetting<QByteArray>("wingeom"));
 	ADD_SETTING(VarSetting<QByteArray>("winstate"));
 	ADD_SETTING(ActionSetting("showcontrols", ui.actionImageControls));
-	ADD_SETTING(ComboSetting("outputeditor", ui.comboBoxOCRMode, 0));
+	ADD_SETTING(ComboSetting("outputeditor", ui.comboBoxOCRMode, OutputModeText));
 
-	m_recognizer->updateLanguagesMenu();
+	m_recognitionMenu->rebuild();
 
 	m_progressWidget = new QWidget(this);
 	m_progressWidget->setLayout(new QHBoxLayout());
@@ -182,11 +197,11 @@ MainWindow::MainWindow(const QStringList& files)
 	m_progressBar->setRange(0, 100);
 	m_progressBar->setMaximumWidth(100);
 	m_progressTimer.setSingleShot(false);
-	connect(&m_progressTimer, SIGNAL(timeout()), this, SLOT(progressUpdate()));
+	connect(&m_progressTimer, &QTimer::timeout, this, &MainWindow::progressUpdate);
 	m_progressWidget->layout()->addWidget(m_progressBar);
 	m_progressCancelButton = new QToolButton();
 	m_progressCancelButton->setIcon(QIcon::fromTheme("dialog-close"));
-	connect(m_progressCancelButton, SIGNAL(clicked(bool)), this, SLOT(progressCancel()));
+	connect(m_progressCancelButton, &QToolButton::clicked, this, &MainWindow::progressCancel);
 	m_progressWidget->layout()->addWidget(m_progressCancelButton);
 	statusBar()->addPermanentWidget(m_progressWidget);
 	m_progressWidget->setVisible(false);
@@ -201,20 +216,38 @@ MainWindow::MainWindow(const QStringList& files)
 
 #if ENABLE_VERSIONCHECK
 	if(ConfigSettings::get<SwitchSetting>("updatecheck")->getValue()) {
-		connect(&m_versionCheckThread, SIGNAL(finished()), this, SLOT(checkVersion()));
-		m_versionCheckThread.start();
+		QFuture<QString> future = QtConcurrent::run([ = ] {
+			QString messages;
+			QString newver = Utils::download(QUrl(CHECKURL), messages, 5000);
+			newver.replace(QRegularExpression("\\s+"), "");
+			QRegularExpression pat(R"(^[\d+\.]+\d+$)");
+			return pat.match(newver).hasMatch() ? newver : QString();
+		});
+		connect(&m_versionWatcher, &QFutureWatcher<QString>::finished, this, [ = ] { checkVersion(m_versionWatcher.future().result()); });
+		m_versionWatcher.setFuture(future);
 	}
 #endif
 
-	m_sourceManager->addSources(files);
+	QStringList hocrFiles;
+	QStringList otherFiles;
+	for(const QString& file : files) {
+		if(file.endsWith(".html", Qt::CaseInsensitive)) {
+			hocrFiles.append(file);
+		} else {
+			otherFiles.append(file);
+		}
+	}
+	m_sourceManager->addSources(otherFiles);
+	if(!hocrFiles.isEmpty()) {
+		if(setOutputMode(OutputModeHOCR)) {
+			static_cast<OutputEditorHOCR*>(m_outputEditor)->open(OutputEditorHOCR::InsertMode::Append, hocrFiles);
+			ui.dockWidgetOutput->setVisible(true);
+		}
+	}
 }
 
 MainWindow::~MainWindow() {
-#if ENABLE_VERSIONCHECK
-	while(m_versionCheckThread.isRunning()) {
-		QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
-	}
-#endif
+	m_versionWatcher.waitForFinished();
 	delete m_acquirer;
 	delete m_outputEditor;
 	delete m_sourceManager;
@@ -230,6 +263,18 @@ void MainWindow::openFiles(const QStringList& files) {
 	m_sourceManager->addSources(files);
 }
 
+void MainWindow::openOutput(const QString& filename) {
+	if(filename.endsWith(".txt", Qt::CaseInsensitive)) {
+		if(setOutputMode(OutputModeText)) {
+			m_outputEditor->open(filename);
+		}
+	} else if(filename.endsWith(".html", Qt::CaseInsensitive)) {
+		if(setOutputMode(OutputModeHOCR)) {
+			m_outputEditor->open(filename);
+		}
+	}
+}
+
 void MainWindow::setOutputPaneVisible(bool visible) {
 	ui.actionToggleOutputPane->setChecked(visible);
 }
@@ -266,16 +311,18 @@ void MainWindow::closeEvent(QCloseEvent* ev) {
 		ev->ignore();
 	} else if(!m_outputEditor->clear()) {
 		ev->ignore();
-	} else if(!isMaximized()) {
-		ConfigSettings::get<VarSetting<QByteArray>>("wingeom")->setValue(saveGeometry());
+	} else {
+		if(!isMaximized()) {
+			ConfigSettings::get<VarSetting<QByteArray>>("wingeom")->setValue(saveGeometry());
+		}
 		ConfigSettings::get<VarSetting<QByteArray>>("winstate")->setValue(saveState());
 	}
 }
 
 void MainWindow::onSourceChanged() {
 	QList<Source*> sources = m_sourceManager->getSelectedSources();
-	if(m_displayer->setSources(sources)) {
-		setWindowTitle(QString("%1 - %2").arg(sources.size() == 1 ? sources.front()->displayname : _("Multiple sources")).arg(PACKAGE_NAME));
+	if(m_displayer->setSources(sources) && !sources.empty()) {
+		setWindowTitle(QString("%1 - %2").arg(sources.size() == 1 ? sources.front()->displayname : _("Multiple sources (%1)").arg(sources.size())).arg(PACKAGE_NAME));
 		if(m_stateStack.top().first == State::Idle) {
 			pushState(State::Normal, _("Ready"));
 		}
@@ -326,33 +373,38 @@ void MainWindow::manageLanguages() {
 
 void MainWindow::showConfig() {
 	m_config->showDialog();
-	m_recognizer->updateLanguagesMenu();
+	m_recognitionMenu->rebuild();
 }
 
-void MainWindow::setOCRMode(int idx) {
+bool MainWindow::setOutputMode(OutputMode mode) {
 	if(m_outputEditor && !m_outputEditor->clear()) {
 		ui.comboBoxOCRMode->blockSignals(true);
 		if(dynamic_cast<OutputEditorText*>(m_outputEditor)) {
-			ui.comboBoxOCRMode->setCurrentIndex(0);
+			ui.comboBoxOCRMode->setCurrentIndex(ui.comboBoxOCRMode->findData(OutputModeText));
 		} else { /*if(dynamic_cast<OutputEditorHOCR*>(m_outputEditor))*/
-			ui.comboBoxOCRMode->setCurrentIndex(1);
+			ui.comboBoxOCRMode->setCurrentIndex(ui.comboBoxOCRMode->findData(OutputModeHOCR));
 		}
 		ui.comboBoxOCRMode->blockSignals(false);
+		return false;
 	} else {
 		delete m_displayerTool;
 		delete m_outputEditor;
-		if(idx == 0) {
+		if(mode == OutputModeText) {
 			m_displayerTool = new DisplayerToolSelect(m_displayer);
 			m_outputEditor = new OutputEditorText();
-		} else { /*if(idx == 1)*/
+		} else { /*if(mode == OutputModeHOCR)*/
 			m_displayerTool = new DisplayerToolHOCR(m_displayer);
 			m_outputEditor = new OutputEditorHOCR(static_cast<DisplayerToolHOCR*>(m_displayerTool));
 		}
 		ui.actionAutodetectLayout->setVisible(m_displayerTool->allowAutodetectOCRAreas());
 		m_displayer->setTool(m_displayerTool);
-		m_outputEditor->setLanguage(m_recognizer->getSelectedLanguage());
-		connect(ui.actionToggleOutputPane, SIGNAL(toggled(bool)), m_outputEditor, SLOT(onVisibilityChanged(bool)));
+		m_outputEditor->setLanguage(m_recognitionMenu->getRecognitionLanguage());
+		connect(ui.actionToggleOutputPane, &QAction::toggled, m_outputEditor, &OutputEditor::onVisibilityChanged);
 		ui.dockWidgetOutput->setWidget(m_outputEditor->getUI());
+		ui.comboBoxOCRMode->blockSignals(true);
+		ui.comboBoxOCRMode->setCurrentIndex(ui.comboBoxOCRMode->findData(mode));
+		ui.comboBoxOCRMode->blockSignals(false);
+		return true;
 	}
 }
 
@@ -368,11 +420,11 @@ void MainWindow::addNotification(const QString& title, const QString& message, c
 	for(const NotificationAction& action : actions) {
 		QToolButton* btn = new QToolButton(frame);
 		btn->setText(action.text);
-		connect(btn, SIGNAL(clicked()), action.target, action.slot);
+		connect(btn, &QToolButton::clicked, action.slot);
 		if(action.close) {
 			btn->setProperty("handle", QVariant::fromValue(reinterpret_cast<void*>(handle)));
 			btn->setProperty("frame", QVariant::fromValue(reinterpret_cast<void*>(frame)));
-			connect(btn, SIGNAL(clicked()), this, SLOT(hideNotification()));
+			connect(btn, &QToolButton::clicked, this, [this] { hideNotification(); });
 		}
 		layout->addWidget(btn);
 	}
@@ -380,7 +432,7 @@ void MainWindow::addNotification(const QString& title, const QString& message, c
 	closeBtn->setIcon(QIcon::fromTheme("dialog-close"));
 	closeBtn->setProperty("handle", QVariant::fromValue(reinterpret_cast<void*>(handle)));
 	closeBtn->setProperty("frame", QVariant::fromValue(reinterpret_cast<void*>(frame)));
-	connect(closeBtn, SIGNAL(clicked()), this, SLOT(hideNotification()));
+	connect(closeBtn, &QToolButton::clicked, this, [this] { hideNotification(); });
 	layout->addWidget(closeBtn);
 	ui.centralwidget->layout()->addWidget(frame);
 	if(handle) {
@@ -401,18 +453,7 @@ void MainWindow::hideNotification(Notification handle) {
 	}
 }
 
-void MainWindow::VersionCheckThread::run() {
-	QString messages;
-	QString newver = Utils::download(QUrl(CHECKURL), messages, 5000);
-	newver.replace(QRegExp("\\s+"), "");
-	QRegExp pat(R"(^[\d+\.]+\d+$)");
-	if(pat.exactMatch(newver)) {
-		m_newestVersion = newver;
-	}
-}
-
-void MainWindow::checkVersion() {
-	QString newver = m_versionCheckThread.getNewestVersion();
+void MainWindow::checkVersion(const QString& newver) {
 	qDebug("Newest version is: %s", qPrintable(newver));
 	if(newver.isEmpty()) {
 		return;
@@ -421,9 +462,9 @@ void MainWindow::checkVersion() {
 
 	if(newver.compare(curver) > 0) {
 		addNotification(_("New version"), _("gImageReader %1 is available").arg(newver), {
-			{_("Download"), this, SLOT(openDownloadUrl()), false},
-			{_("Changelog"), this, SLOT(openChangeLogUrl()), false},
-			{_("Don't notify again"), m_config, SLOT(disableUpdateCheck()), true}
+			{_("Download"), [this]{ openDownloadUrl(); }, false},
+			{_("Changelog"), [this]{ openChangeLogUrl(); }, false},
+			{_("Don't notify again"), [this]{ m_config->disableUpdateCheck(); }, true}
 		});
 	}
 }
@@ -471,25 +512,25 @@ void MainWindow::languageChanged(const Config::Lang& lang) {
 	m_notifierHandle = nullptr;
 	const QString& code = lang.code;
 	if(!code.isEmpty() && !QtSpell::checkLanguageInstalled(code) && ConfigSettings::get<SwitchSetting>("dictinstall")->getValue()) {
-		NotificationAction actionDontShowAgain = {_("Don't show again"), m_config, SLOT(disableDictInstall()), true};
-		NotificationAction actionInstall = {_("Install"), this, SLOT(dictionaryAutoinstall()), false};
+		NotificationAction actionDontShowAgain = {_("Don't show again"), [this]{ m_config->disableDictInstall(); }, true};
+		NotificationAction actionInstall = {_("Install"), [this]{ dictionaryAutoinstall(); }, false};
 #ifdef Q_OS_LINUX
 		if(getConfig()->useSystemDataLocations()) {
 			QDBusConnectionInterface* iface = QDBusConnection::sessionBus().interface();
 			iface->startService("org.freedesktop.PackageKit");
 			if(!iface->isServiceRegistered("org.freedesktop.PackageKit").value()) {
-				actionInstall = {_("Help"), this, SLOT(showHelp()), false}; // TODO #InstallSpelling
+				actionInstall = {_("Help"), [this]{ showHelp(); }, false}; // TODO #InstallSpelling
 				qWarning("Could not find PackageKit on DBus, dictionary autoinstallation will not work");
 			}
 		}
 #endif
-		const QString& name = m_recognizer->getSelectedLanguage().name;
+		const QString& name = m_recognitionMenu->getRecognitionLanguage().name;
 		addNotification(_("Spelling dictionary missing"), _("The spellcheck dictionary for %1 is not installed").arg(name), {actionInstall, actionDontShowAgain}, &m_notifierHandle);
 	}
 }
 
 void MainWindow::dictionaryAutoinstall() {
-	QString code = m_recognizer->getSelectedLanguage().code;
+	QString code = m_recognitionMenu->getRecognitionLanguage().code;
 	pushState(State::Busy, _("Installing spelling dictionary for '%1'").arg(code));
 #ifdef Q_OS_WIN
 	bool isWindows = true;
@@ -513,7 +554,7 @@ void MainWindow::dictionaryAutoinstall() {
 				showHelp("#InstallSpelling");
 			}
 		}
-		m_recognizer->updateLanguagesMenu();
+		m_recognitionMenu->rebuild();
 		popState();
 #endif
 	} else {
@@ -536,16 +577,17 @@ void MainWindow::dictionaryAutoinstall() {
 			}
 			return;
 		}
-		int pos = 0;
 		QString langCode = code.left(code.indexOf('_'));
-		QRegExp langPat(QString(">(%1_?[A-Z]*)<").arg(langCode));
-		QRegExp dictPat(QString(">(%1_?[\\w_]*\\.(dic|aff))<").arg(langCode));
+		QRegularExpression langPat(QString(">(%1_?[A-Z]*)<").arg(langCode));
+		QRegularExpression dictPat(QString(">(%1_?[\\w_]*\\.(dic|aff))<").arg(langCode));
 		QStringList downloaded;
 		QStringList failed;
+		int pos = 0;
+		QRegularExpressionMatch match;
 
-		while((pos = langPat.indexIn(html, pos)) != -1) {
-			QString lang = langPat.cap(1);
-			pos += langPat.matchedLength();
+		while((match = langPat.match(html, pos)).hasMatch()) {
+			QString lang = match.captured(1);
+			pos = match.capturedEnd();
 
 			QString dictHtml = QString(Utils::download(url + lang + "/", messages));
 			if(dictHtml.isEmpty()) {
@@ -553,8 +595,9 @@ void MainWindow::dictionaryAutoinstall() {
 			}
 
 			int dictPos = 0;
-			while((dictPos = dictHtml.indexOf(dictPat, dictPos)) != -1) {
-				QString filename = dictPat.cap(1);
+			QRegularExpressionMatch dictMatch;
+			while((dictMatch = dictPat.match(dictHtml, dictPos)).hasMatch()) {
+				QString filename = dictMatch.captured(1);
 				pushState(State::Busy, _("Downloading '%1'...").arg(filename));
 				QByteArray data = Utils::download(plainurl + lang + "/" + filename, messages);
 				if(!data.isNull()) {
@@ -569,7 +612,7 @@ void MainWindow::dictionaryAutoinstall() {
 					failed.append(filename);
 				}
 				popState();
-				dictPos += dictPat.matchedLength();
+				dictPos += dictMatch.capturedEnd();
 			}
 		}
 
@@ -577,7 +620,7 @@ void MainWindow::dictionaryAutoinstall() {
 		if(!failed.isEmpty()) {
 			QMessageBox::critical(this, _("Error"), _("The following dictionaries could not be downloaded:\n%1\n\nCheck the connectivity and directory permissions.\nHint: If you don't have write permissions in system folders, you can switch to user paths in the settings dialog.").arg(failed.join("\n")));
 		} else if(!downloaded.isEmpty()) {
-			m_recognizer->updateLanguagesMenu();
+			m_recognitionMenu->rebuild();
 			QMessageBox::information(this, _("Dictionaries installed"), _("The following dictionary files were installed:\n%1").arg(downloaded.join("\n")));
 		} else {
 			if(QMessageBox::Help == QMessageBox::critical(this, _("Error"), _("No spelling dictionaries found for '%1'.").arg(code), QMessageBox::Ok | QMessageBox::Help, QMessageBox::Ok)) {
@@ -586,3 +629,8 @@ void MainWindow::dictionaryAutoinstall() {
 		}
 	}
 }
+
+void MainWindow::batchExport() {
+	HOCRBatchExportDialog dialog;
+	dialog.exec();
+}
