@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * OutputTextEdit.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -22,6 +22,7 @@
 #include <QApplication>
 #include <QEventLoop>
 #include <QPainter>
+#include <QRegularExpression>
 #include <QSyntaxHighlighter>
 
 
@@ -34,12 +35,12 @@ private:
 		QTextCharFormat fmt;
 		fmt.setForeground(Qt::gray);
 
-		QRegExp expression("\\s");
-		int index = text.indexOf(expression);
-		while (index >= 0) {
-			int length = expression.matchedLength();
-			setFormat(index, length, fmt);
-			index = text.indexOf(expression, index + length);
+		QRegularExpression expression("\\s");
+		QRegularExpressionMatch match;
+		int offset = 0;
+		while ((match = expression.match(text, offset)).hasMatch()) {
+			offset = match.capturedEnd();
+			setFormat(match.capturedStart(), match.capturedLength(), fmt);
 		}
 	}
 };
@@ -54,8 +55,8 @@ OutputTextEdit::OutputTextEdit(QWidget* parent)
 	m_regionCursor.movePosition(QTextCursor::End, QTextCursor::KeepAnchor);
 	m_entireRegion = true;
 
-	connect(this, SIGNAL(cursorPositionChanged()), this, SLOT(saveRegionBounds()));
-	connect(this, SIGNAL(selectionChanged()), this, SLOT(saveRegionBounds()));
+	connect(this, &OutputTextEdit::cursorPositionChanged, this, &OutputTextEdit::saveRegionBounds);
+	connect(this, &OutputTextEdit::selectionChanged, this, &OutputTextEdit::saveRegionBounds);
 
 	// Force inactive selection to have same color as active selection
 	QColor highlightColor = palette().color(QPalette::Highlight);
@@ -88,7 +89,7 @@ bool OutputTextEdit::findReplace(bool backwards, bool replace, bool matchCase, c
 		return false;
 	}
 
-	QTextDocument::FindFlags flags = 0;
+	QTextDocument::FindFlags flags;
 	Qt::CaseSensitivity cs = Qt::CaseInsensitive;
 	if(backwards) {
 		flags |= QTextDocument::FindBackward;
@@ -133,7 +134,7 @@ bool OutputTextEdit::findReplace(bool backwards, bool replace, bool matchCase, c
 	return true;
 }
 
-bool OutputTextEdit::replaceAll(const QString& searchstr, const QString& replacestr, bool matchCase) {
+int OutputTextEdit::replaceAll(const QString& searchstr, const QString& replacestr, bool matchCase) {
 	QTextCursor cursor =  regionBounds();
 	int end = cursor.position();
 	QString cursel = cursor.selectedText();
@@ -147,7 +148,7 @@ bool OutputTextEdit::replaceAll(const QString& searchstr, const QString& replace
 	} else {
 		cursor.setPosition(std::min(cursor.anchor(), cursor.position()));
 	}
-	QTextDocument::FindFlags flags = 0;
+	QTextDocument::FindFlags flags;
 	if(matchCase) {
 		flags = QTextDocument::FindCaseSensitively;
 	}
@@ -163,10 +164,7 @@ bool OutputTextEdit::replaceAll(const QString& searchstr, const QString& replace
 		++count;
 		QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
 	}
-	if(count == 0) {
-		return false;
-	}
-	return true;
+	return count;
 }
 
 void OutputTextEdit::setDrawWhitespace(bool drawWhitespace) {
@@ -175,7 +173,7 @@ void OutputTextEdit::setDrawWhitespace(bool drawWhitespace) {
 	if(drawWhitespace) {
 		textOption.setFlags(QTextOption::ShowTabsAndSpaces | QTextOption::AddSpaceForLineAndParagraphSeparators);
 	} else {
-		textOption.setFlags(0);
+		textOption.setFlags(QTextOption::Flags());
 	}
 	document()->setDefaultTextOption(textOption);
 }
@@ -283,7 +281,7 @@ void OutputTextEdit::saveRegionBounds() {
 			viewport()->repaint();
 		}
 		// If only one word is selected, don't treat it as a region
-		if(!m_regionCursor.selectedText().contains(QRegExp("\\s"))) {
+		if(!m_regionCursor.selectedText().contains(QRegularExpression("\\s"))) {
 			m_regionCursor.clearSelection();
 		}
 	}
