@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * SubstitutionsManager.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -63,11 +63,7 @@ SubstitutionsManager::SubstitutionsManager(QWidget* parent)
 	m_tableWidget = new QTableWidget(0, 2, this);
 	m_tableWidget->setSelectionBehavior(QAbstractItemView::SelectRows);
 	m_tableWidget->setEditTriggers(QAbstractItemView::CurrentChanged);
-#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
-	m_tableWidget->horizontalHeader()->setResizeMode(QHeaderView::Stretch);
-#else
 	m_tableWidget->horizontalHeader()->setSectionResizeMode(QHeaderView::Stretch);
-#endif
 	m_tableWidget->horizontalHeader()->setVisible(true);
 	m_tableWidget->verticalHeader()->setVisible(false);
 	m_tableWidget->setHorizontalHeaderLabels(QStringList() << _("Search for") << _("Replace with"));
@@ -75,21 +71,21 @@ SubstitutionsManager::SubstitutionsManager(QWidget* parent)
 	QDialogButtonBox* buttonBox = new QDialogButtonBox(QDialogButtonBox::Apply | QDialogButtonBox::Close);
 
 	QVBoxLayout* layout = new QVBoxLayout(this);
-	layout->setMargin(4);
+	layout->setContentsMargins(4, 4, 4, 4);
 	layout->addWidget(toolbar);
 	layout->addWidget(m_tableWidget);
 	layout->addWidget(buttonBox);
 
 	setLayout(layout);
 
-	connect(openAction, SIGNAL(triggered()), this, SLOT(openList()));
-	connect(saveAction, SIGNAL(triggered()), this, SLOT(saveList()));
-	connect(clearAction, SIGNAL(triggered()), this, SLOT(clearList()));
-	connect(addAction, SIGNAL(triggered()), this, SLOT(addRow()));
-	connect(buttonBox->button(QDialogButtonBox::Apply), SIGNAL(clicked()), this, SLOT(emitApplySubstitutions()));
-	connect(buttonBox->button(QDialogButtonBox::Close), SIGNAL(clicked()), this, SLOT(hide()));
-	connect(m_removeAction, SIGNAL(triggered()), this, SLOT(removeRows()));
-	connect(m_tableWidget->selectionModel(), SIGNAL(selectionChanged(QItemSelection, QItemSelection)), this, SLOT(onTableSelectionChanged(QItemSelection, QItemSelection)));
+	connect(openAction, &QAction::triggered, this, &SubstitutionsManager::openList);
+	connect(saveAction, &QAction::triggered, this, &SubstitutionsManager::saveList);
+	connect(clearAction, &QAction::triggered, this, &SubstitutionsManager::clearList);
+	connect(addAction, &QAction::triggered, this, &SubstitutionsManager::addRow);
+	connect(buttonBox->button(QDialogButtonBox::Apply), &QPushButton::clicked, this, &SubstitutionsManager::emitApplySubstitutions);
+	connect(buttonBox->button(QDialogButtonBox::Close), &QPushButton::clicked, this, &SubstitutionsManager::hide);
+	connect(m_removeAction, &QAction::triggered, this, &SubstitutionsManager::removeRows);
+	connect(m_tableWidget->selectionModel(), &QItemSelectionModel::selectionChanged, this, &SubstitutionsManager::onTableSelectionChanged);
 
 	ADD_SETTING(TableSetting("substitutionslist", m_tableWidget));
 }
@@ -159,7 +155,7 @@ bool SubstitutionsManager::saveList() {
 
 bool SubstitutionsManager::clearList() {
 	if(m_tableWidget->rowCount() > 0) {
-		int response = QMessageBox::question(this, _("Save List?"), _("Do you want to save the current list?"), QMessageBox::Save, QMessageBox::Discard, QMessageBox::Cancel);
+		int response = QMessageBox::question(this, _("Save List?"), _("Do you want to save the current list?"), QMessageBox::Save | QMessageBox::Discard | QMessageBox::Cancel);
 		if(response == QMessageBox::Save) {
 			if(!saveList()) {
 				return false;
