@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * HOCRDocument.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -20,19 +20,20 @@
 #include <QDir>
 #include <QDomElement>
 #include <QFileInfo>
+#include <QMenu>
 #include <QIcon>
 #include <QSet>
-#include <QTextStream>
-#include <QtSpell.hpp>
 #include <cmath>
 
 #include "common.hh"
 #include "HOCRDocument.hh"
+#include "HOCRSpellChecker.hh"
 #include "Utils.hh"
 
 
-HOCRDocument::HOCRDocument(QtSpell::TextEditChecker* spell, QObject* parent)
-	: QAbstractItemModel(parent), m_spell(spell) {
+HOCRDocument::HOCRDocument(QObject* parent)
+	: QAbstractItemModel(parent) {
+	m_spell = new HOCRSpellChecker(this);
 }
 
 HOCRDocument::~HOCRDocument() {
@@ -47,8 +48,49 @@ void HOCRDocument::clear() {
 	endResetModel();
 }
 
-void HOCRDocument::recheckSpelling() {
-	recursiveDataChanged(QModelIndex(), {Qt::DisplayRole}, {"ocrx_word"});
+void HOCRDocument::resetMisspelled(const QModelIndex& index) {
+	if(hasChildren(index)) {
+		for(int i = 0, n = rowCount(index); i < n; ++i) {
+			resetMisspelled(this->index(i, 0, index));
+		}
+	} else {
+		mutableItemAtIndex(index)->setMisspelled(-1);
+		emit dataChanged(index, index, {Qt::ForegroundRole});
+	}
+}
+
+void HOCRDocument::addSpellingActions(QMenu* menu, const QModelIndex& index) {
+	QStringList suggestions;
+	QString trimmedWord;
+	bool valid = getItemSpellingSuggestions(index, trimmedWord, suggestions, 16);
+	for(const QString& suggestion : suggestions) {
+		menu->addAction(suggestion, menu, [this, suggestion, index] { setData(index, suggestion, Qt::EditRole); });
+	}
+	if(!trimmedWord.isEmpty()) {
+		if(suggestions.isEmpty()) {
+			menu->addAction(_("No suggestions"))->setEnabled(false);
+		}
+		if(!valid) {
+			menu->addSeparator();
+			menu->addAction(_("Add to dictionary"), menu, [this, trimmedWord, index] {
+				m_spell->addWordToDictionary(trimmedWord);
+				resetMisspelled(index);
+			});
+			menu->addAction(_("Ignore word"), menu, [this, trimmedWord, index] {
+				m_spell->ignoreWord(trimmedWord);
+				resetMisspelled(index);
+			});
+		}
+	}
+}
+
+void HOCRDocument::addWordToDictionary(const QModelIndex& index) {
+	const HOCRItem* item = itemAtIndex(index);
+	if(item && item->isMisspelled()) {
+		QString trimmedWord = HOCRItem::trimmedWord(item->text());
+		m_spell->addWordToDictionary(trimmedWord);
+		resetMisspelled(index);
+	}
 }
 
 QString HOCRDocument::toHTML() const {
@@ -60,13 +102,33 @@ QString HOCRDocument::toHTML() const {
 	return html;
 }
 
-QModelIndex HOCRDocument::addPage(const QDomElement& pageElement, bool cleanGraphics) {
-	int newRow = m_pages.size();
-	beginInsertRows(QModelIndex(), newRow, newRow);
-	m_pages.append(new HOCRPage(pageElement, ++m_pageIdCounter, m_defaultLanguage, cleanGraphics, m_pages.size()));
+
+QModelIndex HOCRDocument::insertPage(int beforeIdx, const QDomElement& pageElement, bool cleanGraphics, const QString& sourceBasePath) {
+	beginInsertRows(QModelIndex(), beforeIdx, beforeIdx);
+	m_pages.insert(beforeIdx, new HOCRPage(pageElement, ++m_pageIdCounter, m_defaultLanguage, cleanGraphics, beforeIdx));
+	if(!sourceBasePath.isEmpty()) {
+		m_pages[beforeIdx]->convertSourcePath(sourceBasePath, true);
+	}
+	for(int i = beforeIdx + 1; i < m_pages.size(); ++i) {
+		m_pages[i]->m_index = i;
+	}
 	endInsertRows();
 	emit dataChanged(index(0, 0), index(m_pages.size() - 1, 0), {Qt::DisplayRole});
-	return index(newRow, 0);
+	return index(beforeIdx, 0);
+}
+
+QModelIndex HOCRDocument::indexAtItem(const HOCRItem* item) const {
+	QList<HOCRItem*> parents;
+	HOCRItem* parent = item->parent();
+	while(parent) {
+		parents.prepend(parent);
+		parent = parent->parent();
+	}
+	QModelIndex idx;
+	for(HOCRItem* parent : parents) {
+		idx = index(parent->index(), 0, idx);
+	}
+	return index(item->index(), 0, idx);
 }
 
 bool HOCRDocument::editItemAttribute(const QModelIndex& index, const QString& name, const QString& value, const QString& attrItemClass) {
@@ -81,7 +143,7 @@ bool HOCRDocument::editItemAttribute(const QModelIndex& index, const QString& na
 		emit dataChanged(colIdx, colIdx, {Qt::DisplayRole});
 	}
 	if(name == "lang") {
-		recursiveDataChanged(index, {Qt::DisplayRole}, {"ocrx_word"});
+		resetMisspelled(index);
 	}
 	emit itemAttributeChanged(index, name, value);
 	if(name == "title:bbox") {
@@ -131,7 +193,7 @@ QModelIndex HOCRDocument::mergeItems(const QModelIndex& parent, int startRow, in
 	if(endRow - startRow <= 0) {
 		return QModelIndex();
 	}
-	QModelIndex targetIndex = parent.child(startRow, 0);
+	QModelIndex targetIndex = index(startRow, 0, parent);
 	HOCRItem* targetItem = mutableItemAtIndex(targetIndex);
 	if(!targetItem || targetItem->itemClass() == "ocr_page") {
 		return QModelIndex();
@@ -143,7 +205,7 @@ QModelIndex HOCRDocument::mergeItems(const QModelIndex& parent, int startRow, in
 		QString text = targetItem->text();
 		beginRemoveRows(parent, startRow + 1, endRow);
 		for(int row = ++startRow; row <= endRow; ++row) {
-			HOCRItem* item = mutableItemAtIndex(parent.child(startRow, 0));
+			HOCRItem* item = mutableItemAtIndex(index(startRow, 0, parent));
 			Q_ASSERT(item);
 			text += item->text();
 			bbox = bbox.united(item->bbox());
@@ -151,13 +213,15 @@ QModelIndex HOCRDocument::mergeItems(const QModelIndex& parent, int startRow, in
 		}
 		endRemoveRows();
 		targetItem->setText(text);
-		emit dataChanged(targetIndex, targetIndex, {Qt::DisplayRole, Qt::ForegroundRole});
+		for(QModelIndex changedIndex : recheckItemSpelling(targetIndex)) {
+			emit dataChanged(changedIndex, changedIndex, {Qt::DisplayRole, Qt::ForegroundRole});
+		}
 	} else {
 		// Merge other items: merge dom trees and bounding boxes
 		QVector<HOCRItem*> moveChilds;
 		beginRemoveRows(parent, startRow + 1, endRow);
 		for(int row = ++startRow; row <= endRow; ++row) {
-			HOCRItem* item = mutableItemAtIndex(parent.child(startRow, 0));
+			HOCRItem* item = mutableItemAtIndex(index(startRow, 0, parent));
 			Q_ASSERT(item);
 			moveChilds.append(item->takeChildren());
 			bbox = bbox.united(item->bbox());
@@ -177,11 +241,11 @@ QModelIndex HOCRDocument::mergeItems(const QModelIndex& parent, int startRow, in
 	return targetIndex;
 }
 
-QModelIndex HOCRDocument::splitItem(const QModelIndex& index, int startRow, int endRow) {
+QModelIndex HOCRDocument::splitItem(const QModelIndex& itemIndex, int startRow, int endRow) {
 	if(endRow - startRow < 0) {
 		return QModelIndex();
 	}
-	HOCRItem* item = mutableItemAtIndex(index);
+	HOCRItem* item = mutableItemAtIndex(itemIndex);
 	if(!item) {
 		return QModelIndex();
 	}
@@ -200,12 +264,12 @@ QModelIndex HOCRDocument::splitItem(const QModelIndex& index, int startRow, int
 	newElement.setAttribute("class", itemClass);
 	newElement.setAttribute("title", HOCRItem::serializeAttrGroup(item->getTitleAttributes()));
 	HOCRItem* newItem = new HOCRItem(newElement, item->page(), item->parent(), item->index() + 1);
-	beginInsertRows(index.parent(), item->index() + 1, item->index() + 1);
+	beginInsertRows(itemIndex.parent(), item->index() + 1, item->index() + 1);
 	insertItem(item->parent(), newItem, item->index() + 1);
 	endInsertRows();
-	QModelIndex newIndex = index.sibling(index.row() + 1, 0);
+	QModelIndex newIndex = itemIndex.sibling(itemIndex.row() + 1, 0);
 	for(int row = 0; row <= (endRow - startRow); ++row) {
-		QModelIndex childIndex = index.child(startRow, 0);
+		QModelIndex childIndex = index(startRow, 0, itemIndex);
 		HOCRItem* child = mutableItemAtIndex(childIndex);
 		Q_ASSERT(child);
 		moveItem(childIndex, newIndex, row);
@@ -214,6 +278,61 @@ QModelIndex HOCRDocument::splitItem(const QModelIndex& index, int startRow, int
 	return newIndex;
 }
 
+QModelIndex HOCRDocument::splitItemText(const QModelIndex& itemIndex, int pos) {
+	HOCRItem* item = mutableItemAtIndex(itemIndex);
+	if(!item || item->itemClass() != "ocrx_word") {
+		return QModelIndex();
+	}
+	// Don't split if it would create empty words
+	if(pos == 0 || pos == item->text().length()) {
+		return itemIndex;
+	}
+	// Compute new bounding box using font metrics with default font
+	QFontMetrics metrics((QFont()));
+	double fraction = metrics.horizontalAdvance(item->text().left(pos)) / double(metrics.horizontalAdvance(item->text()));
+	QRect bbox = item->bbox();
+	QRect leftBBox = bbox;
+	leftBBox.setWidth(bbox.width() * fraction);
+	QRect rightBBox = bbox;
+	rightBBox.setLeft(bbox.left() + bbox.width() * fraction);
+
+	QDomDocument doc;
+	QDomElement newElement = doc.createElement("span");
+	newElement.appendChild(doc.createTextNode(item->text().mid(pos)));
+	newElement.setAttribute("class", "ocrx_word");
+	QMap<QString, QString> attrs = item->getTitleAttributes();
+	attrs["bbox"] = QString("%1 %2 %3 %4").arg(rightBBox.left()).arg(rightBBox.top()).arg(rightBBox.right()).arg(rightBBox.bottom());
+	newElement.setAttribute("title", HOCRItem::serializeAttrGroup(attrs));
+	newElement.setAttribute("lang", item->lang());
+	HOCRItem* newItem = new HOCRItem(newElement, item->page(), item->parent(), item->index() + 1);
+	beginInsertRows(itemIndex.parent(), item->index() + 1, item->index() + 1);
+	insertItem(item->parent(), newItem, item->index() + 1);
+	endInsertRows();
+	setData(itemIndex, item->text().left(pos), Qt::EditRole);
+	editItemAttribute(itemIndex, "title:bbox", QString("%1 %2 %3 %4").arg(leftBBox.left()).arg(leftBBox.top()).arg(leftBBox.right()).arg(leftBBox.bottom()));
+	return itemIndex.sibling(itemIndex.row() + 1, 0);
+}
+
+QModelIndex HOCRDocument::mergeItemText(const QModelIndex& itemIndex, bool mergeNext) {
+	HOCRItem* item = mutableItemAtIndex(itemIndex);
+	if(!item || item->itemClass() != "ocrx_word") {
+		return QModelIndex();
+	}
+	if((!mergeNext && item->index() == 0) || (mergeNext && item->index() == item->parent()->children().size() - 1)) {
+		return QModelIndex();
+	}
+	int offset = mergeNext ? 1 : -1;
+	HOCRItem* otherItem = item->parent()->children()[item->index() + offset];
+	QString newText = mergeNext ? item->text() + otherItem->text() : otherItem->text() + item->text();
+	QRect newBbox = item->bbox().united(otherItem->bbox());
+	setData(itemIndex, newText, Qt::EditRole);
+	editItemAttribute(itemIndex, "title:bbox", QString("%1 %2 %3 %4").arg(newBbox.left()).arg(newBbox.top()).arg(newBbox.right()).arg(newBbox.bottom()));
+	beginRemoveRows(itemIndex.parent(), itemIndex.row() + offset, itemIndex.row() + offset);
+	deleteItem(otherItem);
+	endRemoveRows();
+	return itemIndex;
+}
+
 QModelIndex HOCRDocument::addItem(const QModelIndex& parent, const QDomElement& element) {
 	HOCRItem* parentItem = mutableItemAtIndex(parent);
 	if(!parentItem) {
@@ -241,7 +360,7 @@ bool HOCRDocument::removeItem(const QModelIndex& index) {
 	return true;
 }
 
-QModelIndex HOCRDocument::nextIndex(const QModelIndex& current) {
+QModelIndex HOCRDocument::nextIndex(const QModelIndex& current) const {
 	QModelIndex idx = current;
 	// If the current index is invalid return first index
 	if(!idx.isValid()) {
@@ -249,7 +368,7 @@ QModelIndex HOCRDocument::nextIndex(const QModelIndex& current) {
 	}
 	// If item has children, return first child
 	if(rowCount(idx) > 0) {
-		return idx.child(0, 0);
+		return index(0, 0, idx);
 	}
 	// Return next possible sibling
 	QModelIndex next = idx.sibling(idx.row() + 1, 0);
@@ -264,7 +383,7 @@ QModelIndex HOCRDocument::nextIndex(const QModelIndex& current) {
 	return next;
 }
 
-QModelIndex HOCRDocument::prevIndex(const QModelIndex& current) {
+QModelIndex HOCRDocument::prevIndex(const QModelIndex& current) const {
 	QModelIndex idx = current;
 	// If the current index is invalid return last index
 	if(!idx.isValid()) {
@@ -281,62 +400,112 @@ QModelIndex HOCRDocument::prevIndex(const QModelIndex& current) {
 		idx = index(rowCount() - 1, 0);
 	}
 	while(rowCount(idx) > 0) {
-		idx = idx.child(rowCount(idx) - 1, 0);
+		idx = index(rowCount(idx) - 1, 0, idx);
 	}
 	return idx;
 }
 
-bool HOCRDocument::indexIsMisspelledWord(const QModelIndex& index) const {
-	return !checkItemSpelling(index);
+QModelIndex HOCRDocument::prevOrNextIndex(bool next, const QModelIndex& current, const QString& ocrClass, bool misspelled, bool lowconf) const {
+	QModelIndex start = current;
+	if(!start.isValid()) {
+		start = index(0, 0);
+	}
+	QModelIndex curr = next ? nextIndex(start) : prevIndex(start);
+	while(curr != start) {
+		const HOCRItem* item = itemAtIndex(curr);
+		if(item && item->itemClass() == ocrClass && (!misspelled || indexIsMisspelledWord(curr)) && (!lowconf || item->getTitleAttributes()["x_wconf"].toInt() < 90)) {
+			break;
+		}
+		curr = next ? nextIndex(curr) : prevIndex(curr);
+	};
+	return curr;
 }
 
-bool HOCRDocument::checkItemSpelling(const QModelIndex& index, QStringList* suggestions, int limit) const {
+bool HOCRDocument::indexIsMisspelledWord(const QModelIndex& index) const {
 	const HOCRItem* item = itemAtIndex(index);
-	if(item->itemClass() != "ocrx_word") { return true; }
+	if(item->isMisspelled() == -1) {
+		recheckItemSpelling(index);
+	}
+	return item->isMisspelled() == 1;
+}
+
+QList<QModelIndex> HOCRDocument::recheckItemSpelling(const QModelIndex& index) const {
+	HOCRItem* item = mutableItemAtIndex(index);
+	if(item->itemClass() != "ocrx_word") { return {}; }
 
 	QString prefix, suffix, trimmed = HOCRItem::trimmedWord(item->text(), &prefix, &suffix);
-	if(trimmed.isEmpty()) { return true; }
+	if(trimmed.isEmpty()) {
+		item->setMisspelled(false);
+		return {index};
+	}
 	QString lang = item->spellingLang();
-	if(m_spell->getLanguage() != lang && !(m_spell->setLanguage(lang))) { return true; }
+	if(lang.isEmpty() || (m_spell->getLanguage() != lang && !(m_spell->setLanguage(lang)))) {
+		item->setMisspelled(false);
+		return {index};
+	}
 
 	// check word, including (if requested) setting suggestions; handle hyphenated phrases correctly
-	if(checkSpelling(trimmed, suggestions, limit)) { return true; }
+	if(m_spell->checkSpelling(trimmed)) {
+		item->setMisspelled(false);
+		return {index};
+	}
+	item->setMisspelled(true);
 
 	// handle some hyphenated words
 	// don't bother with words not broken over sibling text lines (ie interrupted by other blocks), it's human hard
 	// don't bother with words broken over three or more lines, it's implausible and this treatment is ^ necessarily incomplete
-	HOCRItem* parent = item->parent();
-	if(!parent) { return false; }
-	HOCRItem* grandparent = parent->parent();
-	if(!grandparent) { return false; }
-	int idx = item->index();
-	int parentIdx = parent->index();
-	QVector<HOCRItem*> parentSiblings = grandparent->children();
-	if(idx == 0 && parentIdx > 0) {
-		HOCRItem* parentPrevSibling = parentSiblings.at(parentIdx - 1);
-		if(!parentPrevSibling) { return false; }
-		QVector<HOCRItem*> cousins = parentPrevSibling->children();
-		if(cousins.size() < 1) { return false; }
-		HOCRItem* prevCousin = cousins.back();
-		if(!prevCousin || prevCousin->itemClass() != "ocrx_word") { return false; }
-		QString prevText = prevCousin->text();
-		if(prevText.isEmpty() || prevText[prevText.size() - 1] != '-') { return false; }
+	HOCRItem* line = item->parent();
+	if(!line) { return {index}; }
+	HOCRItem* paragraph = line->parent();
+	if(!paragraph) { return {index}; }
+	int lineIdx = line->index();
+	const QVector<HOCRItem*>& paragraphLines = paragraph->children();
+	if(item == line->children().front() && lineIdx > 0) {
+		HOCRItem* prevLine = paragraphLines.at(lineIdx - 1);
+		if(!prevLine || prevLine->children().isEmpty()) { return {index}; }
+		HOCRItem* prevWord = prevLine->children().back();
+		if(!prevWord || prevWord->itemClass() != "ocrx_word") { return {index}; }
+		QString prevText = prevWord->text();
+		if(!prevText.endsWith("-")) { return {index}; }
 
 		// don't bother with (reassembled) suggestions for broken words since we can't re-break them
-		return checkSpelling(HOCRItem::trimmedWord(prevText) + trimmed);
-	}
-	QString text = item->text();
-	if(idx + 1 == parent->children().size() && parentIdx + 1 < parentSiblings.size() && !text.isEmpty() && text[text.size() - 1] == '-') {
-		HOCRItem* parentNextSibling = parentSiblings.at(parentIdx + 1);
-		if(!parentNextSibling) { return false; }
-		QVector<HOCRItem*> cousins = parentNextSibling->children();
-		if(cousins.size() < 1) { return false; }
-		HOCRItem* nextCousin = cousins.front();
-		if(!nextCousin || nextCousin->itemClass() != "ocrx_word") { return false; }
+		bool valid = m_spell->checkSpelling(HOCRItem::trimmedWord(prevText) + trimmed);
+		item->setMisspelled(!valid);
+		prevWord->setMisspelled(!valid);
+		return {index, indexAtItem(prevWord)};
+	} else if(item == line->children().back() && lineIdx + 1 < paragraphLines.size() && item->text().endsWith("-")) {
+		HOCRItem* nextLine = paragraphLines.at(lineIdx + 1);
+		if(!nextLine || nextLine->children().isEmpty()) { return {index}; }
+		HOCRItem* nextWord = nextLine->children().front();
+		if(!nextWord || nextWord->itemClass() != "ocrx_word") { return {index}; }
 
-		return checkSpelling(trimmed + HOCRItem::trimmedWord(nextCousin->text()));
+		bool valid = m_spell->checkSpelling(trimmed + HOCRItem::trimmedWord(nextWord->text()));
+		item->setMisspelled(!valid);
+		nextWord->setMisspelled(!valid);
+		return {index, indexAtItem(nextWord)};
 	}
-	return false;
+	return {index};
+}
+
+bool HOCRDocument::getItemSpellingSuggestions(const QModelIndex& index, QString& trimmedWord, QStringList& suggestions, int limit) const {
+	const HOCRItem* item = itemAtIndex(index);
+	if(item->itemClass() != "ocrx_word") { return true; }
+
+	QString prefix, suffix;
+	trimmedWord = HOCRItem::trimmedWord(item->text(), &prefix, &suffix);
+	if(trimmedWord.isEmpty()) {
+		return true;
+	}
+	QString lang = item->spellingLang();
+	if(m_spell->getLanguage() != lang && !(m_spell->setLanguage(lang))) {
+		return true;
+	}
+
+	bool valid = m_spell->checkSpelling(trimmedWord, &suggestions, limit);
+	for(int i = 0, n = suggestions.size(); i < n; ++i) {
+		suggestions[i] = prefix + suggestions[i] + suffix;
+	}
+	return valid;
 }
 
 bool HOCRDocument::referencesSource(const QString& filename) const {
@@ -359,18 +528,18 @@ QModelIndex HOCRDocument::searchPage(const QString& filename, int pageNr) const
 }
 
 QModelIndex HOCRDocument::searchAtCanvasPos(const QModelIndex& pageIndex, const QPoint& pos) const {
-	QModelIndex index = pageIndex;
-	if(!index.isValid()) {
+	QModelIndex idx = pageIndex;
+	if(!idx.isValid()) {
 		return QModelIndex();
 	}
-	const HOCRItem* item = itemAtIndex(index);
+	const HOCRItem* item = itemAtIndex(idx);
 	while(true) {
 		int iChild = 0, nChildren = item->children().size();
 		for(; iChild < nChildren; ++iChild) {
 			const HOCRItem* childItem = item->children()[iChild];
 			if(childItem->bbox().contains(pos)) {
 				item = childItem;
-				index = index.child(iChild, 0);
+				idx = index(iChild, 0, idx);
 				break;
 			}
 		}
@@ -378,7 +547,7 @@ QModelIndex HOCRDocument::searchAtCanvasPos(const QModelIndex& pageIndex, const
 			break;
 		}
 	}
-	return index;
+	return idx;
 }
 
 void HOCRDocument::convertSourcePaths(const QString& basepath, bool absolute) {
@@ -401,6 +570,8 @@ QVariant HOCRDocument::data(const QModelIndex& index, int role) const {
 			return displayRoleForItem(item);
 		case Qt::DecorationRole:
 			return decorationRoleForItem(item);
+		case Qt::ToolTipRole:
+			return tooltipRoleForItem(item);
 		case Qt::ForegroundRole: {
 			bool enabled = item->isEnabled();
 			const HOCRItem* parent = item->parent();
@@ -409,9 +580,9 @@ QVariant HOCRDocument::data(const QModelIndex& index, int role) const {
 				parent = parent->parent();
 			}
 			if(enabled) {
-				return checkItemSpelling(index) ? QVariant() : QVariant(QColor(Qt::red));
+				return indexIsMisspelledWord(index) ? QVariant(QColor(Qt::red)) : QVariant();
 			} else {
-				return checkItemSpelling(index) ? QVariant(QColor(Qt::gray)) : QVariant(QColor(208, 80, 82));
+				return indexIsMisspelledWord(index) ? QVariant(QColor(208, 80, 82)) : QVariant(QColor(Qt::gray));
 			}
 		}
 		case Qt::CheckStateRole:
@@ -435,31 +606,24 @@ bool HOCRDocument::setData(const QModelIndex& index, const QVariant& value, int
 	HOCRItem* item = mutableItemAtIndex(index);
 	if(role == Qt::EditRole && item->itemClass() == "ocrx_word") {
 		item->setText(value.toString());
-		emit dataChanged(index, index, {Qt::DisplayRole, Qt::ForegroundRole});
+		for(QModelIndex changedIndex : recheckItemSpelling(index)) {
+			emit dataChanged(changedIndex, changedIndex, {Qt::DisplayRole, Qt::ForegroundRole});
+		}
+
 		return true;
 	} else if(role == Qt::CheckStateRole) {
 		item->setEnabled(value == Qt::Checked);
-		emit dataChanged(index, index, {Qt::CheckStateRole});
-		recursiveDataChanged(index, {Qt::CheckStateRole});
+		// Get leaf
+		QModelIndex leaf = index;
+		while(hasChildren(leaf)) {
+			leaf = this->index(rowCount(leaf) - 1, 0, leaf);
+		}
+		emit dataChanged(index, leaf, {Qt::CheckStateRole});
 		return true;
 	}
 	return false;
 }
 
-void HOCRDocument::recursiveDataChanged(const QModelIndex& parent, const QVector<int>& roles, const QStringList& itemClasses) {
-	int rows = rowCount(parent);
-	if(rows > 0) {
-		QModelIndex firstChild = index(0, 0, parent);
-		QString childItemClass = itemAtIndex(firstChild)->itemClass();
-		if(itemClasses.isEmpty() || itemClasses.contains(childItemClass)) {
-			emit dataChanged(firstChild, index(rows - 1, 0, parent), roles);
-		}
-		for(int i = 0; i < rows; ++i) {
-			recursiveDataChanged(index(i, 0, parent), roles, itemClasses);
-		}
-	}
-}
-
 void HOCRDocument::recomputeBBoxes(HOCRItem* item) {
 	// Update parent bboxes (except page)
 	while(item && item->parent()) {
@@ -475,7 +639,7 @@ void HOCRDocument::recomputeBBoxes(HOCRItem* item) {
 
 Qt::ItemFlags HOCRDocument::flags(const QModelIndex& index) const {
 	if (!index.isValid()) {
-		return 0;
+		return Qt::ItemFlags();
 	}
 
 	HOCRItem* item = mutableItemAtIndex(index);
@@ -525,7 +689,7 @@ QString HOCRDocument::displayRoleForItem(const HOCRItem* item) const {
 	QString itemClass = item->itemClass();
 	if(itemClass == "ocr_page") {
 		const HOCRPage* page = static_cast<const HOCRPage*>(item);
-		return QString("%1 (%2 %3/%4)").arg(page->title()).arg(tr("Page")).arg(item->index() + 1).arg(m_pages.size());
+		return QString("%1 (%2 %3/%4)").arg(page->title()).arg(_("Page")).arg(item->index() + 1).arg(m_pages.size());
 	} else if(itemClass == "ocr_carea") {
 		return _("Text block");
 	} else if(itemClass == "ocr_par") {
@@ -558,60 +722,13 @@ QIcon HOCRDocument::decorationRoleForItem(const HOCRItem* item) const {
 	return QIcon();
 }
 
-bool HOCRDocument::checkSpelling(const QString& trimmed, QStringList* suggestions, int limit) const {
-	QVector<QStringRef> words = trimmed.splitRef(QRegExp("[\\x2013\\x2014]+"));
-	int perWordLimit = 0;
-	// s = p^w => w = log_p(c) = log(c)/log(p) => p = 10^(log(c)/w)
-	if(limit > 0) { perWordLimit = int(std::pow(10, std::log10(limit) / words.size())); }
-	QList<QList<QString>> wordSuggestions;
-	bool valid = true;
-	bool multipleWords = words.size() > 1;
-	for(const QStringRef& word : words) {
-		QString wordString = word.toString();
-		bool wordValid = m_spell->checkWord(wordString);
-		valid &= wordValid;
-		if(suggestions) {
-			QList<QString> ws = m_spell->getSpellingSuggestions(wordString);
-			if(wordValid && multipleWords) { ws.prepend(wordString); }
-			if(limit == -1) {
-				wordSuggestions.append(ws);
-			} else if(limit > 0) {
-				wordSuggestions.append(ws.mid(0, perWordLimit));
-			}
-		}
-	}
-	if(suggestions) {
-		suggestions->clear();
-		QList<QList<QString>> suggestionCombinations;
-		generateCombinations(wordSuggestions, suggestionCombinations, 0, QList<QString>());
-		for(const QList<QString>& combination : suggestionCombinations) {
-			QString s = "";
-			const QStringRef* originalWord = words.begin();
-			int last = 0;
-			int next;
-			for(const QString& suggestedWord : combination) {
-				next = originalWord->position();
-				s.append(trimmed.midRef(last, next - last));
-				s.append(suggestedWord);
-				last = next + originalWord->length();
-				originalWord++;
-			}
-			s.append(trimmed.midRef(last));
-			suggestions->append(s);
-		}
-	}
-	return valid;
-}
-
-// each suggestion for each word => each word in each suggestion
-void HOCRDocument::generateCombinations(const QList<QList<QString>>& lists, QList<QList<QString>>& results, int depth, const QList<QString> c) const {
-	if(depth == lists.size()) {
-		results.append(c);
-		return;
-	}
-	for(int i = 0; i < lists[depth].size(); ++i) {
-		generateCombinations(lists, results, depth + 1, c + QList<QString>({lists[depth][i]}));
+QString HOCRDocument::tooltipRoleForItem(const HOCRItem* item) const {
+	QString itemClass = item->itemClass();
+	if(itemClass == "ocr_page") {
+		const HOCRPage* page = static_cast<const HOCRPage*>(item);
+		return QString("%1 (%2 %3/%4)").arg(page->title()).arg(_("Page")).arg(item->index() + 1).arg(m_pages.size());
 	}
+	return QString();
 }
 
 void HOCRDocument::insertItem(HOCRItem* parent, HOCRItem* item, int i) {
@@ -651,8 +768,8 @@ QMap<QString, QString> HOCRItem::s_langCache = QMap<QString, QString>();
 
 QMap<QString, QString> HOCRItem::deserializeAttrGroup(const QString& string) {
 	QMap<QString, QString> attrs;
-	for(const QString& attr : string.split(QRegExp("\\s*;\\s*"))) {
-		int splitPos = attr.indexOf(QRegExp("\\s+"));
+	for(const QString& attr : string.split(QRegularExpression("\\s*;\\s*"))) {
+		int splitPos = attr.indexOf(QRegularExpression("\\s+"));
 		attrs.insert(attr.left(splitPos), splitPos > 0 ? attr.mid(splitPos + 1) : "");
 	}
 	return attrs;
@@ -667,16 +784,18 @@ QString HOCRItem::serializeAttrGroup(const QMap<QString, QString>& attrs) {
 }
 
 QString HOCRItem::trimmedWord(const QString& word, QString* prefix, QString* suffix) {
+	QString noaccentWord = Utils::removeDiacritics(word);
 	// correctly trim words with apostrophes or hyphens within them, phrases with dashes, initialisms/acronyms, and numeric citations
-	QRegExp wordRe("^(\\W*)(\\w?|\\w(\\w|[-\\x2013\\x2014'’])*\\w|(\\w+\\.){2,})([\\W\\x00b2\\x00b3\\x00b9\\x2070-\\x207e]*)$");
-	if(wordRe.indexIn(word) != -1) {
+	QRegularExpression wordRe("^(\\W*)(\\w?|\\w(\\w|[-\u2013\u2014'’])*\\w|(\\w+\\.){2,})([\\W\u00b2\u00b3\u00b9\u2070\\-\u207e]*)$");
+	QRegularExpressionMatch match;
+	if((match = wordRe.match(noaccentWord)).hasMatch()) {
 		if(prefix) {
-			*prefix = wordRe.cap(1);
+			*prefix = match.captured(1);
 		}
 		if(suffix) {
-			*suffix = wordRe.cap(5);
+			*suffix = match.captured(5);
 		}
-		return wordRe.cap(2);
+		return word.mid(match.captured(1).length(), word.length() - match.captured(1).length() - match.captured(5).length());
 	}
 	return word;
 }
@@ -693,6 +812,10 @@ HOCRItem::HOCRItem(const QDomElement& element, HOCRPage* page, HOCRItem* parent,
 			m_attrs[attrName] = attributes.item(i).nodeValue();
 		}
 	}
+	// Map ocr_header/ocr_caption/ocr_textfloat to ocr_line
+	if(m_attrs["class"] == "ocr_header" || m_attrs["class"] == "ocr_caption" || m_attrs["class"] == "ocr_textfloat") {
+		m_attrs["class"] = "ocr_line";
+	}
 	// Adjust item id based on pageId
 	if(parent) {
 		QString idClass = itemClass().mid(itemClass().indexOf("_") + 1);
@@ -703,7 +826,7 @@ HOCRItem::HOCRItem(const QDomElement& element, HOCRPage* page, HOCRItem* parent,
 	}
 
 	// Parse item bbox
-	QStringList bbox = m_titleAttrs["bbox"].split(QRegExp("\\s+"));
+	QStringList bbox = m_titleAttrs["bbox"].split(QRegularExpression("\\s+"));
 	if(bbox.size() == 4) {
 		m_bbox.setCoords(bbox[0].toInt(), bbox[1].toInt(), bbox[2].toInt(), bbox[3].toInt());
 	}
@@ -715,6 +838,9 @@ HOCRItem::HOCRItem(const QDomElement& element, HOCRPage* page, HOCRItem* parent,
 	} else if(itemClass() == "ocr_line") {
 		// Depending on the locale, tesseract can use a comma instead of a dot as decimal separator in the baseline...
 		m_titleAttrs["baseline"] = m_titleAttrs["baseline"].replace(",", ".");
+		m_misspelled = false;
+	} else {
+		m_misspelled = false;
 	}
 }
 
@@ -839,7 +965,7 @@ void HOCRItem::setAttribute(const QString& name, const QString& value, const QSt
 		Q_ASSERT(parts[0] == "title");
 		m_titleAttrs[parts[1]] = value;
 		if(name == "title:bbox") {
-			QStringList bbox = value.split(QRegExp("\\s+"));
+			QStringList bbox = value.split(QRegularExpression("\\s+"));
 			Q_ASSERT(bbox.size() == 4);
 			m_bbox.setCoords(bbox[0].toInt(), bbox[1].toInt(), bbox[2].toInt(), bbox[3].toInt());
 		}
@@ -869,11 +995,7 @@ QString HOCRItem::toHtml(int indent) const {
 		if(m_italic) {
 			html += "<em>";
 		}
-#if QT_VERSION >= QT_VERSION_CHECK(5, 0, 0)
 		html += m_text.toHtmlEscaped();
-#else
-		html += Qt::escape(m_text);
-#endif
 		if(m_italic) {
 			html += "</em>";
 		}
@@ -892,20 +1014,21 @@ QString HOCRItem::toHtml(int indent) const {
 }
 
 QPair<double, double> HOCRItem::baseLine() const {
-	static const QRegExp baseLineRx = QRegExp("([+-]?\\d+\\.?\\d*)\\s+([+-]?\\d+\\.?\\d*)");
-	if(baseLineRx.indexIn(m_titleAttrs["baseline"]) != -1) {
-		return qMakePair(baseLineRx.cap(1).toDouble(), baseLineRx.cap(2).toDouble());
+	static const QRegularExpression baseLineRx = QRegularExpression("([+-]?\\d+\\.?\\d*)\\s+([+-]?\\d+\\.?\\d*)");
+	QRegularExpressionMatch match;
+	if((match = baseLineRx.match(m_titleAttrs["baseline"])).hasMatch()) {
+		return qMakePair(match.captured(1).toDouble(), match.captured(2).toDouble());
 	}
 	return qMakePair(0.0, 0.0);
 }
 
-bool HOCRItem::parseChildren(const QDomElement& element, QString language) {
+bool HOCRItem::parseChildren(const QDomElement& element, QString language, const QString& defaultLanguage) {
 	// Determine item language (inherit from parent if not specified)
 	QString elemLang = element.attribute("lang");
 	if(!elemLang.isEmpty()) {
 		auto it = s_langCache.find(elemLang);
 		if(it == s_langCache.end()) {
-			it = s_langCache.insert(elemLang, Utils::getSpellingLanguage(elemLang));
+			it = s_langCache.insert(elemLang, Utils::getSpellingLanguage(elemLang, defaultLanguage));
 		}
 		m_attrs.remove("lang");
 		language = it.value();
@@ -919,7 +1042,7 @@ bool HOCRItem::parseChildren(const QDomElement& element, QString language) {
 	QDomElement childElement = element.firstChildElement();
 	while(!childElement.isNull()) {
 		m_childItems.append(new HOCRItem(childElement, m_pageItem, this, m_childItems.size()));
-		haveWords |= m_childItems.last()->parseChildren(childElement, language);
+		haveWords |= m_childItems.last()->parseChildren(childElement, language, defaultLanguage);
 		childElement = childElement.nextSiblingElement();
 	}
 	return haveWords;
@@ -927,11 +1050,11 @@ bool HOCRItem::parseChildren(const QDomElement& element, QString language) {
 
 ///////////////////////////////////////////////////////////////////////////////
 
-HOCRPage::HOCRPage(const QDomElement& element, int pageId, const QString& language, bool cleanGraphics, int index)
+HOCRPage::HOCRPage(const QDomElement& element, int pageId, const QString& defaultLanguage, bool cleanGraphics, int index)
 	: HOCRItem(element, this, nullptr, index), m_pageId(pageId) {
 	m_attrs["id"] = QString("page_%1").arg(pageId);
 
-	m_sourceFile = m_titleAttrs["image"].replace(QRegExp("^['\"]"), "").replace(QRegExp("['\"]$"), "");
+	m_sourceFile = m_titleAttrs["image"].replace(QRegularExpression("^['\"]"), "").replace(QRegularExpression("['\"]$"), "");
 	m_pageNr = m_titleAttrs["ppageno"].toInt();
 	// Code to handle pageno -> ppageno typo in previous versions of gImageReader
 	if(m_pageNr == 0) {
@@ -940,13 +1063,13 @@ HOCRPage::HOCRPage(const QDomElement& element, int pageId, const QString& langua
 		m_titleAttrs.remove("pageno");
 	}
 	m_angle = m_titleAttrs["rot"].toDouble();
-	m_resolution = m_titleAttrs["res"].toInt();
+	m_resolution = m_titleAttrs.value("res", "100").toInt();
 
 	QDomElement childElement = element.firstChildElement("div");
 	while(!childElement.isNull()) {
 		HOCRItem* item = new HOCRItem(childElement, this, this, m_childItems.size());
 		m_childItems.append(item);
-		if(!item->parseChildren(childElement, language)) {
+		if(!item->parseChildren(childElement, defaultLanguage, defaultLanguage)) {
 			// No word children -> treat as graphic
 			if(cleanGraphics && (item->bbox().width() < 10 || item->bbox().height() < 10)) {
 				// Ignore graphics which are less than 10 x 10
@@ -969,7 +1092,7 @@ void HOCRPage::convertSourcePath(const QString& basepath, bool absolute) {
 	if(absolute && !QFileInfo(m_sourceFile).isAbsolute()) {
 		m_sourceFile = QDir::cleanPath(QDir(basepath).absoluteFilePath(m_sourceFile));
 	} else if(!absolute && QFileInfo(m_sourceFile).isAbsolute() && m_sourceFile.startsWith(basepath)) {
-		m_sourceFile = QString(".%1%2").arg("/").arg(QDir(basepath).relativeFilePath(m_sourceFile));
+		m_sourceFile = QString("./%1").arg(QDir(basepath).relativeFilePath(m_sourceFile));
 	}
 	m_titleAttrs["image"] = QString("'%1'").arg(m_sourceFile);
 }
