@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * HOCROdtExporter.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -18,17 +18,17 @@
  */
 
 #include "ConfigSettings.hh"
-#include "DisplayerToolHOCR.hh"
+#include "Displayer.hh"
 #include "HOCRDocument.hh"
 #include "HOCROdtExporter.hh"
 #include "MainWindow.hh"
-#include "FileDialogs.hh"
 #include "SourceManager.hh"
 #include "Utils.hh"
 
 #include <QDesktopServices>
 #include <QFileInfo>
 #include <QMessageBox>
+#include <QUrl>
 #include <QUuid>
 #include <QXmlStreamWriter>
 #include <quazipfile.h>
@@ -43,19 +43,7 @@ static QString drawNS ("urn:oasis:names:tc:opendocument:xmlns:drawing:1.0");
 static QString xlinkNS ("http://www.w3.org/1999/xlink");
 static QString svgNS ("urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0");
 
-bool HOCROdtExporter::run(const HOCRDocument* hocrdocument, QString& filebasename) {
-	QString suggestion = filebasename;
-	if(suggestion.isEmpty()) {
-		QList<Source*> sources = MAIN->getSourceManager()->getSelectedSources();
-		suggestion = !sources.isEmpty() ? QFileInfo(sources.first()->displayname).baseName() : _("output");
-	}
-
-	QString outname = FileDialogs::saveDialog(_("Save ODT Output..."), suggestion + ".odt", "outputdir", QString("%1 (*.odt)").arg(_("OpenDocument Text Documents")));
-	if(outname.isEmpty()) {
-		return false;
-	}
-	filebasename = QFileInfo(outname).completeBaseName();
-
+bool HOCROdtExporter::run(const HOCRDocument* hocrdocument, const QString& outname, const ExporterSettings* /*settings*/) {
 	QuaZip zip(outname);
 	if(!zip.open( QuaZip::mdCreate )) {
 		QMessageBox::warning(MAIN, _("Export failed"), _("The ODT export failed: unable to write output file."));
@@ -433,5 +421,6 @@ bool HOCROdtExporter::setSource(const QString& sourceFile, int page, int dpi, do
 }
 
 QImage HOCROdtExporter::getSelection(const QRect& bbox) {
-	return m_displayerTool->getSelection(bbox);
+	Displayer* displayer = MAIN->getDisplayer();
+	return displayer->getImage(bbox.translated(displayer->getSceneBoundingRect().toRect().topLeft()));
 }
