@@ -5,7 +5,7 @@
  *   Copyright (C) 2009-2013 Canonical Ltd.
  *   Author: Robert Ancell <robert.ancell@canonical.com>
  * Modifications are:
- *   Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ *   Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -31,7 +31,7 @@
 
 void ScannerSane::init() {
 	moveToThread(&m_thread);
-	connect(&m_thread, SIGNAL(started()), this, SLOT(run()));
+	connect(&m_thread, &QThread::started, this, &ScannerSane::run);
 	m_thread.start();
 }
 
@@ -511,7 +511,8 @@ void ScannerSane::doCompletePage() {
 	QString filename = m_job->params.filename;
 	if(m_job->params.type != ScanType::SINGLE) {
 		QFileInfo finfo(filename);
-		filename = QString("%1/%2_%3.%4").arg(finfo.absolutePath()).arg(finfo.completeBaseName()).arg(m_job->pageNumber).arg(finfo.completeSuffix());
+		filename = QString("%1/%2-%3.%4").arg(finfo.absolutePath()).arg(finfo.completeBaseName()).arg(m_job->pageNumber).arg(finfo.completeSuffix());
+		filename = Utils::makeOutputFilename(filename);
 	}
 	if(!QImage(m_job->imgbuf.data(), m_job->rowstride / 3, m_job->height, m_job->rowstride, QImage::Format_RGB888).save(filename)) {
 		failScan(_("Failed to save image"));
