@@ -1,7 +1,7 @@
 /* -*- Mode: C++; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
 /*
  * ScannerTwain.cc
- * Copyright (C) 2013-2019 Sandro Mani <manisandro@gmail.com>
+ * Copyright (C) 2013-2022 Sandro Mani <manisandro@gmail.com>
  *
  * gImageReader is free software: you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the
@@ -167,7 +167,7 @@ void ScannerTwain::redetect() {
 
 	// Loop through all the sources
 	while (twRC == TWRC_SUCCESS) {
-		sources.append({sourceID.ProductName, sourceID.ProductName});
+		sources.append(Device{sourceID.ProductName, sourceID.ProductName});
 		twRC = call(nullptr, DG_CONTROL, DAT_IDENTITY, MSG_GETNEXT, &sourceID);
 	}
 	emit devicesDetected(sources);
@@ -181,7 +181,7 @@ void ScannerTwain::scan(const Params& params) {
 	/** Open device **/
 	emit scanStateChanged(State::OPEN);
 
-	std::strncpy(m_srcID.ProductName, params.device.toLocal8Bit().data(), sizeof(m_srcID.ProductName));
+	std::strncpy(m_srcID.ProductName, params.device.toLocal8Bit().data(), sizeof(m_srcID.ProductName) - 1);
 	// State 3 to 4
 	if(call(nullptr, DG_CONTROL, DAT_IDENTITY, MSG_OPENDS, &m_srcID) != TWRC_SUCCESS) {
 		failScan(_("Unable to connect to scanner"));
